@@ -3,7 +3,7 @@
  *  Here we test manipulations on GiNaC's indexed objects. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -55,7 +55,7 @@ static unsigned check_equal_simplify(const ex &e1, const ex &e2, const scalar_pr
 	return 0;
 }
 
-static unsigned delta_check(void)
+static unsigned delta_check()
 {
 	// checks identities of the delta tensor
 
@@ -81,7 +81,7 @@ static unsigned delta_check(void)
 	return result;
 }
 
-static unsigned metric_check(void)
+static unsigned metric_check()
 {
 	// checks identities of the metric tensor
 
@@ -114,7 +114,7 @@ static unsigned metric_check(void)
 	return result;
 }
 
-static unsigned epsilon_check(void)
+static unsigned epsilon_check()
 {
 	// checks identities of the epsilon tensor
 
@@ -136,8 +136,10 @@ static unsigned epsilon_check(void)
 	// contraction with symmetric tensor is zero
 	result += check_equal_simplify(lorentz_eps(mu, nu, rho, sigma) * indexed(d, sy_symm(), mu_co, nu_co), 0);
 	result += check_equal_simplify(lorentz_eps(mu, nu, rho, sigma) * indexed(d, sy_symm(), nu_co, sigma_co, rho_co), 0);
-	ex e = lorentz_eps(mu, nu, rho, sigma) * indexed(d, sy_symm(), mu_co, tau);
-	result += check_equal_simplify(e, e);
+	result += check_equal_simplify(lorentz_eps(mu, nu, rho, sigma) * indexed(d, mu_co) * indexed(d, nu_co), 0);
+	result += check_equal_simplify(lorentz_eps(mu_co, nu, rho, sigma) * indexed(d, mu) * indexed(d, nu_co), 0);
+	ex e = lorentz_eps(mu, nu, rho, sigma) * indexed(d, mu_co) - lorentz_eps(mu_co, nu, rho, sigma) * indexed(d, mu);
+	result += check_equal_simplify(e, 0);
 
 	// contractions of epsilon tensors
 	result += check_equal_simplify(lorentz_eps(mu, nu, rho, sigma) * lorentz_eps(mu_co, nu_co, rho_co, sigma_co), -24);
@@ -146,14 +148,19 @@ static unsigned epsilon_check(void)
 	return result;
 }
 
-static unsigned symmetry_check(void)
+DECLARE_FUNCTION_2P(symm_fcn)
+REGISTER_FUNCTION(symm_fcn, set_symmetry(sy_symm(0, 1)));
+DECLARE_FUNCTION_2P(anti_fcn)
+REGISTER_FUNCTION(anti_fcn, set_symmetry(sy_anti(0, 1)));
+
+static unsigned symmetry_check()
 {
 	// check symmetric/antisymmetric objects
 
 	unsigned result = 0;
 
 	idx i(symbol("i"), 3), j(symbol("j"), 3), k(symbol("k"), 3), l(symbol("l"), 3);
-	symbol A("A"), B("B");
+	symbol A("A"), B("B"), C("C");
 	ex e;
 
 	result += check_equal(indexed(A, sy_symm(), i, j), indexed(A, sy_symm(), j, i));
@@ -185,10 +192,17 @@ static unsigned symmetry_check(void)
 	result += check_equal(symmetrize(e), 0);
 	result += check_equal(antisymmetrize(e), e);
 
+	e = (indexed(A, sy_anti(), i, j, k, l) * (indexed(B, j) * indexed(C, k) + indexed(B, k) * indexed(C, j)) + indexed(B, i, l)).expand();
+	result += check_equal_simplify(e, indexed(B, i, l));
+
+	result += check_equal(symm_fcn(0, 1) + symm_fcn(1, 0), 2*symm_fcn(0, 1));
+	result += check_equal(anti_fcn(0, 1) + anti_fcn(1, 0), 0);
+	result += check_equal(anti_fcn(0, 0), 0);
+
 	return result;
 }
 
-static unsigned scalar_product_check(void)
+static unsigned scalar_product_check()
 {
 	// check scalar product replacement
 
@@ -211,7 +225,7 @@ static unsigned scalar_product_check(void)
 	return result;
 }
 
-static unsigned edyn_check(void)
+static unsigned edyn_check()
 {
 	// Relativistic electrodynamics
 
@@ -227,19 +241,17 @@ static unsigned edyn_check(void)
 
 	// Lorentz transformation matrix (boost along x axis)
 	matrix L(4, 4);
-	L(0, 0) = gamma;
-	L(0, 1) = -beta*gamma;
-	L(1, 0) = -beta*gamma;
-	L(1, 1) = gamma;
-	L(2, 2) = 1; L(3, 3) = 1;
+	L =       gamma, -beta*gamma, 0, 0,
+	    -beta*gamma,       gamma, 0, 0,
+	              0,           0, 1, 0,
+	              0,           0, 0, 1;
 
 	// Electromagnetic field tensor
-	matrix F(4, 4, lst(
-		 0, -Ex, -Ey, -Ez,
+	matrix F(4, 4);
+	F =  0, -Ex, -Ey, -Ez,
 		Ex,   0, -Bz,  By,
 		Ey,  Bz,   0, -Bx,
-		Ez, -By,  Bx,   0
-	));
+		Ez, -By,  Bx,   0;
 
 	// Indices
 	symbol s_mu("mu"), s_nu("nu"), s_rho("rho"), s_sigma("sigma");
@@ -298,14 +310,14 @@ static unsigned edyn_check(void)
 	return result;
 }
 
-static unsigned spinor_check(void)
+static unsigned spinor_check()
 {
 	// check identities of the spinor metric
 
 	unsigned result = 0;
 
 	symbol psi("psi");
-	spinidx A(symbol("A"), 2), B(symbol("B"), 2), C(symbol("C"), 2);
+	spinidx A(symbol("A")), B(symbol("B")), C(symbol("C")), D(symbol("D"));
 	ex A_co = A.toggle_variance(), B_co = B.toggle_variance();
 	ex e;
 
@@ -327,13 +339,15 @@ static unsigned spinor_check(void)
 	result += check_equal_simplify(e, -indexed(psi, A_co));
 	e = spinor_metric(A_co, B_co) * indexed(psi, A);
 	result += check_equal_simplify(e, indexed(psi, B_co));
+	e = spinor_metric(D, A) * spinor_metric(A_co, B_co) * spinor_metric(B, C) - spinor_metric(D, A_co) * spinor_metric(A, B_co) * spinor_metric(B, C);
+	result += check_equal_simplify(e, 0);
 
 	return result;
 }
 
-static unsigned dummy_check(void)
+static unsigned dummy_check()
 {
-	// check dummy index renaming
+	// check dummy index renaming/repositioning
 
 	unsigned result = 0;
 
@@ -352,10 +366,19 @@ static unsigned dummy_check(void)
 	e = indexed(p, mu, mu.toggle_variance()) - indexed(p, nu, nu.toggle_variance());
 	result += check_equal_simplify(e, 0);
 
+	e = indexed(p, mu.toggle_variance(), nu, mu) * indexed(q, i)
+	  - indexed(p, mu, nu, mu.toggle_variance()) * indexed(q, i);
+	result += check_equal_simplify(e, 0);
+
+	e = indexed(p, mu, mu.toggle_variance()) - indexed(p, nu.toggle_variance(), nu);
+	result += check_equal_simplify(e, 0);
+	e = indexed(p, mu.toggle_variance(), mu) - indexed(p, nu, nu.toggle_variance());
+	result += check_equal_simplify(e, 0);
+
 	return result;
 }
 
-unsigned exam_indexed(void)
+unsigned exam_indexed()
 {
 	unsigned result = 0;
 	
