@@ -3,7 +3,7 @@
  *  Series expansion test (Laurent and Taylor series). */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -28,19 +28,21 @@ static unsigned check_series(const ex &e, const ex &point, const ex &d, int orde
 {
 	ex es = e.series(x==point, order);
 	ex ep = ex_to<pseries>(es).convert_to_poly();
-	if (!(ep - d).is_zero()) {
+	if (!(ep - d).expand().is_zero()) {
 		clog << "series expansion of " << e << " at " << point
 		     << " erroneously returned " << ep << " (instead of " << d
 		     << ")" << endl;
-		(ep-d).printtree(clog);
+		clog << tree << (ep-d) << dflt;
 		return 1;
 	}
 	return 0;
 }
 
 // Series expansion
-static unsigned exam_series1(void)
+static unsigned exam_series1()
 {
+	using GiNaC::log;
+
 	unsigned result = 0;
 	ex e, d;
 	
@@ -69,15 +71,15 @@ static unsigned exam_series1(void)
 	result += check_series(e, 1, d);
 	
 	e = pow(x + pow(x, 3), -1);
-	d = pow(x, -1) - x + pow(x, 3) - pow(x, 5) + Order(pow(x, 7));
+	d = pow(x, -1) - x + pow(x, 3) - pow(x, 5) + pow(x, 7) + Order(pow(x, 8));
 	result += check_series(e, 0, d);
 	
 	e = pow(pow(x, 2) + pow(x, 4), -1);
-	d = pow(x, -2) - 1 + pow(x, 2) - pow(x, 4) + Order(pow(x, 6));
+	d = pow(x, -2) - 1 + pow(x, 2) - pow(x, 4) + pow(x, 6) + Order(pow(x, 8));
 	result += check_series(e, 0, d);
 	
 	e = pow(sin(x), -2);
-	d = pow(x, -2) + numeric(1,3) + pow(x, 2) / 15 + pow(x, 4) * 2/189 + Order(pow(x, 5));
+	d = pow(x, -2) + numeric(1,3) + pow(x, 2) / 15 + pow(x, 4) * 2/189 + pow(x, 6) / 675  + Order(pow(x, 8));
 	result += check_series(e, 0, d);
 	
 	e = sin(x) / cos(x);
@@ -85,7 +87,7 @@ static unsigned exam_series1(void)
 	result += check_series(e, 0, d);
 	
 	e = cos(x) / sin(x);
-	d = pow(x, -1) - x / 3 - pow(x, 3) / 45 - pow(x, 5) * 2/945 + Order(pow(x, 6));
+	d = pow(x, -1) - x / 3 - pow(x, 3) / 45 - pow(x, 5) * 2/945 - pow(x, 7) / 4725 + Order(pow(x, 8));
 	result += check_series(e, 0, d);
 	
 	e = pow(numeric(2), x);
@@ -103,24 +105,48 @@ static unsigned exam_series1(void)
 	result += check_series(e, 0, d, 1);
 	result += check_series(e, 0, d, 2);
 	
+	e = pow(x, 8) * pow(pow(x,3)+ pow(x + pow(x,3), 2), -2);
+	d = pow(x, 4) - 2*pow(x, 5) + Order(pow(x, 6));
+	result += check_series(e, 0, d, 6);
+	
+	e = cos(x) * pow(sin(x)*(pow(x, 5) + 4 * pow(x, 2)), -3);
+	d = pow(x, -9) / 64 - 3 * pow(x, -6) / 256 - pow(x, -5) / 960 + 535 * pow(x, -3) / 96768
+	    + pow(x, -2) / 1280 - pow(x, -1) / 14400 - numeric(283, 129024) - 2143 * x / 5322240
+	    + Order(pow(x, 2));
+	result += check_series(e, 0, d, 2);
+	
+	e = sqrt(1+x*x) * sqrt(1+2*x*x);
+	d = 1 + Order(pow(x, 2));
+	result += check_series(e, 0, d, 2);
+
+	symbol a("a");
+	e = pow(x, 4) * sin(a) + pow(x, 2);
+	d = pow(x, 2) + Order(pow(x, 3));
+	result += check_series(e, 0, d, 3);
+
+	symbol b("b");
+	e = log(a*x + b*x*x*log(x));
+	d = log(a*x) + b/a*log(x)*x - pow(b/a, 2)/2*pow(log(x)*x, 2) + Order(pow(x, 3));
+	result += check_series(e, 0, d, 3);
+
 	return result;
 }
 
 // Series addition
-static unsigned exam_series2(void)
+static unsigned exam_series2()
 {
 	unsigned result = 0;
 	ex e, d;
 	
 	e = pow(sin(x), -1).series(x==0, 8) + pow(sin(-x), -1).series(x==0, 12);
-	d = Order(pow(x, 6));
+	d = Order(pow(x, 8));
 	result += check_series(e, 0, d);
 	
 	return result;
 }
 
 // Series multiplication
-static unsigned exam_series3(void)
+static unsigned exam_series3()
 {
 	unsigned result = 0;
 	ex e, d;
@@ -133,7 +159,7 @@ static unsigned exam_series3(void)
 }
 
 // Series exponentiation
-static unsigned exam_series4(void)
+static unsigned exam_series4()
 {
 	unsigned result = 0;
 	ex e, d;
@@ -142,15 +168,16 @@ static unsigned exam_series4(void)
 	d = 4 - 4*pow(x, 2) + 4*pow(x, 4)/3 + Order(pow(x, 5));
 	result += check_series(e, 0, d);
 	
-	e = pow(tgamma(x), 2).series(x==0, 3);
-	d = pow(x,-2) - 2*Euler/x + (pow(Pi,2)/6+2*pow(Euler,2)) + Order(x);
+	e = pow(tgamma(x), 2).series(x==0, 2);
+	d = pow(x,-2) - 2*Euler/x + (pow(Pi,2)/6+2*pow(Euler,2)) 
+		+ x*(-4*pow(Euler, 3)/3 -pow(Pi,2)*Euler/3 - 2*zeta(3)/3) + Order(pow(x, 2));
 	result += check_series(e, 0, d);
 	
 	return result;
 }
 
 // Order term handling
-static unsigned exam_series5(void)
+static unsigned exam_series5()
 {
 	unsigned result = 0;
 	ex e, d;
@@ -170,7 +197,7 @@ static unsigned exam_series5(void)
 }
 
 // Series expansion of tgamma(-1)
-static unsigned exam_series6(void)
+static unsigned exam_series6()
 {
 	ex e = tgamma(2*x);
 	ex d = pow(x+1,-1)*numeric(1,4) +
@@ -202,26 +229,25 @@ static unsigned exam_series6(void)
 }
 	
 // Series expansion of tan(x==Pi/2)
-static unsigned exam_series7(void)
+static unsigned exam_series7()
 {
 	ex e = tan(x*Pi/2);
 	ex d = pow(x-1,-1)/Pi*(-2) + pow(x-1,1)*Pi/6 + pow(x-1,3)*pow(Pi,3)/360
 	      +pow(x-1,5)*pow(Pi,5)/15120 + pow(x-1,7)*pow(Pi,7)/604800
-	      +Order(pow(x-1,8));
-	return check_series(e,1,d,8);
+	      +Order(pow(x-1,9));
+	return check_series(e,1,d,9);
 }
 
 // Series expansion of log(sin(x==0))
-static unsigned exam_series8(void)
+static unsigned exam_series8()
 {
 	ex e = log(sin(x));
-	ex d = log(x) - pow(x,2)/6 - pow(x,4)/180 - pow(x,6)/2835
-	      +Order(pow(x,8));
-	return check_series(e,0,d,8);
+	ex d = log(x) - pow(x,2)/6 - pow(x,4)/180 - pow(x,6)/2835 - pow(x,8)/37800 + Order(pow(x,9));
+	return check_series(e,0,d,9);
 }
 
 // Series expansion of Li2(sin(x==0))
-static unsigned exam_series9(void)
+static unsigned exam_series9()
 {
 	ex e = Li2(sin(x));
 	ex d = x + pow(x,2)/4 - pow(x,3)/18 - pow(x,4)/48
@@ -231,8 +257,10 @@ static unsigned exam_series9(void)
 }
 
 // Series expansion of Li2((x==2)^2), caring about branch-cut
-static unsigned exam_series10(void)
+static unsigned exam_series10()
 {
+	using GiNaC::log;
+
 	ex e = Li2(pow(x,2));
 	ex d = Li2(4) + (-log(3) + I*Pi*csgn(I-I*pow(x,2))) * (x-2)
 	       + (numeric(-2,3) + log(3)/4 - I*Pi/4*csgn(I-I*pow(x,2))) * pow(x-2,2)
@@ -243,8 +271,10 @@ static unsigned exam_series10(void)
 }
 
 // Series expansion of logarithms around branch points
-static unsigned exam_series11(void)
+static unsigned exam_series11()
 {
+	using GiNaC::log;
+
 	unsigned result = 0;
 	ex e, d;
 	symbol a("a");
@@ -276,15 +306,17 @@ static unsigned exam_series11(void)
 	result += check_series(e,0,d,5);
 	
 	e = log((1-x)/x);
-	d = log(1-x) - (x-1) + pow(x-1,2)/2 - pow(x-1,3)/3 + Order(pow(x-1,4));
-	result += check_series(e,1,d,4);
+	d = log(1-x) - (x-1) + pow(x-1,2)/2 - pow(x-1,3)/3  + pow(x-1,4)/4 + Order(pow(x-1,5));
+	result += check_series(e,1,d,5);
 	
 	return result;
 }
 
 // Series expansion of other functions around branch points
-static unsigned exam_series12(void)
+static unsigned exam_series12()
 {
+	using GiNaC::log;
+
 	unsigned result = 0;
 	ex e, d;
 	
@@ -309,7 +341,7 @@ static unsigned exam_series12(void)
 }
 
 
-unsigned exam_pseries(void)
+unsigned exam_pseries()
 {
 	unsigned result = 0;
 	
