@@ -3,7 +3,7 @@
  *  Interface to GiNaC's sums of expressions. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -41,32 +41,30 @@ public:
 	add(const exvector & v);
 	add(const epvector & v);
 	add(const epvector & v, const ex & oc);
-	add(epvector * vp, const ex & oc);
+	add(std::auto_ptr<epvector> vp, const ex & oc);
 	
 	// functions overriding virtual functions from base classes
 public:
-	void print(const print_context & c, unsigned level = 0) const;
-	unsigned precedence(void) const {return 40;}
+	unsigned precedence() const {return 40;}
 	bool info(unsigned inf) const;
 	int degree(const ex & s) const;
 	int ldegree(const ex & s) const;
 	ex coeff(const ex & s, int n=1) const;
 	ex eval(int level=0) const;
-	ex evalm(void) const;
+	ex evalm() const;
 	ex series(const relational & r, int order, unsigned options = 0) const;
-	ex normal(lst &sym_lst, lst &repl_lst, int level=0) const;
-	numeric integer_content(void) const;
+	ex normal(exmap & repl, exmap & rev_lookup, int level=0) const;
+	numeric integer_content() const;
 	ex smod(const numeric &xi) const;
-	numeric max_coefficient(void) const;
-	exvector get_free_indices(void) const;
-	ex simplify_ncmul(const exvector & v) const;
+	numeric max_coefficient() const;
+	exvector get_free_indices() const;
+	ex eval_ncmul(const exvector & v) const;
 protected:
 	ex derivative(const symbol & s) const;
-	bool is_equal_same_type(const basic & other) const;
-	unsigned return_type(void) const;
-	unsigned return_type_tinfo(void) const;
+	unsigned return_type() const;
+	unsigned return_type_tinfo() const;
 	ex thisexpairseq(const epvector & v, const ex & oc) const;
-	ex thisexpairseq(epvector * vp, const ex & oc) const;
+	ex thisexpairseq(std::auto_ptr<epvector> vp, const ex & oc) const;
 	expair split_ex_to_pair(const ex & e) const;
 	expair combine_ex_with_coeff_to_pair(const ex & e,
 	                                     const ex & c) const;
@@ -74,6 +72,14 @@ protected:
 	                                       const ex & c) const;
 	ex recombine_pair_to_ex(const expair & p) const;
 	ex expand(unsigned options=0) const;
+
+	// non-virtual functions in this class
+protected:
+	void print_add(const print_context & c, const char *openbrace, const char *closebrace, const char *mul_sym, unsigned level) const;
+	void do_print(const print_context & c, unsigned level) const;
+	void do_print_latex(const print_latex & c, unsigned level) const;
+	void do_print_csrc(const print_csrc & c, unsigned level) const;
+	void do_print_python_repr(const print_python_repr & c, unsigned level) const;
 };
 
 // utility functions
