@@ -3,7 +3,7 @@
  *  Implementation of GiNaC's ABC. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -34,51 +34,66 @@
 #include "lst.h"
 #include "ncmul.h"
 #include "relational.h"
+#include "operators.h"
 #include "wildcard.h"
-#include "print.h"
 #include "archive.h"
 #include "utils.h"
 
 namespace GiNaC {
 
-GINAC_IMPLEMENT_REGISTERED_CLASS_NO_CTORS(basic, void)
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(basic, void,
+  print_func<print_context>(&basic::do_print).
+  print_func<print_tree>(&basic::do_print_tree).
+  print_func<print_python_repr>(&basic::do_print_python_repr))
 
 //////////
-// default ctor, dtor, copy ctor, assignment operator and helpers
+// default constructor, destructor, copy constructor and assignment operator
 //////////
 
 // public
 
-basic::basic(const basic & other) : tinfo_key(TINFO_basic), flags(0), refcount(0)
+/** basic copy constructor: implicitly assumes that the other class is of
+ *  the exact same type (as it's used by duplicate()), so it can copy the
+ *  tinfo_key and the hash value. */
+basic::basic(const basic & other) : tinfo_key(other.tinfo_key), flags(other.flags & ~status_flags::dynallocated), hashvalue(other.hashvalue)
 {
-	copy(other);
+	GINAC_ASSERT(typeid(*this) == typeid(other));
 }
 
+/** basic assignment operator: the other object might be of a derived class. */
 const basic & basic::operator=(const basic & other)
 {
-	if (this != &other) {
-		destroy(true);
-		copy(other);
+	unsigned fl = other.flags & ~status_flags::dynallocated;
+	if (tinfo_key != other.tinfo_key) {
+		// The other object is of a derived class, so clear the flags as they
+		// might no longer apply (especially hash_calculated). Oh, and don't
+		// copy the tinfo_key: it is already set correctly for this object.
+		fl &= ~(status_flags::evaluated | status_flags::expanded | status_flags::hash_calculated);
+	} else {
+		// The objects are of the exact same class, so copy the hash value.
+		hashvalue = other.hashvalue;
 	}
+	flags = fl;
+	set_refcount(0);
 	return *this;
 }
 
 // protected
 
-// none (all conditionally inlined)
+// none (all inlined)
 
 //////////
-// other ctors
+// other constructors
 //////////
 
-// none (all conditionally inlined)
+// none (all inlined)
 
 //////////
 // archiving
 //////////
 
 /** Construct object from archive_node. */
-basic::basic(const archive_node &n, const lst &sym_lst) : flags(0), refcount(0)
+basic::basic(const archive_node &n, lst &sym_lst) : flags(0)
 {
 	// Reconstruct tinfo_key from class name
 	std::string class_name;
@@ -103,23 +118,85 @@ void basic::archive(archive_node &n) const
 
 // public
 
-/** Output to stream.
+/** Output to stream. This performs double dispatch on the dynamic type of
+ *  *this and the dynamic type of the supplied print context.
  *  @param c print context object that describes the output formatting
  *  @param level value that is used to identify the precedence or indentation
  *               level for placing parentheses and formatting */
 void basic::print(const print_context & c, unsigned level) const
 {
-	if (is_of_type(c, print_tree)) {
+	print_dispatch(get_class_info(), c, level);
+}
 
-		c.s << std::string(level, ' ') << class_name()
-		    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
-		    << ", nops=" << nops()
-		    << std::endl;
-		for (unsigned i=0; i<nops(); ++i)
-			op(i).print(c, level + static_cast<const print_tree &>(c).delta_indent);
+/** Like print(), but dispatch to the specified class. Can be used by
+ *  implementations of print methods to dispatch to the method of the
+ *  superclass.
+ *
+ *  @see basic::print */
+void basic::print_dispatch(const registered_class_info & ri, const print_context & c, unsigned level) const
+{
+	// Double dispatch on object type and print_context type
+	const registered_class_info * reg_info = &ri;
+	const print_context_class_info * pc_info = &c.get_class_info();
 
-	} else
-		c.s << "[" << class_name() << " object]";
+next_class:
+	const std::vector<print_functor> & pdt = reg_info->options.get_print_dispatch_table();
+
+next_context:
+	unsigned id = pc_info->options.get_id();
+	if (id >= pdt.size() || !(pdt[id].is_valid())) {
+
+		// Method not found, try parent print_context class
+		const print_context_class_info * parent_pc_info = pc_info->get_parent();
+		if (parent_pc_info) {
+			pc_info = parent_pc_info;
+			goto next_context;
+		}
+
+		// Method still not found, try parent class
+		const registered_class_info * parent_reg_info = reg_info->get_parent();
+		if (parent_reg_info) {
+			reg_info = parent_reg_info;
+			pc_info = &c.get_class_info();
+			goto next_class;
+		}
+
+		// Method still not found. This shouldn't happen because basic (the
+		// base class of the algebraic hierarchy) registers a method for
+		// print_context (the base class of the print context hierarchy),
+		// so if we end up here, there's something wrong with the class
+		// registry.
+		throw (std::runtime_error(std::string("basic::print(): method for ") + class_name() + "/" + c.class_name() + " not found"));
+
+	} else {
+
+		// Call method
+		pdt[id](*this, c, level);
+	}
+}
+
+/** Default output to stream. */
+void basic::do_print(const print_context & c, unsigned level) const
+{
+	c.s << "[" << class_name() << " object]";
+}
+
+/** Tree output to stream. */
+void basic::do_print_tree(const print_tree & c, unsigned level) const
+{
+	c.s << std::string(level, ' ') << class_name() << " @" << this
+	    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec;
+	if (nops())
+		c.s << ", nops=" << nops();
+	c.s << std::endl;
+	for (size_t i=0; i<nops(); ++i)
+		op(i).print(c, level + c.delta_indent);
+}
+
+/** Python parsable output to stream. */
+void basic::do_print_python_repr(const print_python_repr & c, unsigned level) const
+{
+	c.s << class_name() << "()";
 }
 
 /** Little wrapper around print to be called within a debugger.
@@ -127,8 +204,9 @@ void basic::print(const print_context & c, unsigned level) const
  *  debugger because it might not know what cout is.  This method can be
  *  invoked with no argument and it will simply print to stdout.
  *
- *  @see basic::print */
-void basic::dbgprint(void) const
+ *  @see basic::print
+ *  @see basic::dbgprinttree */
+void basic::dbgprint() const
 {
 	this->print(std::cerr);
 	std::cerr << std::endl;
@@ -136,27 +214,18 @@ void basic::dbgprint(void) const
 
 /** Little wrapper around printtree to be called within a debugger.
  *
- *  @see basic::dbgprint
- *  @see basic::printtree */
-void basic::dbgprinttree(void) const
+ *  @see basic::dbgprint */
+void basic::dbgprinttree() const
 {
 	this->print(print_tree(std::cerr));
 }
 
-/** Return relative operator precedence (for parenthizing output). */
-unsigned basic::precedence(void) const
+/** Return relative operator precedence (for parenthezing output). */
+unsigned basic::precedence() const
 {
 	return 70;
 }
 
-/** Create a new copy of this on the heap.  One can think of this as simulating
- *  a virtual copy constructor which is needed for instance by the refcounted
- *  construction of an ex from a basic. */
-basic * basic::duplicate() const
-{
-	return new basic(*this);
-}
-
 /** Information about the object.
  *
  *  @see class info_flags */
@@ -167,7 +236,7 @@ bool basic::info(unsigned inf) const
 }
 
 /** Number of operands/members. */
-unsigned basic::nops() const
+size_t basic::nops() const
 {
 	// iterating from 0 to nops() on atomic objects should be an empty loop,
 	// and accessing their elements is a range error.  Container objects should
@@ -176,30 +245,44 @@ unsigned basic::nops() const
 }
 
 /** Return operand/member at position i. */
-ex basic::op(int i) const
+ex basic::op(size_t i) const
 {
-	return (const_cast<basic *>(this))->let_op(i);
+	throw(std::range_error(std::string("basic::op(): ") + class_name() + std::string(" has no operands")));
 }
 
 /** Return modifyable operand/member at position i. */
-ex & basic::let_op(int i)
+ex & basic::let_op(size_t i)
 {
-	throw(std::out_of_range("op() out of range"));
+	ensure_if_modifiable();
+	throw(std::range_error(std::string("basic::let_op(): ") + class_name() + std::string(" has no operands")));
 }
 
 ex basic::operator[](const ex & index) const
 {
-	if (is_ex_exactly_of_type(index,numeric))
-		return op(ex_to<numeric>(index).to_int());
+	if (is_exactly_a<numeric>(index))
+		return op(static_cast<size_t>(ex_to<numeric>(index).to_int()));
 
-	throw(std::invalid_argument("non-numeric indices not supported by this type"));
+	throw(std::invalid_argument(std::string("non-numeric indices not supported by ") + class_name()));
 }
 
-ex basic::operator[](int i) const
+ex basic::operator[](size_t i) const
 {
 	return op(i);
 }
 
+ex & basic::operator[](const ex & index)
+{
+	if (is_exactly_a<numeric>(index))
+		return let_op(ex_to<numeric>(index).to_int());
+
+	throw(std::invalid_argument(std::string("non-numeric indices not supported by ") + class_name()));
+}
+
+ex & basic::operator[](size_t i)
+{
+	return let_op(i);
+}
+
 /** Test for occurrence of a pattern.  An object 'has' a pattern if it matches
  *  the pattern itself or one of the children 'has' it.  As a consequence
  *  (according to the definition of children) given e=x+y+z, e.has(x) is true
@@ -209,7 +292,7 @@ bool basic::has(const ex & pattern) const
 	lst repl_lst;
 	if (match(pattern, repl_lst))
 		return true;
-	for (unsigned i=0; i<nops(); i++)
+	for (size_t i=0; i<nops(); i++)
 		if (op(i).has(pattern))
 			return true;
 	
@@ -220,17 +303,16 @@ bool basic::has(const ex & pattern) const
  *  sub-expressions (one level only, not recursively). */
 ex basic::map(map_function & f) const
 {
-	unsigned num = nops();
+	size_t num = nops();
 	if (num == 0)
 		return *this;
 
 	basic *copy = duplicate();
 	copy->setflag(status_flags::dynallocated);
 	copy->clearflag(status_flags::hash_calculated | status_flags::expanded);
-	ex e(*copy);
-	for (unsigned i=0; i<num; i++)
-		e.let_op(i) = f(e.op(i));
-	return e.eval();
+	for (size_t i=0; i<num; i++)
+		copy->let_op(i) = f(copy->op(i));
+	return *copy;
 }
 
 /** Return degree of highest power in object s. */
@@ -260,7 +342,7 @@ ex basic::coeff(const ex & s, int n) const
 ex basic::collect(const ex & s, bool distributed) const
 {
 	ex x;
-	if (is_ex_of_type(s, lst)) {
+	if (is_a<lst>(s)) {
 
 		// List of objects specified
 		if (s.nops() == 0)
@@ -271,7 +353,7 @@ ex basic::collect(const ex & s, bool distributed) const
 		else if (distributed) {
 
 			// Get lower/upper degree of all symbols in list
-			int num = s.nops();
+			size_t num = s.nops();
 			struct sym_info {
 				ex sym;
 				int ldeg, deg;
@@ -280,7 +362,7 @@ ex basic::collect(const ex & s, bool distributed) const
 			};
 			sym_info *si = new sym_info[num];
 			ex c = *this;
-			for (int i=0; i<num; i++) {
+			for (size_t i=0; i<num; i++) {
 				si[i].sym = s.op(i);
 				si[i].ldeg = si[i].cnt = this->ldegree(si[i].sym);
 				si[i].deg = this->degree(si[i].sym);
@@ -291,14 +373,14 @@ ex basic::collect(const ex & s, bool distributed) const
 
 				// Calculate coeff*x1^c1*...*xn^cn
 				ex y = _ex1;
-				for (int i=0; i<num; i++) {
+				for (size_t i=0; i<num; i++) {
 					int cnt = si[i].cnt;
 					y *= power(si[i].sym, cnt);
 				}
 				x += y * si[num - 1].coeff;
 
 				// Increment counters
-				int n = num - 1;
+				size_t n = num - 1;
 				while (true) {
 					++si[n].cnt;
 					if (si[n].cnt <= si[n].deg) {
@@ -308,7 +390,7 @@ ex basic::collect(const ex & s, bool distributed) const
 							c = *this;
 						else
 							c = si[n - 1].coeff;
-						for (int i=n; i<num; i++)
+						for (size_t i=n; i<num; i++)
 							c = si[i].coeff = c.coeff(si[i].sym, si[i].cnt);
 						break;
 					}
@@ -325,8 +407,13 @@ done:		delete[] si;
 
 			// Recursive form
 			x = *this;
-			for (int n=s.nops()-1; n>=0; n--)
+			size_t n = s.nops() - 1;
+			while (true) {
 				x = x.collect(s[n]);
+				if (n == 0)
+					break;
+				n--;
+			}
 		}
 
 	} else {
@@ -344,7 +431,7 @@ done:		delete[] si;
 ex basic::eval(int level) const
 {
 	// There is nothing to do for basic objects:
-	return this->hold();
+	return hold();
 }
 
 /** Function object to be applied by basic::evalf(). */
@@ -377,7 +464,7 @@ struct evalm_map_function : public map_function {
 } map_evalm;
 
 /** Evaluate sums, products and integer powers of matrices. */
-ex basic::evalm(void) const
+ex basic::evalm() const
 {
 	if (nops() == 0)
 		return *this;
@@ -459,14 +546,14 @@ bool basic::match(const ex & pattern, lst & repl_lst) const
 	Bog is the king of Pattern.
 */
 
-	if (is_ex_exactly_of_type(pattern, wildcard)) {
+	if (is_exactly_a<wildcard>(pattern)) {
 
 		// Wildcard matches anything, but check whether we already have found
 		// a match for that wildcard first (if so, it the earlier match must
 		// be the same expression)
-		for (unsigned i=0; i<repl_lst.nops(); i++) {
-			if (repl_lst.op(i).op(0).is_equal(pattern))
-				return is_equal(ex_to<basic>(repl_lst.op(i).op(1)));
+		for (lst::const_iterator it = repl_lst.begin(); it != repl_lst.end(); ++it) {
+			if (it->op(0).is_equal(pattern))
+				return is_equal(ex_to<basic>(it->op(1)));
 		}
 		repl_lst.append(pattern == *this);
 		return true;
@@ -491,7 +578,7 @@ bool basic::match(const ex & pattern, lst & repl_lst) const
 			return false;
 
 		// Otherwise the subexpressions must match one-to-one
-		for (unsigned i=0; i<nops(); i++)
+		for (size_t i=0; i<nops(); i++)
 			if (!op(i).match(pattern.op(i), repl_lst))
 				return false;
 
@@ -500,28 +587,61 @@ bool basic::match(const ex & pattern, lst & repl_lst) const
 	}
 }
 
-/** Substitute a set of objects by arbitrary expressions. The ex returned
- *  will already be evaluated. */
-ex basic::subs(const lst & ls, const lst & lr, bool no_pattern) const
+/** Helper function for subs(). Does not recurse into subexpressions. */
+ex basic::subs_one_level(const exmap & m, unsigned options) const
 {
-	GINAC_ASSERT(ls.nops() == lr.nops());
+	exmap::const_iterator it;
 
-	if (no_pattern) {
-		for (unsigned i=0; i<ls.nops(); i++) {
-			if (is_equal(ex_to<basic>(ls.op(i))))
-				return lr.op(i);
-		}
+	if (options & subs_options::no_pattern) {
+		it = m.find(*this);
+		if (it != m.end())
+			return it->second;
 	} else {
-		for (unsigned i=0; i<ls.nops(); i++) {
+		for (it = m.begin(); it != m.end(); ++it) {
 			lst repl_lst;
-			if (match(ex_to<basic>(ls.op(i)), repl_lst))
-				return lr.op(i).subs(repl_lst, true); // avoid infinite recursion when re-substituting the wildcards
+			if (match(ex_to<basic>(it->first), repl_lst))
+				return it->second.subs(repl_lst, options | subs_options::no_pattern); // avoid infinite recursion when re-substituting the wildcards
 		}
 	}
 
 	return *this;
 }
 
+/** Substitute a set of objects by arbitrary expressions. The ex returned
+ *  will already be evaluated. */
+ex basic::subs(const exmap & m, unsigned options) const
+{
+	size_t num = nops();
+	if (num) {
+
+		// Substitute in subexpressions
+		for (size_t i=0; i<num; i++) {
+			const ex & orig_op = op(i);
+			const ex & subsed_op = orig_op.subs(m, options);
+			if (!are_ex_trivially_equal(orig_op, subsed_op)) {
+
+				// Something changed, clone the object
+				basic *copy = duplicate();
+				copy->setflag(status_flags::dynallocated);
+				copy->clearflag(status_flags::hash_calculated | status_flags::expanded);
+
+				// Substitute the changed operand
+				copy->let_op(i++) = subsed_op;
+
+				// Substitute the other operands
+				for (; i<num; i++)
+					copy->let_op(i) = op(i).subs(m, options);
+
+				// Perform substitutions on the new object as a whole
+				return copy->subs_one_level(m, options);
+			}
+		}
+	}
+
+	// Nothing changed or no subexpressions
+	return subs_one_level(m, options);
+}
+
 /** Default interface of nth derivative ex::diff(s, n).  It should be called
  *  instead of ::derivative(s) for first derivatives and for nth derivatives it
  *  just recurses down.
@@ -549,14 +669,19 @@ ex basic::diff(const symbol & s, unsigned nth) const
 }
 
 /** Return a vector containing the free indices of an expression. */
-exvector basic::get_free_indices(void) const
+exvector basic::get_free_indices() const
 {
 	return exvector(); // return an empty exvector
 }
 
-ex basic::simplify_ncmul(const exvector & v) const
+ex basic::conjugate() const
 {
-	return simplified_ncmul(v);
+	return *this;
+}
+
+ex basic::eval_ncmul(const exvector & v) const
+{
+	return hold_ncmul(v);
 }
 
 // protected
@@ -619,12 +744,12 @@ bool basic::match_same_type(const basic & other) const
 	return true;
 }
 
-unsigned basic::return_type(void) const
+unsigned basic::return_type() const
 {
 	return return_types::commutative;
 }
 
-unsigned basic::return_type_tinfo(void) const
+unsigned basic::return_type_tinfo() const
 {
 	return tinfo();
 }
@@ -635,17 +760,14 @@ unsigned basic::return_type_tinfo(void) const
  *  members.  For this reason it is well suited for container classes but
  *  atomic classes should override this implementation because otherwise they
  *  would all end up with the same hashvalue. */
-unsigned basic::calchash(void) const
+unsigned basic::calchash() const
 {
 	unsigned v = golden_ratio_hash(tinfo());
-	for (unsigned i=0; i<nops(); i++) {
-		v = rotate_left_31(v);
-		v ^= (const_cast<basic *>(this))->op(i).gethash();
+	for (size_t i=0; i<nops(); i++) {
+		v = rotate_left(v);
+		v ^= this->op(i).gethash();
 	}
-	
-	// mask out numeric hashes:
-	v &= 0x7FFFFFFFU;
-	
+
 	// store calculated hash value only if object is already evaluated
 	if (flags & status_flags::evaluated) {
 		setflag(status_flags::hash_calculated);
@@ -681,82 +803,43 @@ ex basic::expand(unsigned options) const
 
 // public
 
-/** Substitute objects in an expression (syntactic substitution) and return
- *  the result as a new expression.  There are two valid types of
- *  replacement arguments: 1) a relational like object==ex and 2) a list of
- *  relationals lst(object1==ex1,object2==ex2,...), which is converted to
- *  subs(lst(object1,object2,...),lst(ex1,ex2,...)). */
-ex basic::subs(const ex & e, bool no_pattern) const
-{
-	if (e.info(info_flags::relation_equal)) {
-		return subs(lst(e), no_pattern);
-	}
-	if (!e.info(info_flags::list)) {
-		throw(std::invalid_argument("basic::subs(ex): argument must be a list"));
-	}
-	lst ls;
-	lst lr;
-	for (unsigned i=0; i<e.nops(); i++) {
-		ex r = e.op(i);
-		if (!r.info(info_flags::relation_equal)) {
-			throw(std::invalid_argument("basic::subs(ex): argument must be a list of equations"));
-		}
-		ls.append(r.op(0));
-		lr.append(r.op(1));
-	}
-	return subs(ls, lr, no_pattern);
-}
-
-/** Compare objects to establish canonical ordering.
+/** Compare objects syntactically to establish canonical ordering.
  *  All compare functions return: -1 for *this less than other, 0 equal,
  *  1 greater. */
 int basic::compare(const basic & other) const
 {
-	unsigned hash_this = gethash();
-	unsigned hash_other = other.gethash();
-	
+	const unsigned hash_this = gethash();
+	const unsigned hash_other = other.gethash();
 	if (hash_this<hash_other) return -1;
 	if (hash_this>hash_other) return 1;
-	
-	unsigned typeid_this = tinfo();
-	unsigned typeid_other = other.tinfo();
-	
-	if (typeid_this<typeid_other) {
+
+	const unsigned typeid_this = tinfo();
+	const unsigned typeid_other = other.tinfo();
+	if (typeid_this==typeid_other) {
+		GINAC_ASSERT(typeid(*this)==typeid(other));
+// 		int cmpval = compare_same_type(other);
+// 		if (cmpval!=0) {
+// 			std::cout << "hash collision, same type: " 
+// 			          << *this << " and " << other << std::endl;
+// 			this->print(print_tree(std::cout));
+// 			std::cout << " and ";
+// 			other.print(print_tree(std::cout));
+// 			std::cout << std::endl;
+// 		}
+// 		return cmpval;
+		return compare_same_type(other);
+	} else {
 // 		std::cout << "hash collision, different types: " 
 // 		          << *this << " and " << other << std::endl;
 // 		this->print(print_tree(std::cout));
 // 		std::cout << " and ";
 // 		other.print(print_tree(std::cout));
 // 		std::cout << std::endl;
-		return -1;
+		return (typeid_this<typeid_other ? -1 : 1);
 	}
-	if (typeid_this>typeid_other) {
-// 		std::cout << "hash collision, different types: " 
-// 		          << *this << " and " << other << std::endl;
-// 		this->print(print_tree(std::cout));
-// 		std::cout << " and ";
-// 		other.print(print_tree(std::cout));
-// 		std::cout << std::endl;
-		return 1;
-	}
-	
-	GINAC_ASSERT(typeid(*this)==typeid(other));
-	
-// 	int cmpval = compare_same_type(other);
-// 	if ((cmpval!=0) && (hash_this<0x80000000U)) {
-// 		std::cout << "hash collision, same type: " 
-// 		          << *this << " and " << other << std::endl;
-// 		this->print(print_tree(std::cout));
-// 		std::cout << " and ";
-// 		other.print(print_tree(std::cout));
-// 		std::cout << std::endl;
-// 	}
-// 	return cmpval;
-	
-	return compare_same_type(other);
 }
 
-/** Test for equality.
+/** Test for syntactic equality.
  *  This is only a quick test, meaning objects should be in the same domain.
  *  You might have to .expand(), .normal() objects first, depending on the
  *  domain of your computation, to get a more reliable answer.
@@ -779,18 +862,18 @@ bool basic::is_equal(const basic & other) const
 /** Stop further evaluation.
  *
  *  @see basic::eval */
-const basic & basic::hold(void) const
+const basic & basic::hold() const
 {
 	return setflag(status_flags::evaluated);
 }
 
 /** Ensure the object may be modified without hurting others, throws if this
  *  is not the case. */
-void basic::ensure_if_modifiable(void) const
+void basic::ensure_if_modifiable() const
 {
-	if (this->refcount>1)
+	if (get_refcount() > 1)
 		throw(std::runtime_error("cannot modify multiply referenced object"));
-	clearflag(status_flags::hash_calculated);
+	clearflag(status_flags::hash_calculated | status_flags::evaluated);
 }
 
 //////////
