@@ -3,7 +3,7 @@
  *  Interface to GiNaC's ABC. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,26 +23,30 @@
 #ifndef __GINAC_BASIC_H__
 #define __GINAC_BASIC_H__
 
+#include <cstddef> // for size_t
 #include <vector>
+#include <map>
 // CINT needs <algorithm> to work properly with <vector>
 #include <algorithm>
 
 #include "flags.h"
 #include "tinfos.h"
+#include "ptr.h"
 #include "assertion.h"
 #include "registrar.h"
 
 namespace GiNaC {
 
 class ex;
+class ex_is_less;
 class symbol;
-class lst;
 class numeric;
 class relational;
 class archive_node;
 class print_context;
 
 typedef std::vector<ex> exvector;
+typedef std::map<ex, ex, ex_is_less> exmap;
 
 
 /** Function object for map(). */
@@ -53,95 +57,163 @@ struct map_function {
 };
 
 
-/** This class is the ABC (abstract base class) of GiNaC's class hierarchy.
- *  It is responsible for the reference counting. */
-class basic
+/** Degenerate base class for visitors. basic and derivative classes
+ *  support Robert C. Martin's Acyclic Visitor pattern (cf.
+ *  http://objectmentor.com/publications/acv.pdf). */
+class visitor {
+protected:
+	virtual ~visitor() {}
+};
+
+
+/** This class is the ABC (abstract base class) of GiNaC's class hierarchy. */
+class basic : public refcounted
 {
 	GINAC_DECLARE_REGISTERED_CLASS_NO_CTORS(basic, void)
 	
 	friend class ex;
 	
-	// default ctor, dtor, copy ctor, assignment operator and helpers
+	// default constructor, destructor, copy constructor and assignment operator
+protected:
+	basic() : tinfo_key(TINFO_basic), flags(0) {}
+
 public:
-	basic() : tinfo_key(TINFO_basic), flags(0), refcount(0) {}
-	/** basic dtor, virtual because class ex will delete objects via ptr. */
+	/** basic destructor, virtual because class ex will delete objects of
+	 *  derived classes via a basic*. */
 	virtual ~basic()
 	{
-		destroy(false);
-		GINAC_ASSERT((!(flags & status_flags::dynallocated))||(refcount==0));
+		GINAC_ASSERT((!(flags & status_flags::dynallocated)) || (get_refcount() == 0));
 	}
 	basic(const basic & other);
 	const basic & operator=(const basic & other);
+
 protected:
-	/** For use by copy ctor and assignment operator. */
-	void copy(const basic & other)
-	{
-		flags = other.flags & ~status_flags::dynallocated;
-		hashvalue = other.hashvalue;
-		tinfo_key = other.tinfo_key;
-	}
-	/** For use by dtor and assignment operator. */
-	virtual void destroy(bool call_parent) { }
-	
-	// other ctors
-	/** ctor with specified tinfo_key */
-	basic(unsigned ti) : tinfo_key(ti), flags(0), refcount(0) {}
+	/** Constructor with specified tinfo_key (used by derived classes instead
+	 *  of the default constructor to avoid assigning tinfo_key twice). */
+	basic(unsigned ti) : tinfo_key(ti), flags(0) {}
 	
 	// new virtual functions which can be overridden by derived classes
 public: // only const functions please (may break reference counting)
-	virtual basic * duplicate() const;
+
+	/** Create a clone of this object on the heap.  One can think of this as
+	 *  simulating a virtual copy constructor which is needed for instance by
+	 *  the refcounted construction of an ex from a basic. */
+	virtual basic * duplicate() const { return new basic(*this); }
+
+	// evaluation
+	virtual ex eval(int level = 0) const;
+	virtual ex evalf(int level = 0) const;
+	virtual ex evalm() const;
+protected:
+	virtual ex eval_ncmul(const exvector & v) const;
+public:
+	virtual ex eval_indexed(const basic & i) const;
+
+	// printing
 	virtual void print(const print_context & c, unsigned level = 0) const;
-	virtual void dbgprint(void) const;
-	virtual void dbgprinttree(void) const;
-	virtual unsigned precedence(void) const;
+	virtual void dbgprint() const;
+	virtual void dbgprinttree() const;
+	virtual unsigned precedence() const;
+
+	// info
 	virtual bool info(unsigned inf) const;
-	virtual unsigned nops() const;
-	virtual ex op(int i) const;
-	virtual ex & let_op(int i);
+
+	// operand access
+	virtual size_t nops() const;
+	virtual ex op(size_t i) const;
 	virtual ex operator[](const ex & index) const;
-	virtual ex operator[](int i) const;
-	virtual ex expand(unsigned options = 0) const;
+	virtual ex operator[](size_t i) const;
+	virtual ex & let_op(size_t i);
+	virtual ex & operator[](const ex & index);
+	virtual ex & operator[](size_t i);
+
+	// pattern matching
 	virtual bool has(const ex & other) const;
+	virtual bool match(const ex & pattern, lst & repl_lst) const;
+protected:
+	virtual bool match_same_type(const basic & other) const;
+public:
+
+	// substitutions
+	virtual ex subs(const exmap & m, unsigned options = 0) const;
+
+	// function mapping
 	virtual ex map(map_function & f) const;
+
+	// visitors and tree traversal
+	virtual void accept(GiNaC::visitor & v) const
+	{
+		if (visitor *p = dynamic_cast<visitor *>(&v))
+			p->visit(*this);
+	}
+
+	// degree/coeff
 	virtual int degree(const ex & s) const;
 	virtual int ldegree(const ex & s) const;
 	virtual ex coeff(const ex & s, int n = 1) const;
+
+	// expand/collect
+	virtual ex expand(unsigned options = 0) const;
 	virtual ex collect(const ex & s, bool distributed = false) const;
-	virtual ex eval(int level = 0) const;
-	virtual ex evalf(int level = 0) const;
-	virtual ex evalm(void) const;
+
+	// differentiation and series expansion
+protected:
+	virtual ex derivative(const symbol & s) const;
+public:
 	virtual ex series(const relational & r, int order, unsigned options = 0) const;
-	virtual bool match(const ex & pattern, lst & repl_lst) const;
-	virtual ex subs(const lst & ls, const lst & lr, bool no_pattern = false) const;
-	virtual ex normal(lst &sym_lst, lst &repl_lst, int level = 0) const;
-	virtual ex to_rational(lst &repl_lst) const;
-	virtual numeric integer_content(void) const;
+
+	// rational functions
+	virtual ex normal(exmap & repl, exmap & rev_lookup, int level = 0) const;
+	virtual ex to_rational(exmap & repl) const;
+	virtual ex to_polynomial(exmap & repl) const;
+
+	// polynomial algorithms
+	virtual numeric integer_content() const;
 	virtual ex smod(const numeric &xi) const;
-	virtual numeric max_coefficient(void) const;
-	virtual exvector get_free_indices(void) const;
-	virtual ex eval_indexed(const basic & i) const;
+	virtual numeric max_coefficient() const;
+
+	// indexed objects
+	virtual exvector get_free_indices() const;
 	virtual ex add_indexed(const ex & self, const ex & other) const;
 	virtual ex scalar_mul_indexed(const ex & self, const numeric & other) const;
 	virtual bool contract_with(exvector::iterator self, exvector::iterator other, exvector & v) const;
-	virtual unsigned return_type(void) const;
-	virtual unsigned return_type_tinfo(void) const;
-protected: // functions that should be called from class ex only
-	virtual ex derivative(const symbol & s) const;
+
+	// noncommutativity
+	virtual unsigned return_type() const;
+	virtual unsigned return_type_tinfo() const;
+
+	// complex conjugation
+	virtual ex conjugate() const;
+
+	// functions that should be called from class ex only
+protected:
 	virtual int compare_same_type(const basic & other) const;
 	virtual bool is_equal_same_type(const basic & other) const;
-	virtual bool match_same_type(const basic & other) const;
-	virtual unsigned calchash(void) const;
-	virtual ex simplify_ncmul(const exvector & v) const;
+
+	virtual unsigned calchash() const;
 	
 	// non-virtual functions in this class
 public:
-	ex subs(const ex & e, bool no_pattern = false) const;
-	ex diff(const symbol & s, unsigned nth=1) const;
+	/** Like print(), but dispatch to the specified class. Can be used by
+	 *  implementations of print methods to dispatch to the method of the
+	 *  superclass.
+	 *
+	 *  @see basic::print */
+	template <class T>
+	void print_dispatch(const print_context & c, unsigned level) const
+	{
+		print_dispatch(T::get_class_info_static(), c, level);
+	}
+
+	void print_dispatch(const registered_class_info & ri, const print_context & c, unsigned level) const;
+
+	ex subs_one_level(const exmap & m, unsigned options) const;
+	ex diff(const symbol & s, unsigned nth = 1) const;
 	int compare(const basic & other) const;
 	bool is_equal(const basic & other) const;
-	const basic & hold(void) const;
-	unsigned gethash(void) const { if (flags & status_flags::hash_calculated) return hashvalue; else return calchash(); }
-	unsigned tinfo(void) const {return tinfo_key;}
+	const basic & hold() const;
+	unsigned gethash() const { if (flags & status_flags::hash_calculated) return hashvalue; else return calchash(); }
+	unsigned tinfo() const {return tinfo_key;}
 
 	/** Set some status_flags. */
 	const basic & setflag(unsigned f) const {flags |= f; return *this;}
@@ -150,66 +222,41 @@ public:
 	const basic & clearflag(unsigned f) const {flags &= ~f; return *this;}
 
 protected:
-	void ensure_if_modifiable(void) const;
+	void ensure_if_modifiable() const;
+
+	void do_print(const print_context & c, unsigned level) const;
+	void do_print_tree(const print_tree & c, unsigned level) const;
+	void do_print_python_repr(const print_python_repr & c, unsigned level) const;
 	
 	// member variables
 protected:
 	unsigned tinfo_key;                 ///< typeinfo
 	mutable unsigned flags;             ///< of type status_flags
 	mutable unsigned hashvalue;         ///< hash value
-private:
-	unsigned refcount;                  ///< Number of reference counts
 };
 
+
 // global variables
 
 extern int max_recursion_level;
 
+
 // convenience type checker template functions
 
 /** Check if obj is a T, including base classes. */
 template <class T>
 inline bool is_a(const basic &obj)
 {
-	return dynamic_cast<const T *>(&obj)!=0;
+	return dynamic_cast<const T *>(&obj) != 0;
 }
 
 /** Check if obj is a T, not including base classes.  This one is just an
  *  inefficient default.  It should in all time-critical cases be overridden
- *  by template specializations that don't create a temporary. */
+ *  by template specializations that use the TINFO_* constants directly. */
 template <class T>
-inline bool is_exactly_a(const class basic &obj)
+inline bool is_exactly_a(const basic &obj)
 {
-	const T foo; return foo.tinfo()==obj.tinfo();
-}
-
-/** Check if ex is a handle to a T, including base classes. */
-template <class T>
-inline bool is_a(const ex &obj)
-{
-	return is_a<T>(*obj.bp);
-}
-
-/** Check if ex is a handle to a T, not including base classes. */
-template <class T>
-inline bool is_exactly_a(const ex &obj)
-{
-	return is_exactly_a<T>(*obj.bp);
-}
-
-/** Return a reference to the basic-derived class T object embedded in an
- *  expression.  This is fast but unsafe: the result is undefined if the
- *  expression does not contain a T object at its top level.  Hence, you
- *  should generally check the type of e first.
- *
- *  @param e expression
- *  @return reference to pseries object
- *  @see is_exactly_a<class T>() */
-template <class T>
-inline const T &ex_to(const ex &e)
-{
-	GINAC_ASSERT(is_a<T>(e));
-	return static_cast<const T &>(*e.bp);
+	return obj.tinfo() == T::get_class_info_static().options.get_id();
 }
 
 } // namespace GiNaC
