@@ -3,7 +3,7 @@
  *  Implementation of GiNaC's clifford algebra (Dirac gamma) objects. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -32,20 +32,39 @@
 #include "symmetry.h"
 #include "lst.h"
 #include "relational.h"
+#include "operators.h"
 #include "mul.h"
-#include "print.h"
 #include "archive.h"
 #include "utils.h"
 
 namespace GiNaC {
 
-GINAC_IMPLEMENT_REGISTERED_CLASS(clifford, indexed)
-GINAC_IMPLEMENT_REGISTERED_CLASS(diracone, tensor)
-GINAC_IMPLEMENT_REGISTERED_CLASS(diracgamma, tensor)
-GINAC_IMPLEMENT_REGISTERED_CLASS(diracgamma5, tensor)
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(clifford, indexed,
+  print_func<print_dflt>(&clifford::do_print_dflt).
+  print_func<print_latex>(&clifford::do_print_latex))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(diracone, tensor,
+  print_func<print_dflt>(&diracone::do_print).
+  print_func<print_latex>(&diracone::do_print_latex))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(diracgamma, tensor,
+  print_func<print_dflt>(&diracgamma::do_print).
+  print_func<print_latex>(&diracgamma::do_print_latex))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(diracgamma5, tensor,
+  print_func<print_dflt>(&diracgamma5::do_print).
+  print_func<print_latex>(&diracgamma5::do_print_latex))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(diracgammaL, tensor,
+  print_func<print_context>(&diracgammaL::do_print).
+  print_func<print_latex>(&diracgammaL::do_print_latex))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(diracgammaR, tensor,
+  print_func<print_context>(&diracgammaR::do_print).
+  print_func<print_latex>(&diracgammaR::do_print_latex))
 
 //////////
-// default ctor, dtor, copy ctor, assignment operator and helpers
+// default constructors
 //////////
 
 clifford::clifford() : representation_label(0)
@@ -53,16 +72,11 @@ clifford::clifford() : representation_label(0)
 	tinfo_key = TINFO_clifford;
 }
 
-void clifford::copy(const clifford & other)
-{
-	inherited::copy(other);
-	representation_label = other.representation_label;
-}
-
-DEFAULT_DESTROY(clifford)
-DEFAULT_CTORS(diracone)
-DEFAULT_CTORS(diracgamma)
-DEFAULT_CTORS(diracgamma5)
+DEFAULT_CTOR(diracone)
+DEFAULT_CTOR(diracgamma)
+DEFAULT_CTOR(diracgamma5)
+DEFAULT_CTOR(diracgammaL)
+DEFAULT_CTOR(diracgammaR)
 
 //////////
 // other constructors
@@ -90,7 +104,7 @@ clifford::clifford(unsigned char rl, const exvector & v, bool discardable) : inh
 	tinfo_key = TINFO_clifford;
 }
 
-clifford::clifford(unsigned char rl, exvector * vp) : inherited(sy_none(), vp), representation_label(rl)
+clifford::clifford(unsigned char rl, std::auto_ptr<exvector> vp) : inherited(sy_none(), vp), representation_label(rl)
 {
 	tinfo_key = TINFO_clifford;
 }
@@ -99,7 +113,7 @@ clifford::clifford(unsigned char rl, exvector * vp) : inherited(sy_none(), vp),
 // archiving
 //////////
 
-clifford::clifford(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+clifford::clifford(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	unsigned rl;
 	n.find_unsigned("label", rl);
@@ -116,6 +130,8 @@ DEFAULT_UNARCHIVE(clifford)
 DEFAULT_ARCHIVING(diracone)
 DEFAULT_ARCHIVING(diracgamma)
 DEFAULT_ARCHIVING(diracgamma5)
+DEFAULT_ARCHIVING(diracgammaL)
+DEFAULT_ARCHIVING(diracgammaR)
 
 //////////
 // functions overriding virtual functions from base classes
@@ -142,33 +158,45 @@ bool clifford::match_same_type(const basic & other) const
 	return representation_label == o.representation_label;
 }
 
-void clifford::print(const print_context & c, unsigned level) const
+static bool is_dirac_slash(const ex & seq0)
 {
-	if (!is_a<diracgamma5>(seq[0]) && !is_a<diracgamma>(seq[0]) && !is_a<diracone>(seq[0])) {
-
-		// dirac_slash() object is printed differently
-		if (is_a<print_tree>(c))
-			inherited::print(c, level);
-		else if (is_a<print_latex>(c)) {
-			c.s << "{";
-			seq[0].print(c, level);
-			c.s << "\\hspace{-1.0ex}/}";
-		} else {
-			seq[0].print(c, level);
-			c.s << "\\";
-		}
+	return !is_a<diracgamma5>(seq0) && !is_a<diracgammaL>(seq0) &&
+	       !is_a<diracgammaR>(seq0) && !is_a<diracgamma>(seq0) &&
+	       !is_a<diracone>(seq0);
+}
 
+void clifford::do_print_dflt(const print_dflt & c, unsigned level) const
+{
+	// dirac_slash() object is printed differently
+	if (is_dirac_slash(seq[0])) {
+		seq[0].print(c, level);
+		c.s << "\\";
 	} else
-		inherited::print(c, level);
+		this->print_dispatch<inherited>(c, level);
+}
+
+void clifford::do_print_latex(const print_latex & c, unsigned level) const
+{
+	// dirac_slash() object is printed differently
+	if (is_dirac_slash(seq[0])) {
+		c.s << "{";
+		seq[0].print(c, level);
+		c.s << "\\hspace{-1.0ex}/}";
+	} else
+		this->print_dispatch<inherited>(c, level);
 }
 
 DEFAULT_COMPARE(diracone)
 DEFAULT_COMPARE(diracgamma)
 DEFAULT_COMPARE(diracgamma5)
+DEFAULT_COMPARE(diracgammaL)
+DEFAULT_COMPARE(diracgammaR)
 
 DEFAULT_PRINT_LATEX(diracone, "ONE", "\\mathbb{1}")
 DEFAULT_PRINT_LATEX(diracgamma, "gamma", "\\gamma")
 DEFAULT_PRINT_LATEX(diracgamma5, "gamma5", "{\\gamma^5}")
+DEFAULT_PRINT_LATEX(diracgammaL, "gammaL", "{\\gamma_L}")
+DEFAULT_PRINT_LATEX(diracgammaR, "gammaR", "{\\gamma_R}")
 
 /** This function decomposes gamma~mu -> (1, mu) and a\ -> (a.ix, ix) */
 static void base_and_index(const ex & c, ex & b, ex & i)
@@ -179,6 +207,9 @@ static void base_and_index(const ex & c, ex & b, ex & i)
 	if (is_a<diracgamma>(c.op(0))) { // proper dirac gamma object
 		i = c.op(1);
 		b = _ex1;
+	} else if (is_a<diracgamma5>(c.op(0)) || is_a<diracgammaL>(c.op(0)) || is_a<diracgammaR>(c.op(0))) { // gamma5/L/R
+		i = _ex0;
+		b = _ex1;
 	} else { // slash object, generate new dummy index
 		varidx ix((new symbol)->setflag(status_flags::dynallocated), ex_to<idx>(c.op(1)).get_dim());
 		b = indexed(c.op(0), ix.toggle_variance());
@@ -194,9 +225,15 @@ bool diracgamma::contract_with(exvector::iterator self, exvector::iterator other
 	GINAC_ASSERT(is_a<diracgamma>(self->op(0)));
 	unsigned char rl = ex_to<clifford>(*self).get_representation_label();
 
+	ex dim = ex_to<idx>(self->op(1)).get_dim();
+	if (other->nops() > 1)
+		dim = minimal_dim(dim, ex_to<idx>(other->op(1)).get_dim());
+
 	if (is_a<clifford>(*other)) {
 
-		ex dim = ex_to<idx>(self->op(1)).get_dim();
+		// Contraction only makes sense if the represenation labels are equal
+		if (ex_to<clifford>(*other).get_representation_label() != rl)
+			return false;
 
 		// gamma~mu gamma.mu = dim ONE
 		if (other - self == 1) {
@@ -259,15 +296,22 @@ bool diracgamma::contract_with(exvector::iterator self, exvector::iterator other
 			*other = _ex1;
 			return true;
 		}
+
+	} else if (is_a<symbol>(other->op(0)) && other->nops() == 2) {
+
+		// x.mu gamma~mu -> x-slash
+		*self = dirac_slash(other->op(0), dim, rl);
+		*other = _ex1;
+		return true;
 	}
 
 	return false;
 }
 
 /** Perform automatic simplification on noncommutative product of clifford
- *  objects. This removes superfluous ONEs, permutes gamma5's to the front
+ *  objects. This removes superfluous ONEs, permutes gamma5/L/R's to the front
  *  and removes squares of gamma objects. */
-ex clifford::simplify_ncmul(const exvector & v) const
+ex clifford::eval_ncmul(const exvector & v) const
 {
 	exvector s;
 	s.reserve(v.size());
@@ -283,17 +327,67 @@ ex clifford::simplify_ncmul(const exvector & v) const
 	bool something_changed = false;
 	int sign = 1;
 
-	// Anticommute gamma5's to the front
+	// Anticommute gamma5/L/R's to the front
 	if (s.size() >= 2) {
 		exvector::iterator first = s.begin(), next_to_last = s.end() - 2;
 		while (true) {
 			exvector::iterator it = next_to_last;
 			while (true) {
 				exvector::iterator it2 = it + 1;
-				if (is_a<clifford>(*it) && is_a<clifford>(*it2) && !is_a<diracgamma5>(it->op(0)) && is_a<diracgamma5>(it2->op(0))) {
-					it->swap(*it2);
-					sign = -sign;
-					something_changed = true;
+				if (is_a<clifford>(*it) && is_a<clifford>(*it2)) {
+					ex e1 = it->op(0), e2 = it2->op(0);
+
+					if (is_a<diracgamma5>(e2)) {
+
+						if (is_a<diracgammaL>(e1) || is_a<diracgammaR>(e1)) {
+
+							// gammaL/R gamma5 -> gamma5 gammaL/R
+							it->swap(*it2);
+							something_changed = true;
+
+						} else if (!is_a<diracgamma5>(e1)) {
+
+							// gamma5 gamma5 -> gamma5 gamma5 (do nothing)
+							// x gamma5 -> -gamma5 x
+							it->swap(*it2);
+							sign = -sign;
+							something_changed = true;
+						}
+
+					} else if (is_a<diracgammaL>(e2)) {
+
+						if (is_a<diracgammaR>(e1)) {
+
+							// gammaR gammaL -> 0
+							return _ex0;
+
+						} else if (!is_a<diracgammaL>(e1) && !is_a<diracgamma5>(e1)) {
+
+							// gammaL gammaL -> gammaL gammaL (do nothing)
+							// gamma5 gammaL -> gamma5 gammaL (do nothing)
+							// x gammaL -> gammaR x
+							it->swap(*it2);
+							*it = clifford(diracgammaR(), ex_to<clifford>(*it).get_representation_label());
+							something_changed = true;
+						}
+
+					} else if (is_a<diracgammaR>(e2)) {
+
+						if (is_a<diracgammaL>(e1)) {
+
+							// gammaL gammaR -> 0
+							return _ex0;
+
+						} else if (!is_a<diracgammaR>(e1) && !is_a<diracgamma5>(e1)) {
+
+							// gammaR gammaR -> gammaR gammaR (do nothing)
+							// gamma5 gammaR -> gamma5 gammaR (do nothing)
+							// x gammaR -> gammaL x
+							it->swap(*it2);
+							*it = clifford(diracgammaL(), ex_to<clifford>(*it).get_representation_label());
+							something_changed = true;
+						}
+					}
 				}
 				if (it == first)
 					break;
@@ -305,12 +399,6 @@ ex clifford::simplify_ncmul(const exvector & v) const
 		}
 	}
 
-	// Remove squares of gamma5
-	while (s.size() >= 2 && is_a<clifford>(s[0]) && is_a<clifford>(s[1]) && is_a<diracgamma5>(s[0].op(0)) && is_a<diracgamma5>(s[1].op(0))) {
-		s.erase(s.begin(), s.begin() + 2);
-		something_changed = true;
-	}
-
 	// Remove equal adjacent gammas
 	if (s.size() >= 2) {
 		exvector::iterator it, itend = s.end() - 1;
@@ -319,7 +407,14 @@ ex clifford::simplify_ncmul(const exvector & v) const
 			ex & b = it[1];
 			if (!is_a<clifford>(a) || !is_a<clifford>(b))
 				continue;
-			if (is_a<diracgamma>(a.op(0)) && is_a<diracgamma>(b.op(0))) {
+
+			const ex & ag = a.op(0);
+			const ex & bg = b.op(0);
+			bool a_is_diracgamma = is_a<diracgamma>(ag);
+			bool b_is_diracgamma = is_a<diracgamma>(bg);
+
+			if (a_is_diracgamma && b_is_diracgamma) {
+
 				const ex & ia = a.op(1);
 				const ex & ib = b.op(1);
 				if (ia.is_equal(ib)) { // gamma~alpha gamma~alpha -> g~alpha~alpha
@@ -327,15 +422,46 @@ ex clifford::simplify_ncmul(const exvector & v) const
 					b = dirac_ONE(representation_label);
 					something_changed = true;
 				}
-			} else if (!is_a<diracgamma>(a.op(0)) && !is_a<diracgamma>(b.op(0))) {
-				const ex & ba = a.op(0);
-				const ex & bb = b.op(0);
-				if (ba.is_equal(bb)) { // a\ a\ -> a^2
-					varidx ix((new symbol)->setflag(status_flags::dynallocated), ex_to<idx>(a.op(1)).get_dim());
-					a = indexed(ba, ix) * indexed(bb, ix.toggle_variance());
-					b = dirac_ONE(representation_label);
-					something_changed = true;
-				}
+
+			} else if ((is_a<diracgamma5>(ag) && is_a<diracgamma5>(bg))) {
+
+				// Remove squares of gamma5
+				a = dirac_ONE(representation_label);
+				b = dirac_ONE(representation_label);
+				something_changed = true;
+
+			} else if ((is_a<diracgammaL>(ag) && is_a<diracgammaL>(bg))
+			        || (is_a<diracgammaR>(ag) && is_a<diracgammaR>(bg))) {
+
+				// Remove squares of gammaL/R
+				b = dirac_ONE(representation_label);
+				something_changed = true;
+
+			} else if (is_a<diracgammaL>(ag) && is_a<diracgammaR>(bg)) {
+
+				// gammaL and gammaR are orthogonal
+				return _ex0;
+
+			} else if (is_a<diracgamma5>(ag) && is_a<diracgammaL>(bg)) {
+
+				// gamma5 gammaL -> -gammaL
+				a = dirac_ONE(representation_label);
+				sign = -sign;
+				something_changed = true;
+
+			} else if (is_a<diracgamma5>(ag) && is_a<diracgammaR>(bg)) {
+
+				// gamma5 gammaR -> gammaR
+				a = dirac_ONE(representation_label);
+				something_changed = true;
+
+			} else if (!a_is_diracgamma && !b_is_diracgamma && ag.is_equal(bg)) {
+
+				// a\ a\ -> a^2
+				varidx ix((new symbol)->setflag(status_flags::dynallocated), ex_to<idx>(a.op(1)).minimal_dim(ex_to<idx>(b.op(1))));
+				a = indexed(ag, ix) * indexed(ag, ix.toggle_variance());
+				b = dirac_ONE(representation_label);
+				something_changed = true;
 			}
 		}
 	}
@@ -343,21 +469,36 @@ ex clifford::simplify_ncmul(const exvector & v) const
 	if (s.empty())
 		return clifford(diracone(), representation_label) * sign;
 	if (something_changed)
-		return nonsimplified_ncmul(s) * sign;
+		return reeval_ncmul(s) * sign;
 	else
-		return simplified_ncmul(s) * sign;
+		return hold_ncmul(s) * sign;
 }
 
-ex clifford::thisexprseq(const exvector & v) const
+ex clifford::thiscontainer(const exvector & v) const
 {
 	return clifford(representation_label, v);
 }
 
-ex clifford::thisexprseq(exvector * vp) const
+ex clifford::thiscontainer(std::auto_ptr<exvector> vp) const
 {
 	return clifford(representation_label, vp);
 }
 
+ex diracgamma5::conjugate() const
+{	
+	return _ex_1 * (*this);
+}
+
+ex diracgammaL::conjugate() const
+{
+	return (new diracgammaR)->setflag(status_flags::dynallocated);
+}
+
+ex diracgammaR::conjugate() const
+{
+	return (new diracgammaL)->setflag(status_flags::dynallocated);
+}
+
 //////////
 // global functions
 //////////
@@ -380,14 +521,14 @@ ex dirac_gamma5(unsigned char rl)
 	return clifford(diracgamma5(), rl);
 }
 
-ex dirac_gamma6(unsigned char rl)
+ex dirac_gammaL(unsigned char rl)
 {
-	return clifford(diracone(), rl) + clifford(diracgamma5(), rl);
+	return clifford(diracgammaL(), rl);
 }
 
-ex dirac_gamma7(unsigned char rl)
+ex dirac_gammaR(unsigned char rl)
 {
-	return clifford(diracone(), rl) - clifford(diracgamma5(), rl);
+	return clifford(diracgammaR(), rl);
 }
 
 ex dirac_slash(const ex & e, const ex & dim, unsigned char rl)
@@ -414,13 +555,13 @@ static bool is_clifford_tinfo(unsigned ti)
 
 /** Take trace of a string of an even number of Dirac gammas given a vector
  *  of indices. */
-static ex trace_string(exvector::const_iterator ix, unsigned num)
+static ex trace_string(exvector::const_iterator ix, size_t num)
 {
 	// Tr gamma.mu gamma.nu = 4 g.mu.nu
 	if (num == 2)
 		return lorentz_g(ix[0], ix[1]);
 
-	// Tr gamma.mu gamma.nu gamma.rho gamma.sig = 4 (g.mu.nu g.rho.sig + g.nu.rho g.mu.sig - g.mu.rho g.nu.sig
+	// Tr gamma.mu gamma.nu gamma.rho gamma.sig = 4 (g.mu.nu g.rho.sig + g.nu.rho g.mu.sig - g.mu.rho g.nu.sig )
 	else if (num == 4)
 		return lorentz_g(ix[0], ix[1]) * lorentz_g(ix[2], ix[3])
 		     + lorentz_g(ix[1], ix[2]) * lorentz_g(ix[0], ix[3])
@@ -436,8 +577,8 @@ static ex trace_string(exvector::const_iterator ix, unsigned num)
 	exvector v(num - 2);
 	int sign = 1;
 	ex result;
-	for (unsigned i=1; i<num; i++) {
-		for (unsigned n=1, j=0; n<num; n++) {
+	for (size_t i=1; i<num; i++) {
+		for (size_t n=1, j=0; n<num; n++) {
 			if (n == i)
 				continue;
 			v[j++] = ix[n];
@@ -452,17 +593,21 @@ ex dirac_trace(const ex & e, unsigned char rl, const ex & trONE)
 {
 	if (is_a<clifford>(e)) {
 
-		if (ex_to<clifford>(e).get_representation_label() == rl
-		 && is_a<diracone>(e.op(0)))
+		if (!ex_to<clifford>(e).get_representation_label() == rl)
+			return _ex0;
+		const ex & g = e.op(0);
+		if (is_a<diracone>(g))
 			return trONE;
+		else if (is_a<diracgammaL>(g) || is_a<diracgammaR>(g))
+			return trONE/2;
 		else
 			return _ex0;
 
-	} else if (is_ex_exactly_of_type(e, mul)) {
+	} else if (is_exactly_a<mul>(e)) {
 
 		// Trace of product: pull out non-clifford factors
 		ex prod = _ex1;
-		for (unsigned i=0; i<e.nops(); i++) {
+		for (size_t i=0; i<e.nops(); i++) {
 			const ex &o = e.op(i);
 			if (is_clifford_tinfo(o.return_type_tinfo(), rl))
 				prod *= dirac_trace(o, rl, trONE);
@@ -471,19 +616,22 @@ ex dirac_trace(const ex & e, unsigned char rl, const ex & trONE)
 		}
 		return prod;
 
-	} else if (is_ex_exactly_of_type(e, ncmul)) {
+	} else if (is_exactly_a<ncmul>(e)) {
 
 		if (!is_clifford_tinfo(e.return_type_tinfo(), rl))
 			return _ex0;
 
-		// Expand product, if necessary
-		ex e_expanded = e.expand();
+		// Substitute gammaL/R and expand product, if necessary
+		ex e_expanded = e.subs(lst(
+			dirac_gammaL(rl) == (dirac_ONE(rl)-dirac_gamma5(rl))/2,
+			dirac_gammaR(rl) == (dirac_ONE(rl)+dirac_gamma5(rl))/2
+		), subs_options::no_pattern).expand();
 		if (!is_a<ncmul>(e_expanded))
 			return dirac_trace(e_expanded, rl, trONE);
 
 		// gamma5 gets moved to the front so this check is enough
 		bool has_gamma5 = is_a<diracgamma5>(e.op(0).op(0));
-		unsigned num = e.nops();
+		size_t num = e.nops();
 
 		if (has_gamma5) {
 
@@ -493,42 +641,44 @@ ex dirac_trace(const ex & e, unsigned char rl, const ex & trONE)
 				return _ex0;
 
 			// Tr gamma5 gamma.mu gamma.nu gamma.rho gamma.sigma = 4I * epsilon(mu, nu, rho, sigma)
+			// (the epsilon is always 4-dimensional)
 			if (num == 5) {
 				ex b1, i1, b2, i2, b3, i3, b4, i4;
 				base_and_index(e.op(1), b1, i1);
 				base_and_index(e.op(2), b2, i2);
 				base_and_index(e.op(3), b3, i3);
 				base_and_index(e.op(4), b4, i4);
-				return trONE * I * (eps0123(i1, i2, i3, i4) * b1 * b2 * b3 * b4).simplify_indexed();
+				return trONE * I * (lorentz_eps(ex_to<idx>(i1).replace_dim(_ex4), ex_to<idx>(i2).replace_dim(_ex4), ex_to<idx>(i3).replace_dim(_ex4), ex_to<idx>(i4).replace_dim(_ex4)) * b1 * b2 * b3 * b4).simplify_indexed();
 			}
 
 	   		// Tr gamma5 S_2k =
 			//   I/4! * epsilon0123.mu1.mu2.mu3.mu4 * Tr gamma.mu1 gamma.mu2 gamma.mu3 gamma.mu4 S_2k
+			// (the epsilon is always 4-dimensional)
 			exvector ix(num-1), bv(num-1);
-			for (unsigned i=1; i<num; i++)
+			for (size_t i=1; i<num; i++)
 				base_and_index(e.op(i), bv[i-1], ix[i-1]);
 			num--;
 			int *iv = new int[num];
 			ex result;
-			for (unsigned i=0; i<num-3; i++) {
+			for (size_t i=0; i<num-3; i++) {
 				ex idx1 = ix[i];
-				for (unsigned j=i+1; j<num-2; j++) {
+				for (size_t j=i+1; j<num-2; j++) {
 					ex idx2 = ix[j];
-					for (unsigned k=j+1; k<num-1; k++) {
+					for (size_t k=j+1; k<num-1; k++) {
 						ex idx3 = ix[k];
-						for (unsigned l=k+1; l<num; l++) {
+						for (size_t l=k+1; l<num; l++) {
 							ex idx4 = ix[l];
 							iv[0] = i; iv[1] = j; iv[2] = k; iv[3] = l;
 							exvector v;
 							v.reserve(num - 4);
-							for (unsigned n=0, t=4; n<num; n++) {
+							for (size_t n=0, t=4; n<num; n++) {
 								if (n == i || n == j || n == k || n == l)
 									continue;
 								iv[t++] = n;
 								v.push_back(ix[n]);
 							}
 							int sign = permutation_sign(iv, iv + num);
-							result += sign * eps0123(idx1, idx2, idx3, idx4)
+							result += sign * lorentz_eps(ex_to<idx>(idx1).replace_dim(_ex4), ex_to<idx>(idx2).replace_dim(_ex4), ex_to<idx>(idx3).replace_dim(_ex4), ex_to<idx>(idx4).replace_dim(_ex4))
 							        * trace_string(v.begin(), num - 4);
 						}
 					}
@@ -552,7 +702,7 @@ ex dirac_trace(const ex & e, unsigned char rl, const ex & trONE)
 			}
 
 			exvector iv(num), bv(num);
-			for (unsigned i=0; i<num; i++)
+			for (size_t i=0; i<num; i++)
 				base_and_index(e.op(i), bv[i], iv[i]);
 
 			return trONE * (trace_string(iv.begin(), num) * mul(bv)).simplify_indexed();
@@ -571,21 +721,21 @@ ex dirac_trace(const ex & e, unsigned char rl, const ex & trONE)
 ex canonicalize_clifford(const ex & e)
 {
 	// Scan for any ncmul objects
-	lst srl;
+	exmap srl;
 	ex aux = e.to_rational(srl);
-	for (unsigned i=0; i<srl.nops(); i++) {
+	for (exmap::iterator i = srl.begin(); i != srl.end(); ++i) {
 
-		ex lhs = srl.op(i).lhs();
-		ex rhs = srl.op(i).rhs();
+		ex lhs = i->first;
+		ex rhs = i->second;
 
-		if (is_ex_exactly_of_type(rhs, ncmul)
+		if (is_exactly_a<ncmul>(rhs)
 		 && rhs.return_type() == return_types::noncommutative
 		 && is_clifford_tinfo(rhs.return_type_tinfo())) {
 
 			// Expand product, if necessary
 			ex rhs_expanded = rhs.expand();
 			if (!is_a<ncmul>(rhs_expanded)) {
-				srl.let_op(i) = (lhs == canonicalize_clifford(rhs_expanded));
+				i->second = canonicalize_clifford(rhs_expanded);
 				continue;
 
 			} else if (!is_a<clifford>(rhs.op(0)))
@@ -593,12 +743,12 @@ ex canonicalize_clifford(const ex & e)
 
 			exvector v;
 			v.reserve(rhs.nops());
-			for (unsigned j=0; j<rhs.nops(); j++)
+			for (size_t j=0; j<rhs.nops(); j++)
 				v.push_back(rhs.op(j));
 
 			// Stupid recursive bubble sort because we only want to swap adjacent gammas
 			exvector::iterator it = v.begin(), next_to_last = v.end() - 1;
-			if (is_a<diracgamma5>(it->op(0)))
+			if (is_a<diracgamma5>(it->op(0)) || is_a<diracgammaL>(it->op(0)) || is_a<diracgammaR>(it->op(0)))
 				++it;
 			while (it != next_to_last) {
 				if (it[0].compare(it[1]) > 0) {
@@ -607,12 +757,12 @@ ex canonicalize_clifford(const ex & e)
 					base_and_index(it[0], b1, i1);
 					base_and_index(it[1], b2, i2);
 					it[0] = (lorentz_g(i1, i2) * b1 * b2).simplify_indexed();
-					it[1] = _ex2;
+					it[1] = v.size() == 2 ? _ex2 * dirac_ONE(ex_to<clifford>(it[1]).get_representation_label()) : _ex2;
 					ex sum = ncmul(v);
 					it[0] = save1;
 					it[1] = save0;
 					sum -= ncmul(v, true);
-					srl.let_op(i) = (lhs == canonicalize_clifford(sum));
+					i->second = canonicalize_clifford(sum);
 					goto next_sym;
 				}
 				++it;
@@ -620,7 +770,7 @@ ex canonicalize_clifford(const ex & e)
 next_sym:	;
 		}
 	}
-	return aux.subs(srl).simplify_indexed();
+	return aux.subs(srl, subs_options::no_pattern).simplify_indexed();
 }
 
 } // namespace GiNaC
