@@ -3,7 +3,7 @@
  *  Interface to GiNaC's color (SU(3) Lie algebra) objects. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -47,20 +47,20 @@ public:
 
 	// internal constructors
 	color(unsigned char rl, const exvector & v, bool discardable = false);
-	color(unsigned char rl, exvector * vp); // vp will be deleted
+	color(unsigned char rl, std::auto_ptr<exvector> vp);
 
 	// functions overriding virtual functions from base classes
 protected:
+	ex eval_ncmul(const exvector & v) const;
 	bool match_same_type(const basic & other) const;
-	ex simplify_ncmul(const exvector & v) const;
-	ex thisexprseq(const exvector & v) const;
-	ex thisexprseq(exvector * vp) const;
-	unsigned return_type(void) const { return return_types::noncommutative; }
-	unsigned return_type_tinfo(void) const { return TINFO_color + representation_label; }
+	ex thiscontainer(const exvector & v) const;
+	ex thiscontainer(std::auto_ptr<exvector> vp) const;
+	unsigned return_type() const { return return_types::noncommutative; }
+	unsigned return_type_tinfo() const { return TINFO_color + representation_label; }
 
 	// non-virtual functions in this class
 public:
-	unsigned char get_representation_label(void) const {return representation_label;}
+	unsigned char get_representation_label() const {return representation_label;}
 
 	// member variables
 private:
@@ -73,9 +73,10 @@ class su3one : public tensor
 {
 	GINAC_DECLARE_REGISTERED_CLASS(su3one, tensor)
 
-	// functions overriding virtual functions from base classes
-public:
-	void print(const print_context & c, unsigned level = 0) const;
+	// non-virtual functions in this class
+protected:
+	void do_print(const print_context & c, unsigned level) const;
+	void do_print_latex(const print_latex & c, unsigned level) const;
 };
 
 /** This class represents an su(3) generator. */
@@ -85,8 +86,12 @@ class su3t : public tensor
 
 	// functions overriding virtual functions from base classes
 public:
-	void print(const print_context & c, unsigned level = 0) const;
 	bool contract_with(exvector::iterator self, exvector::iterator other, exvector & v) const;
+
+	// non-virtual functions in this class
+protected:
+	void do_print(const print_context & c, unsigned level) const;
+	void do_print_latex(const print_latex & c, unsigned level) const;
 };
 
 /** This class represents the tensor of antisymmetric su(3) structure
@@ -97,9 +102,13 @@ class su3f : public tensor
 
 	// functions overriding virtual functions from base classes
 public:
-	void print(const print_context & c, unsigned level = 0) const;
 	ex eval_indexed(const basic & i) const;
 	bool contract_with(exvector::iterator self, exvector::iterator other, exvector & v) const;
+
+	// non-virtual functions in this class
+protected:
+	void do_print(const print_context & c, unsigned level) const;
+	void do_print_latex(const print_latex & c, unsigned level) const;
 };
 
 /** This class represents the tensor of symmetric su(3) structure constants. */
@@ -109,9 +118,13 @@ class su3d : public tensor
 
 	// functions overriding virtual functions from base classes
 public:
-	void print(const print_context & c, unsigned level = 0) const;
 	ex eval_indexed(const basic & i) const;
 	bool contract_with(exvector::iterator self, exvector::iterator other, exvector & v) const;
+
+	// non-virtual functions in this class
+protected:
+	void do_print(const print_context & c, unsigned level) const;
+	void do_print_latex(const print_latex & c, unsigned level) const;
 };
 
 
