@@ -1,4 +1,24 @@
-$maxargs=13;
+#  This perl script automatically generates function.h and function.cpp
+
+#  function.pl options: \$maxargs=${maxargs}
+# 
+#  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
+# 
+#  This program is free software; you can redistribute it and/or modify
+#  it under the terms of the GNU General Public License as published by
+#  the Free Software Foundation; either version 2 of the License, or
+#  (at your option) any later version.
+# 
+#  This program is distributed in the hope that it will be useful,
+#  but WITHOUT ANY WARRANTY; without even the implied warranty of
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#  GNU General Public License for more details.
+# 
+#  You should have received a copy of the GNU General Public License
+#  along with this program; if not, write to the Free Software
+#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+$maxargs=14;
 
 sub generate_seq {
 	my ($seq_template,$n)=@_;
@@ -39,10 +59,9 @@ sub generate {
 $declare_function_macro = generate(
 	<<'END_OF_DECLARE_FUNCTION_MACRO','typename T${N}','const T${N} & p${N}','GiNaC::ex(p${N})');
 #define DECLARE_FUNCTION_${N}P(NAME) \\
-extern const unsigned function_index_##NAME; \\
-template<${SEQ1}> \\
-inline const GiNaC::function NAME(${SEQ2}) { \\
-	return GiNaC::function(function_index_##NAME, ${SEQ3}); \\
+class NAME##_SERIAL { public: static unsigned serial; }; \\
+template<${SEQ1}> const GiNaC::function NAME(${SEQ2}) { \\
+	return GiNaC::function(NAME##_SERIAL::serial, ${SEQ3}); \\
 }
 
 END_OF_DECLARE_FUNCTION_MACRO
@@ -55,6 +74,10 @@ $typedef_evalf_funcp=generate(
 'typedef ex (* evalf_funcp_${N})(${SEQ1});'."\n",
 'const ex &','','');
 
+$typedef_conjugate_funcp=generate(
+'typedef ex (* conjugate_funcp_${N})(${SEQ1});'."\n",
+'const ex &','','');
+
 $typedef_derivative_funcp=generate(
 'typedef ex (* derivative_funcp_${N})(${SEQ1}, unsigned);'."\n",
 'const ex &','','');
@@ -63,14 +86,30 @@ $typedef_series_funcp=generate(
 'typedef ex (* series_funcp_${N})(${SEQ1}, const relational &, int, unsigned);'."\n",
 'const ex &','','');
 
+$typedef_print_funcp=generate(
+'typedef void (* print_funcp_${N})(${SEQ1}, const print_context &);'."\n",
+'const ex &','','');
+
 $eval_func_interface=generate('    function_options & eval_func(eval_funcp_${N} e);'."\n",'','','');
 
 $evalf_func_interface=generate('    function_options & evalf_func(evalf_funcp_${N} ef);'."\n",'','','');
 
+$conjugate_func_interface=generate('    function_options & conjugate_func(conjugate_funcp_${N} d);'."\n",'','','');
+
 $derivative_func_interface=generate('    function_options & derivative_func(derivative_funcp_${N} d);'."\n",'','','');
 
 $series_func_interface=generate('    function_options & series_func(series_funcp_${N} s);'."\n",'','','');
 
+$print_func_interface=generate(
+	<<'END_OF_PRINT_FUNC_INTERFACE','','','');
+    template <class Ctx> function_options & print_func(print_funcp_${N} p)
+    {
+    	test_and_set_nparams(${N});
+    	set_print_func(Ctx::get_class_info_static().options.get_id(), print_funcp(p));
+    	return *this;
+    }
+END_OF_PRINT_FUNC_INTERFACE
+
 $constructors_interface=generate(
 '    function(unsigned ser, ${SEQ1});'."\n",
 'const ex & param${N}','','');
@@ -87,33 +126,46 @@ END_OF_CONSTRUCTORS_IMPLEMENTATION
 $eval_switch_statement=generate(
 	<<'END_OF_EVAL_SWITCH_STATEMENT','seq[${N}-1]','','');
 	case ${N}:
-		eval_result = ((eval_funcp_${N})(registered_functions()[serial].eval_f))(${SEQ1});
+		eval_result = ((eval_funcp_${N})(opt.eval_f))(${SEQ1});
 		break;
 END_OF_EVAL_SWITCH_STATEMENT
 
 $evalf_switch_statement=generate(
 	<<'END_OF_EVALF_SWITCH_STATEMENT','eseq[${N}-1]','','');
 	case ${N}:
-		return ((evalf_funcp_${N})(registered_functions()[serial].evalf_f))(${SEQ1});
+		return ((evalf_funcp_${N})(opt.evalf_f))(${SEQ1});
 END_OF_EVALF_SWITCH_STATEMENT
 
+$conjugate_switch_statement=generate(
+	<<'END_OF_DIFF_SWITCH_STATEMENT','seq[${N}-1]','','');
+	case ${N}:
+		return ((conjugate_funcp_${N})(opt.conjugate_f))(${SEQ1});
+END_OF_DIFF_SWITCH_STATEMENT
+
 $diff_switch_statement=generate(
 	<<'END_OF_DIFF_SWITCH_STATEMENT','seq[${N}-1]','','');
 	case ${N}:
-		return ((derivative_funcp_${N})(registered_functions()[serial].derivative_f))(${SEQ1},diff_param);
+		return ((derivative_funcp_${N})(opt.derivative_f))(${SEQ1},diff_param);
 END_OF_DIFF_SWITCH_STATEMENT
 
 $series_switch_statement=generate(
 	<<'END_OF_SERIES_SWITCH_STATEMENT','seq[${N}-1]','','');
 	case ${N}:
 		try {
-			res = ((series_funcp_${N})(registered_functions()[serial].series_f))(${SEQ1},r,order,options);
+			res = ((series_funcp_${N})(opt.series_f))(${SEQ1},r,order,options);
 		} catch (do_taylor) {
 			res = basic::series(r, order, options);
 		}
 		return res;
 END_OF_SERIES_SWITCH_STATEMENT
 
+$print_switch_statement=generate(
+	<<'END_OF_PRINT_SWITCH_STATEMENT','seq[${N}-1]','','');
+		case ${N}:
+			((print_funcp_${N})(pdt[id]))(${SEQ1}, c);
+			break;
+END_OF_PRINT_SWITCH_STATEMENT
+
 $eval_func_implementation=generate(
 	<<'END_OF_EVAL_FUNC_IMPLEMENTATION','','','');
 function_options & function_options::eval_func(eval_funcp_${N} e)
@@ -134,6 +186,16 @@ function_options & function_options::evalf_func(evalf_funcp_${N} ef)
 }
 END_OF_EVALF_FUNC_IMPLEMENTATION
 
+$conjugate_func_implementation=generate(
+	<<'END_OF_CONJUGATE_FUNC_IMPLEMENTATION','','','');
+function_options & function_options::conjugate_func(conjugate_funcp_${N} c)
+{
+	test_and_set_nparams(${N});
+	conjugate_f = conjugate_funcp(c);
+	return *this;
+}
+END_OF_CONJUGATE_FUNC_IMPLEMENTATION
+
 $derivative_func_implementation=generate(
 	<<'END_OF_DERIVATIVE_FUNC_IMPLEMENTATION','','','');
 function_options & function_options::derivative_func(derivative_funcp_${N} d)
@@ -164,7 +226,7 @@ $interface=<<END_OF_INTERFACE;
  *  Please do not modify it directly, edit the perl script instead!
  *  function.pl options: \$maxargs=${maxargs}
  *
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -197,7 +259,7 @@ $declare_function_macro
 // end of generated lines
 
 #define REGISTER_FUNCTION(NAME,OPT) \\
-const unsigned function_index_##NAME= \\
+unsigned NAME##_SERIAL::serial = \\
 	GiNaC::function::register_new(GiNaC::function_options(#NAME).OPT);
 
 namespace GiNaC {
@@ -207,22 +269,29 @@ class symmetry;
 
 typedef ex (* eval_funcp)();
 typedef ex (* evalf_funcp)();
+typedef ex (* conjugate_funcp)();
 typedef ex (* derivative_funcp)();
 typedef ex (* series_funcp)();
+typedef void (* print_funcp)();
 
 // the following lines have been generated for max. ${maxargs} parameters
 $typedef_eval_funcp
 $typedef_evalf_funcp
+$typedef_conjugate_funcp
 $typedef_derivative_funcp
 $typedef_series_funcp
+$typedef_print_funcp
 // end of generated lines
 
 // Alternatively, an exvector may be passed into the static function, instead
 // of individual ex objects.  Then, the number of arguments is not limited.
 typedef ex (* eval_funcp_exvector)(const exvector &);
 typedef ex (* evalf_funcp_exvector)(const exvector &);
+typedef ex (* conjugate_funcp_exvector)(const exvector &);
 typedef ex (* derivative_funcp_exvector)(const exvector &, unsigned);
 typedef ex (* series_funcp_exvector)(const exvector &, const relational &, int, unsigned);
+typedef void (* print_funcp_exvector)(const exvector &, const print_context &);
+
 
 class function_options
 {
@@ -232,32 +301,47 @@ public:
 	function_options();
 	function_options(std::string const & n, std::string const & tn=std::string());
 	~function_options();
-	void initialize(void);
+	void initialize();
+
+	function_options & dummy() { return *this; }
 	function_options & set_name(std::string const & n, std::string const & tn=std::string());
 	function_options & latex_name(std::string const & tn);
 // the following lines have been generated for max. ${maxargs} parameters
 $eval_func_interface
 $evalf_func_interface
+$conjugate_func_interface
 $derivative_func_interface
 $series_func_interface
+$print_func_interface
 // end of generated lines
 	function_options & eval_func(eval_funcp_exvector e);
 	function_options & evalf_func(evalf_funcp_exvector ef);
+	function_options & conjugate_func(conjugate_funcp_exvector d);
 	function_options & derivative_func(derivative_funcp_exvector d);
 	function_options & series_func(series_funcp_exvector s);
 
+	template <class Ctx> function_options & print_func(print_funcp_exvector p)
+	{
+		print_use_exvector_args = true;
+		set_print_func(Ctx::get_class_info_static().options.get_id(), print_funcp(p));
+		return *this;
+	}
+
 	function_options & set_return_type(unsigned rt, unsigned rtt=0);
-	function_options & do_not_evalf_params(void);
+	function_options & do_not_evalf_params();
 	function_options & remember(unsigned size, unsigned assoc_size=0,
 	                            unsigned strategy=remember_strategies::delete_never);
 	function_options & overloaded(unsigned o);
 	function_options & set_symmetry(const symmetry & s);
-	void test_and_set_nparams(unsigned n);
-	std::string get_name(void) const { return name; }
-	unsigned get_nparams(void) const { return nparams; }
-	bool has_derivative(void) const { return derivative_f != NULL; }
+
+	std::string get_name() const { return name; }
+	unsigned get_nparams() const { return nparams; }
 
 protected:
+	bool has_derivative() const { return derivative_f != NULL; }
+	void test_and_set_nparams(unsigned n);
+	void set_print_func(unsigned id, print_funcp f);
+
 	std::string name;
 	std::string TeX_name;
 
@@ -265,8 +349,10 @@ protected:
 
 	eval_funcp eval_f;
 	evalf_funcp evalf_f;
+	conjugate_funcp conjugate_f;
 	derivative_funcp derivative_f;
 	series_funcp series_f;
+	std::vector<print_funcp> print_dispatch_table;
 
 	bool evalf_params_first;
 
@@ -281,14 +367,22 @@ protected:
 
 	bool eval_use_exvector_args;
 	bool evalf_use_exvector_args;
+	bool conjugate_use_exvector_args;
 	bool derivative_use_exvector_args;
 	bool series_use_exvector_args;
+	bool print_use_exvector_args;
 
 	unsigned functions_with_same_name;
 
 	ex symtree;
 };
 
+
+/** Exception class thrown by classes which provide their own series expansion
+ *  to signal that ordinary Taylor expansion is safe. */
+class do_taylor {};
+
+
 /** The class function is used to implement builtin functions like sin, cos...
 	and user defined functions */
 class function : public exprseq
@@ -297,7 +391,7 @@ class function : public exprseq
 
 	// CINT has a linking problem
 #ifndef __MAKECINT__
-	friend void ginsh_get_ginac_functions(void);
+	friend void ginsh_get_ginac_functions();
 #endif // def __MAKECINT__
 
 	friend class remember_table_entry;
@@ -306,7 +400,7 @@ class function : public exprseq
 
 // member functions
 
-	// other ctors
+	// other constructors
 public:
 	function(unsigned ser);
 	// the following lines have been generated for max. ${maxargs} parameters
@@ -314,25 +408,26 @@ $constructors_interface
 	// end of generated lines
 	function(unsigned ser, const exprseq & es);
 	function(unsigned ser, const exvector & v, bool discardable = false);
-	function(unsigned ser, exvector * vp); // vp will be deleted
+	function(unsigned ser, std::auto_ptr<exvector> vp);
 
 	// functions overriding virtual functions from base classes
 public:
 	void print(const print_context & c, unsigned level = 0) const;
-	unsigned precedence(void) const {return 70;}
+	unsigned precedence() const {return 70;}
 	ex expand(unsigned options=0) const;
 	ex eval(int level=0) const;
 	ex evalf(int level=0) const;
-	unsigned calchash(void) const;
+	unsigned calchash() const;
 	ex series(const relational & r, int order, unsigned options = 0) const;
-	ex thisexprseq(const exvector & v) const;
-	ex thisexprseq(exvector * vp) const;
+	ex thiscontainer(const exvector & v) const;
+	ex thiscontainer(std::auto_ptr<exvector> vp) const;
+	ex conjugate() const;
 protected:
 	ex derivative(const symbol & s) const;
 	bool is_equal_same_type(const basic & other) const;
 	bool match_same_type(const basic & other) const;
-	unsigned return_type(void) const;
-	unsigned return_type_tinfo(void) const;
+	unsigned return_type() const;
+	unsigned return_type_tinfo() const;
 	
 	// new virtual functions which can be overridden by derived classes
 	// none
@@ -340,15 +435,15 @@ protected:
 	// non-virtual functions in this class
 protected:
 	ex pderivative(unsigned diff_param) const; // partial differentiation
-	static std::vector<function_options> & registered_functions(void);
+	static std::vector<function_options> & registered_functions();
 	bool lookup_remember_table(ex & result) const;
 	void store_remember_table(ex const & result) const;
 public:
 	static unsigned register_new(function_options const & opt);
 	static unsigned current_serial;
 	static unsigned find_function(const std::string &name, unsigned nparams);
-	unsigned get_serial(void) const {return serial;}
-	std::string get_name(void) const;
+	unsigned get_serial() const {return serial;}
+	std::string get_name() const;
 
 // member variables
 
@@ -364,8 +459,15 @@ template<> inline bool is_exactly_a<function>(const basic & obj)
 	return obj.tinfo()==TINFO_function;
 }
 
-#define is_ex_the_function(OBJ, FUNCNAME) \\
-	(GiNaC::is_exactly_a<GiNaC::function>(OBJ) && GiNaC::ex_to<GiNaC::function>(OBJ).get_serial() == function_index_##FUNCNAME)
+template <typename T>
+inline bool is_the_function(const ex & x)
+{
+	return is_exactly_a<function>(x)
+	    && ex_to<function>(x).get_serial() == T::serial;
+}
+
+// Check whether OBJ is the specified symbolic function.
+#define is_ex_the_function(OBJ, FUNCNAME) (GiNaC::is_the_function<FUNCNAME##_SERIAL>(OBJ))
 
 } // namespace GiNaC
 
@@ -383,7 +485,7 @@ $implementation=<<END_OF_IMPLEMENTATION;
  *  Please do not modify it directly, edit the perl script instead!
  *  function.pl options: \$maxargs=${maxargs}
  *
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -406,6 +508,7 @@ $implementation=<<END_OF_IMPLEMENTATION;
 #include <list>
 
 #include "function.h"
+#include "operators.h"
 #include "fderivative.h"
 #include "ex.h"
 #include "lst.h"
@@ -431,7 +534,7 @@ function_options::function_options()
 function_options::function_options(std::string const & n, std::string const & tn)
 {
 	initialize();
-	set_name(n,tn);
+	set_name(n, tn);
 }
 
 function_options::~function_options()
@@ -439,17 +542,19 @@ function_options::~function_options()
 	// nothing to clean up at the moment
 }
 
-void function_options::initialize(void)
+void function_options::initialize()
 {
-	set_name("unnamed_function","\\\\mbox{unnamed}");
+	set_name("unnamed_function", "\\\\mbox{unnamed}");
 	nparams = 0;
-	eval_f = evalf_f = derivative_f = series_f = 0;
+	eval_f = evalf_f = conjugate_f = derivative_f = series_f = 0;
 	evalf_params_first = true;
 	use_return_type = false;
 	eval_use_exvector_args = false;
 	evalf_use_exvector_args = false;
+	conjugate_use_exvector_args = false;
 	derivative_use_exvector_args = false;
 	series_use_exvector_args = false;
+	print_use_exvector_args = false;
 	use_remember = false;
 	functions_with_same_name = 1;
 	symtree = 0;
@@ -475,6 +580,7 @@ function_options & function_options::latex_name(std::string const & tn)
 // the following lines have been generated for max. ${maxargs} parameters
 $eval_func_implementation
 $evalf_func_implementation
+$conjugate_func_implementation
 $derivative_func_implementation
 $series_func_implementation
 // end of generated lines
@@ -491,6 +597,12 @@ function_options& function_options::evalf_func(evalf_funcp_exvector ef)
 	evalf_f = evalf_funcp(ef);
 	return *this;
 }
+function_options& function_options::conjugate_func(conjugate_funcp_exvector c)
+{
+	conjugate_use_exvector_args = true;
+	conjugate_f = conjugate_funcp(c);
+	return *this;
+}
 function_options& function_options::derivative_func(derivative_funcp_exvector d)
 {
 	derivative_use_exvector_args = true;
@@ -504,7 +616,6 @@ function_options& function_options::series_func(series_funcp_exvector s)
 	return *this;
 }
 
-
 function_options & function_options::set_return_type(unsigned rt, unsigned rtt)
 {
 	use_return_type = true;
@@ -513,7 +624,7 @@ function_options & function_options::set_return_type(unsigned rt, unsigned rtt)
 	return *this;
 }
 
-function_options & function_options::do_not_evalf_params(void)
+function_options & function_options::do_not_evalf_params()
 {
 	evalf_params_first = false;
 	return *this;
@@ -549,13 +660,20 @@ void function_options::test_and_set_nparams(unsigned n)
 	} else if (nparams!=n) {
 		// we do not throw an exception here because this code is
 		// usually executed before main(), so the exception could not
-		// caught anyhow
-		std::cerr << "WARNING: number of parameters ("
+		// be caught anyhow
+		std::cerr << "WARNING: " << name << "(): number of parameters ("
 		          << n << ") differs from number set before (" 
 		          << nparams << ")" << std::endl;
 	}
 }
 
+void function_options::set_print_func(unsigned id, print_funcp f)
+{
+	if (id >= print_dispatch_table.size())
+		print_dispatch_table.resize(id + 1);
+	print_dispatch_table[id] = f;
+}
+
 /** This can be used as a hook for external applications. */
 unsigned function::current_serial = 0;
 
@@ -563,7 +681,7 @@ unsigned function::current_serial = 0;
 GINAC_IMPLEMENT_REGISTERED_CLASS(function, exprseq)
 
 //////////
-// default ctor, dtor, copy ctor, assignment operator and helpers
+// default constructor
 //////////
 
 // public
@@ -573,22 +691,8 @@ function::function() : serial(0)
 	tinfo_key = TINFO_function;
 }
 
-// protected
-
-void function::copy(const function & other)
-{
-	inherited::copy(other);
-	serial = other.serial;
-}
-
-void function::destroy(bool call_parent)
-{
-	if (call_parent)
-		inherited::destroy(call_parent);
-}
-
 //////////
-// other ctors
+// other constructors
 //////////
 
 // public
@@ -605,6 +709,10 @@ $constructors_implementation
 function::function(unsigned ser, const exprseq & es) : exprseq(es), serial(ser)
 {
 	tinfo_key = TINFO_function;
+
+	// Force re-evaluation even if the exprseq was already evaluated
+	// (the exprseq copy constructor copies the flags)
+	clearflag(status_flags::evaluated);
 }
 
 function::function(unsigned ser, const exvector & v, bool discardable) 
@@ -613,7 +721,7 @@ function::function(unsigned ser, const exvector & v, bool discardable)
 	tinfo_key = TINFO_function;
 }
 
-function::function(unsigned ser, exvector * vp) 
+function::function(unsigned ser, std::auto_ptr<exvector> vp) 
   : exprseq(vp), serial(ser)
 {
 	tinfo_key = TINFO_function;
@@ -624,7 +732,7 @@ function::function(unsigned ser, exvector * vp)
 //////////
 
 /** Construct object from archive_node. */
-function::function(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+function::function(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	// Find serial number by function name
 	std::string s;
@@ -644,7 +752,7 @@ function::function(const archive_node &n, const lst &sym_lst) : inherited(n, sym
 }
 
 /** Unarchive the object. */
-ex function::unarchive(const archive_node &n, const lst &sym_lst)
+ex function::unarchive(const archive_node &n, lst &sym_lst)
 {
 	return (new function(n, sym_lst))->setflag(status_flags::dynallocated);
 }
@@ -666,44 +774,67 @@ void function::archive(archive_node &n) const
 void function::print(const print_context & c, unsigned level) const
 {
 	GINAC_ASSERT(serial<registered_functions().size());
+	const function_options &opt = registered_functions()[serial];
+	const std::vector<print_funcp> &pdt = opt.print_dispatch_table;
 
-	if (is_of_type(c, print_tree)) {
+	// Dynamically dispatch on print_context type
+	const print_context_class_info *pc_info = &c.get_class_info();
 
-		c.s << std::string(level, ' ') << class_name() << " "
-		    << registered_functions()[serial].name
-		    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
-		    << ", nops=" << nops()
-		    << std::endl;
-		unsigned delta_indent = static_cast<const print_tree &>(c).delta_indent;
-		for (unsigned i=0; i<seq.size(); ++i)
-			seq[i].print(c, level + delta_indent);
-		c.s << std::string(level + delta_indent, ' ') << "=====" << std::endl;
+next_context:
+	unsigned id = pc_info->options.get_id();
+	if (id >= pdt.size() || pdt[id] == NULL) {
 
-	} else if (is_of_type(c, print_csrc)) {
+		// Method not found, try parent print_context class
+		const print_context_class_info *parent_pc_info = pc_info->get_parent();
+		if (parent_pc_info) {
+			pc_info = parent_pc_info;
+			goto next_context;
+		}
 
-		// Print function name in lowercase
-		std::string lname = registered_functions()[serial].name;
-		unsigned num = lname.size();
-		for (unsigned i=0; i<num; i++)
-			lname[i] = tolower(lname[i]);
-		c.s << lname << "(";
+		// Method still not found, use default output
+		if (is_a<print_tree>(c)) {
 
-		// Print arguments, separated by commas
-		exvector::const_iterator it = seq.begin(), itend = seq.end();
-		while (it != itend) {
-			it->print(c);
-			++it;
-			if (it != itend)
-				c.s << ",";
+			c.s << std::string(level, ' ') << class_name() << " "
+			    << opt.name << " @" << this
+			    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
+			    << ", nops=" << nops()
+			    << std::endl;
+			unsigned delta_indent = static_cast<const print_tree &>(c).delta_indent;
+			for (size_t i=0; i<seq.size(); ++i)
+				seq[i].print(c, level + delta_indent);
+			c.s << std::string(level + delta_indent, ' ') << "=====" << std::endl;
+
+		} else if (is_a<print_csrc>(c)) {
+
+			// Print function name in lowercase
+			std::string lname = opt.name;
+			size_t num = lname.size();
+			for (size_t i=0; i<num; i++)
+				lname[i] = tolower(lname[i]);
+			c.s << lname;
+			printseq(c, '(', ',', ')', exprseq::precedence(), function::precedence());
+
+		} else if (is_a<print_latex>(c)) {
+			c.s << opt.TeX_name;
+			printseq(c, '(', ',', ')', exprseq::precedence(), function::precedence());
+		} else {
+			c.s << opt.name;
+			printseq(c, '(', ',', ')', exprseq::precedence(), function::precedence());
 		}
-		c.s << ")";
 
-	} else if (is_of_type(c, print_latex)) {
-		c.s << registered_functions()[serial].TeX_name;
-		printseq(c, '(', ',', ')', exprseq::precedence(), function::precedence());
 	} else {
-		c.s << registered_functions()[serial].name;
-		printseq(c, '(', ',', ')', exprseq::precedence(), function::precedence());
+
+		// Method found, call it
+		current_serial = serial;
+		if (opt.print_use_exvector_args)
+			((print_funcp_exvector)pdt[id])(seq, c);
+		else switch (opt.nparams) {
+			// the following lines have been generated for max. ${maxargs} parameters
+${print_switch_statement}
+			// end of generated lines
+		default:
+			throw(std::logic_error("function::print(): invalid nparams"));
+		}
 	}
 }
 
@@ -718,13 +849,12 @@ ex function::expand(unsigned options) const
 
 ex function::eval(int level) const
 {
-	GINAC_ASSERT(serial<registered_functions().size());
-
 	if (level>1) {
 		// first evaluate children, then we will end up here again
 		return function(serial,evalchildren(level));
 	}
 
+	GINAC_ASSERT(serial<registered_functions().size());
 	const function_options &opt = registered_functions()[serial];
 
 	// Canonicalize argument order according to the symmetry properties
@@ -736,7 +866,7 @@ ex function::eval(int level) const
 			// Something has changed while sorting arguments, more evaluations later
 			if (sig == 0)
 				return _ex0;
-			return ex(sig) * thisexprseq(v);
+			return ex(sig) * thiscontainer(v);
 		}
 	}
 
@@ -750,8 +880,8 @@ ex function::eval(int level) const
 		return eval_result;
 	}
 	current_serial = serial;
-	if (registered_functions()[serial].eval_use_exvector_args)
-		eval_result = ((eval_funcp_exvector)(registered_functions()[serial].eval_f))(seq);
+	if (opt.eval_use_exvector_args)
+		eval_result = ((eval_funcp_exvector)(opt.eval_f))(seq);
 	else
 	switch (opt.nparams) {
 		// the following lines have been generated for max. ${maxargs} parameters
@@ -769,29 +899,31 @@ ${eval_switch_statement}
 ex function::evalf(int level) const
 {
 	GINAC_ASSERT(serial<registered_functions().size());
+	const function_options &opt = registered_functions()[serial];
 
 	// Evaluate children first
 	exvector eseq;
-	if (level == 1)
+	if (level == 1 || !(opt.evalf_params_first))
 		eseq = seq;
 	else if (level == -max_recursion_level)
 		throw(std::runtime_error("max recursion level reached"));
-	else
+	else {
 		eseq.reserve(seq.size());
-	--level;
-	exvector::const_iterator it = seq.begin(), itend = seq.end();
-	while (it != itend) {
-		eseq.push_back(it->evalf(level));
-		++it;
+		--level;
+		exvector::const_iterator it = seq.begin(), itend = seq.end();
+		while (it != itend) {
+			eseq.push_back(it->evalf(level));
+			++it;
+		}
 	}
-	
-	if (registered_functions()[serial].evalf_f==0) {
+
+	if (opt.evalf_f==0) {
 		return function(serial,eseq).hold();
 	}
 	current_serial = serial;
-	if (registered_functions()[serial].evalf_use_exvector_args)
-		return ((evalf_funcp_exvector)(registered_functions()[serial].evalf_f))(seq);
-	switch (registered_functions()[serial].nparams) {
+	if (opt.evalf_use_exvector_args)
+		return ((evalf_funcp_exvector)(opt.evalf_f))(seq);
+	switch (opt.nparams) {
 		// the following lines have been generated for max. ${maxargs} parameters
 ${evalf_switch_statement}
 		// end of generated lines
@@ -799,14 +931,14 @@ ${evalf_switch_statement}
 	throw(std::logic_error("function::evalf(): invalid nparams"));
 }
 
-unsigned function::calchash(void) const
+unsigned function::calchash() const
 {
 	unsigned v = golden_ratio_hash(golden_ratio_hash(tinfo()) ^ serial);
-	for (unsigned i=0; i<nops(); i++) {
-		v = rotate_left_31(v);
+	for (size_t i=0; i<nops(); i++) {
+		v = rotate_left(v);
 		v ^= this->op(i).gethash();
 	}
-	v &= 0x7FFFFFFFU;
+
 	if (flags & status_flags::evaluated) {
 		setflag(status_flags::hash_calculated);
 		hashvalue = v;
@@ -814,14 +946,14 @@ unsigned function::calchash(void) const
 	return v;
 }
 
-ex function::thisexprseq(const exvector & v) const
+ex function::thiscontainer(const exvector & v) const
 {
-	return function(serial,v);
+	return function(serial, v);
 }
 
-ex function::thisexprseq(exvector * vp) const
+ex function::thiscontainer(std::auto_ptr<exvector> vp) const
 {
-	return function(serial,vp);
+	return function(serial, vp);
 }
 
 /** Implementation of ex::series for functions.
@@ -829,21 +961,22 @@ ex function::thisexprseq(exvector * vp) const
 ex function::series(const relational & r, int order, unsigned options) const
 {
 	GINAC_ASSERT(serial<registered_functions().size());
+	const function_options &opt = registered_functions()[serial];
 
-	if (registered_functions()[serial].series_f==0) {
+	if (opt.series_f==0) {
 		return basic::series(r, order);
 	}
 	ex res;
 	current_serial = serial;
-	if (registered_functions()[serial].series_use_exvector_args) {
+	if (opt.series_use_exvector_args) {
 		try {
-			res = ((series_funcp_exvector)(registered_functions()[serial].series_f))(seq, r, order, options);
+			res = ((series_funcp_exvector)(opt.series_f))(seq, r, order, options);
 		} catch (do_taylor) {
 			res = basic::series(r, order, options);
 		}
 		return res;
 	}
-	switch (registered_functions()[serial].nparams) {
+	switch (opt.nparams) {
 		// the following lines have been generated for max. ${maxargs} parameters
 ${series_switch_statement}
 		// end of generated lines
@@ -851,6 +984,28 @@ ${series_switch_statement}
 	throw(std::logic_error("function::series(): invalid nparams"));
 }
 
+/** Implementation of ex::conjugate for functions. */
+ex function::conjugate() const
+{
+	GINAC_ASSERT(serial<registered_functions().size());
+	const function_options & opt = registered_functions()[serial];
+
+	if (opt.conjugate_f==0) {
+		return exprseq::conjugate();
+	}
+
+	if (opt.conjugate_use_exvector_args) {
+		return ((conjugate_funcp_exvector)(opt.conjugate_f))(seq);
+	}
+
+	switch (opt.nparams) {
+		// the following lines have been generated for max. ${maxargs} parameters
+${conjugate_switch_statement}
+		// end of generated lines
+	}
+	throw(std::logic_error("function::conjugate(): invalid nparams"));
+}
+
 // protected
 
 /** Implementation of ex::diff() for functions. It applies the chain rule,
@@ -860,14 +1015,14 @@ ex function::derivative(const symbol & s) const
 {
 	ex result;
 
-	if (serial == function_index_Order) {
+	if (serial == Order_SERIAL::serial) {
 		// Order Term function only differentiates the argument
 		return Order(seq[0].diff(s));
 	} else {
 		// Chain rule
 		ex arg_diff;
-		unsigned num = seq.size();
-		for (unsigned i=0; i<num; i++) {
+		size_t num = seq.size();
+		for (size_t i=0; i<num; i++) {
 			arg_diff = seq[i].diff(s);
 			// We apply the chain rule only when it makes sense.  This is not
 			// just for performance reasons but also to allow functions to
@@ -883,7 +1038,7 @@ ex function::derivative(const symbol & s) const
 
 int function::compare_same_type(const basic & other) const
 {
-	GINAC_ASSERT(is_of_type(other, function));
+	GINAC_ASSERT(is_a<function>(other));
 	const function & o = static_cast<const function &>(other);
 
 	if (serial != o.serial)
@@ -894,7 +1049,7 @@ int function::compare_same_type(const basic & other) const
 
 bool function::is_equal_same_type(const basic & other) const
 {
-	GINAC_ASSERT(is_of_type(other, function));
+	GINAC_ASSERT(is_a<function>(other));
 	const function & o = static_cast<const function &>(other);
 
 	if (serial != o.serial)
@@ -905,26 +1060,46 @@ bool function::is_equal_same_type(const basic & other) const
 
 bool function::match_same_type(const basic & other) const
 {
-	GINAC_ASSERT(is_of_type(other, function));
+	GINAC_ASSERT(is_a<function>(other));
 	const function & o = static_cast<const function &>(other);
 
 	return serial == o.serial;
 }
 
-unsigned function::return_type(void) const
+unsigned function::return_type() const
 {
-	if (seq.empty())
-		return return_types::commutative;
-	else
-		return seq.begin()->return_type();
+	GINAC_ASSERT(serial<registered_functions().size());
+	const function_options &opt = registered_functions()[serial];
+
+	if (opt.use_return_type) {
+		// Return type was explicitly specified
+		return opt.return_type;
+	} else {
+		// Default behavior is to use the return type of the first
+		// argument. Thus, exp() of a matrix behaves like a matrix, etc.
+		if (seq.empty())
+			return return_types::commutative;
+		else
+			return seq.begin()->return_type();
+	}
 }
 
-unsigned function::return_type_tinfo(void) const
+unsigned function::return_type_tinfo() const
 {
-	if (seq.empty())
-		return tinfo_key;
-	else
-		return seq.begin()->return_type_tinfo();
+	GINAC_ASSERT(serial<registered_functions().size());
+	const function_options &opt = registered_functions()[serial];
+
+	if (opt.use_return_type) {
+		// Return type was explicitly specified
+		return opt.return_type_tinfo;
+	} else {
+		// Default behavior is to use the return type of the first
+		// argument. Thus, exp() of a matrix behaves like a matrix, etc.
+		if (seq.empty())
+			return tinfo_key;
+		else
+			return seq.begin()->return_type_tinfo();
+	}
 }
 
 //////////
@@ -942,15 +1117,16 @@ unsigned function::return_type_tinfo(void) const
 ex function::pderivative(unsigned diff_param) const // partial differentiation
 {
 	GINAC_ASSERT(serial<registered_functions().size());
+	const function_options &opt = registered_functions()[serial];
 	
 	// No derivative defined? Then return abstract derivative object
-	if (registered_functions()[serial].derivative_f == NULL)
+	if (opt.derivative_f == NULL)
 		return fderivative(serial, diff_param, seq);
 
 	current_serial = serial;
-	if (registered_functions()[serial].derivative_use_exvector_args)
-		return ((derivative_funcp_exvector)(registered_functions()[serial].derivative_f))(seq, diff_param);
-	switch (registered_functions()[serial].nparams) {
+	if (opt.derivative_use_exvector_args)
+		return ((derivative_funcp_exvector)(opt.derivative_f))(seq, diff_param);
+	switch (opt.nparams) {
 		// the following lines have been generated for max. ${maxargs} parameters
 ${diff_switch_statement}
 		// end of generated lines
@@ -958,7 +1134,7 @@ ${diff_switch_statement}
 	throw(std::logic_error("function::pderivative(): no diff function defined"));
 }
 
-std::vector<function_options> & function::registered_functions(void)
+std::vector<function_options> & function::registered_functions()
 {
 	static std::vector<function_options> * rf = new std::vector<function_options>;
 	return *rf;
@@ -966,20 +1142,20 @@ std::vector<function_options> & function::registered_functions(void)
 
 bool function::lookup_remember_table(ex & result) const
 {
-	return remember_table::remember_tables()[serial].lookup_entry(*this,result);
+	return remember_table::remember_tables()[this->serial].lookup_entry(*this,result);
 }
 
 void function::store_remember_table(ex const & result) const
 {
-	remember_table::remember_tables()[serial].add_entry(*this,result);
+	remember_table::remember_tables()[this->serial].add_entry(*this,result);
 }
 
 // public
 
 unsigned function::register_new(function_options const & opt)
 {
-	unsigned same_name = 0;
-	for (unsigned i=0; i<registered_functions().size(); ++i) {
+	size_t same_name = 0;
+	for (size_t i=0; i<registered_functions().size(); ++i) {
 		if (registered_functions()[i].name==opt.name) {
 			++same_name;
 		}
@@ -1019,7 +1195,7 @@ unsigned function::find_function(const std::string &name, unsigned nparams)
 }
 
 /** Return the print name of the function. */
-std::string function::get_name(void) const
+std::string function::get_name() const
 {
 	GINAC_ASSERT(serial<registered_functions().size());
 	return registered_functions()[serial].name;
