@@ -3,7 +3,7 @@
  *  This include file includes all other public GiNaC headers. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -57,14 +57,13 @@
 #include "inifcns.h"
 #include "fderivative.h"
 #include "operators.h"
+#include "hash_map.h"
 
-#ifndef GINAC_BASE_ONLY
 #include "idx.h"
 #include "indexed.h"
 #include "tensor.h"
 #include "color.h"
 #include "clifford.h"
-#endif // ndef GINAC_BASE_ONLY
 
 #ifdef __MAKECINT__
 #pragma link C++ nestedclass;
