@@ -3,7 +3,7 @@
  *  Implementation of GiNaC's indices. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,23 +21,36 @@
  */
 
 #include <iostream>
+#include <sstream>
 #include <stdexcept>
 
 #include "idx.h"
 #include "symbol.h"
 #include "lst.h"
-#include "print.h"
+#include "relational.h"
+#include "operators.h"
 #include "archive.h"
 #include "utils.h"
 
 namespace GiNaC {
 
-GINAC_IMPLEMENT_REGISTERED_CLASS(idx, basic)
-GINAC_IMPLEMENT_REGISTERED_CLASS(varidx, idx)
-GINAC_IMPLEMENT_REGISTERED_CLASS(spinidx, varidx)
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(idx, basic,
+  print_func<print_context>(&idx::do_print).
+  print_func<print_latex>(&idx::do_print_latex).
+  print_func<print_tree>(&idx::do_print_tree))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(varidx, idx,
+  print_func<print_context>(&varidx::do_print).
+  // print_latex inherited from idx
+  print_func<print_tree>(&varidx::do_print_tree))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(spinidx, varidx,
+  print_func<print_context>(&spinidx::do_print).
+  print_func<print_latex>(&spinidx::do_print_latex).
+  print_func<print_tree>(&spinidx::do_print_tree))
 
 //////////
-// default ctor, dtor, copy ctor, assignment operator and helpers
+// default constructor
 //////////
 
 idx::idx() : inherited(TINFO_idx) {}
@@ -52,29 +65,6 @@ spinidx::spinidx() : dotted(false)
 	tinfo_key = TINFO_spinidx;
 }
 
-void idx::copy(const idx & other)
-{
-	inherited::copy(other);
-	value = other.value;
-	dim = other.dim;
-}
-
-void varidx::copy(const varidx & other)
-{
-	inherited::copy(other);
-	covariant = other.covariant;
-}
-
-void spinidx::copy(const spinidx & other)
-{
-	inherited::copy(other);
-	dotted = other.dotted;
-}
-
-DEFAULT_DESTROY(idx)
-DEFAULT_DESTROY(varidx)
-DEFAULT_DESTROY(spinidx)
-
 //////////
 // other constructors
 //////////
@@ -100,18 +90,18 @@ spinidx::spinidx(const ex & v, const ex & d, bool cov, bool dot) : inherited(v,
 // archiving
 //////////
 
-idx::idx(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+idx::idx(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	n.find_ex("value", value, sym_lst);
 	n.find_ex("dim", dim, sym_lst);
 }
 
-varidx::varidx(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+varidx::varidx(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	n.find_bool("covariant", covariant);
 }
 
-spinidx::spinidx(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+spinidx::spinidx(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	n.find_bool("dotted", dotted);
 }
@@ -143,110 +133,92 @@ DEFAULT_UNARCHIVE(spinidx)
 // functions overriding virtual functions from base classes
 //////////
 
-void idx::print(const print_context & c, unsigned level) const
+void idx::print_index(const print_context & c, unsigned level) const
 {
-	if (is_of_type(c, print_tree)) {
-
-		c.s << std::string(level, ' ') << class_name()
-		    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
-		    << std::endl;
-		unsigned delta_indent = static_cast<const print_tree &>(c).delta_indent;
-		value.print(c, level + delta_indent);
-		dim.print(c, level + delta_indent);
-
-	} else {
-
-		if (is_a<print_latex>(c))
-			c.s << "{";
-		else
-			c.s << ".";
-		bool need_parens = !(is_ex_exactly_of_type(value, numeric) || is_ex_of_type(value, symbol));
-		if (need_parens)
-			c.s << "(";
-		value.print(c);
-		if (need_parens)
-			c.s << ")";
-		if (is_a<print_latex>(c))
-			c.s << "}";
+	bool need_parens = !(is_exactly_a<numeric>(value) || is_a<symbol>(value));
+	if (need_parens)
+		c.s << "(";
+	value.print(c);
+	if (need_parens)
+		c.s << ")";
+	if (c.options & print_options::print_index_dimensions) {
+		c.s << "[";
+		dim.print(c);
+		c.s << "]";
 	}
 }
 
-void varidx::print(const print_context & c, unsigned level) const
+void idx::do_print(const print_context & c, unsigned level) const
 {
-	if (is_of_type(c, print_tree)) {
-
-		c.s << std::string(level, ' ') << class_name()
-		    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
-		    << (covariant ? ", covariant" : ", contravariant")
-		    << std::endl;
-		unsigned delta_indent = static_cast<const print_tree &>(c).delta_indent;
-		value.print(c, level + delta_indent);
-		dim.print(c, level + delta_indent);
-
-	} else {
-		if (is_a<print_latex>(c))
-			c.s << "{";
-		else {
-			if (covariant)
-				c.s << ".";
-			else
-				c.s << "~";
-		}
-		bool need_parens = !(is_ex_exactly_of_type(value, numeric) || is_ex_of_type(value, symbol));
-		if (need_parens)
-			c.s << "(";
-		value.print(c);
-		if (need_parens)
-			c.s << ")";
-		if (is_a<print_latex>(c))
-			c.s << "}";
-	}
+	c.s << ".";
+	print_index(c, level);
 }
 
-void spinidx::print(const print_context & c, unsigned level) const
+void idx::do_print_latex(const print_latex & c, unsigned level) const
 {
-	if (is_of_type(c, print_tree)) {
+	c.s << "{";
+	print_index(c, level);
+	c.s << "}";
+}
 
-		c.s << std::string(level, ' ') << class_name()
-		    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
-		    << (covariant ? ", covariant" : ", contravariant")
-		    << (dotted ? ", dotted" : ", undotted")
-		    << std::endl;
-		unsigned delta_indent = static_cast<const print_tree &>(c).delta_indent;
-		value.print(c, level + delta_indent);
-		dim.print(c, level + delta_indent);
+void idx::do_print_tree(const print_tree & c, unsigned level) const
+{
+	c.s << std::string(level, ' ') << class_name() << " @" << this
+	    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
+	    << std::endl;
+	value.print(c, level +  c.delta_indent);
+	dim.print(c, level + c.delta_indent);
+}
 
-	} else {
+void varidx::do_print(const print_context & c, unsigned level) const
+{
+	if (covariant)
+		c.s << ".";
+	else
+		c.s << "~";
+	print_index(c, level);
+}
 
-		bool is_tex = is_of_type(c, print_latex);
-		if (is_tex) {
-			if (covariant)
-				c.s << "_{";
-			else
-				c.s << "^{";
-		} else {
-			if (covariant)
-				c.s << ".";
-			else
-				c.s << "~";
-		}
-		if (dotted) {
-			if (is_tex)
-				c.s << "\\dot{";
-			else
-				c.s << "*";
-		}
-		bool need_parens = !(is_ex_exactly_of_type(value, numeric) || is_ex_of_type(value, symbol));
-		if (need_parens)
-			c.s << "(";
-		value.print(c);
-		if (need_parens)
-			c.s << ")";
-		if (is_tex && dotted)
-			c.s << "}";
-		if (is_tex)
-			c.s << "}";
-	}
+void varidx::do_print_tree(const print_tree & c, unsigned level) const
+{
+	c.s << std::string(level, ' ') << class_name() << " @" << this
+	    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
+	    << (covariant ? ", covariant" : ", contravariant")
+	    << std::endl;
+	value.print(c, level + c.delta_indent);
+	dim.print(c, level + c.delta_indent);
+}
+
+void spinidx::do_print(const print_context & c, unsigned level) const
+{
+	if (covariant)
+		c.s << ".";
+	else
+		c.s << "~";
+	if (dotted)
+		c.s << "*";
+	print_index(c, level);
+}
+
+void spinidx::do_print_latex(const print_latex & c, unsigned level) const
+{
+	if (dotted)
+		c.s << "\\dot{";
+	else
+		c.s << "{";
+	print_index(c, level);
+	c.s << "}";
+}
+
+void spinidx::do_print_tree(const print_tree & c, unsigned level) const
+{
+	c.s << std::string(level, ' ') << class_name() << " @" << this
+	    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
+	    << (covariant ? ", covariant" : ", contravariant")
+	    << (dotted ? ", dotted" : ", undotted")
+	    << std::endl;
+	value.print(c, level + c.delta_indent);
+	dim.print(c, level + c.delta_indent);
 }
 
 bool idx::info(unsigned inf) const
@@ -256,18 +228,27 @@ bool idx::info(unsigned inf) const
 	return inherited::info(inf);
 }
 
-unsigned idx::nops() const
+size_t idx::nops() const
 {
 	// don't count the dimension as that is not really a sub-expression
 	return 1;
 }
 
-ex & idx::let_op(int i)
+ex idx::op(size_t i) const
 {
 	GINAC_ASSERT(i == 0);
 	return value;
 }
 
+ex idx::map(map_function & f) const
+{
+	idx *copy = duplicate();
+	copy->setflag(status_flags::dynallocated);
+	copy->clearflag(status_flags::hash_calculated);
+	copy->value = f(value);
+	return *copy;
+}
+
 /** Returns order relation between two indices of the same type. The order
  *  must be such that dummy indices lie next to each other. */
 int idx::compare_same_type(const basic & other) const
@@ -301,6 +282,7 @@ int varidx::compare_same_type(const basic & other) const
 	// Check variance last so dummy indices will end up next to each other
 	if (covariant != o.covariant)
 		return covariant ? -1 : 1;
+
 	return 0;
 }
 
@@ -311,6 +293,7 @@ bool varidx::match_same_type(const basic & other) const
 
 	if (covariant != o.covariant)
 		return false;
+
 	return inherited::match_same_type(other);
 }
 
@@ -340,6 +323,29 @@ bool spinidx::match_same_type(const basic & other) const
 	return inherited::match_same_type(other);
 }
 
+unsigned idx::calchash() const
+{
+	// NOTE: The code in simplify_indexed() assumes that canonically
+	// ordered sequences of indices have the two members of dummy index
+	// pairs lying next to each other. The hash values for indices must
+	// be devised accordingly. The easiest (only?) way to guarantee the
+	// desired ordering is to make indices with the same value have equal
+	// hash keys. That is, the hash values must not depend on the index
+	// dimensions or other attributes (variance etc.).
+	// The compare_same_type() methods will take care of the rest.
+	unsigned v = golden_ratio_hash(tinfo());
+	v = rotate_left(v);
+	v ^= value.gethash();
+
+	// Store calculated hash value only if object is already evaluated
+	if (flags & status_flags::evaluated) {
+		setflag(status_flags::hash_calculated);
+		hashvalue = v;
+	}
+
+	return v;
+}
+
 /** By default, basic::evalf would evaluate the index value but we don't want
  *  a.1 to become a.(1.0). */
 ex idx::evalf(int level) const
@@ -347,32 +353,29 @@ ex idx::evalf(int level) const
 	return *this;
 }
 
-ex idx::subs(const lst & ls, const lst & lr, bool no_pattern) const
+ex idx::subs(const exmap & m, unsigned options) const
 {
-	GINAC_ASSERT(ls.nops() == lr.nops());
-
 	// First look for index substitutions
-	for (unsigned i=0; i<ls.nops(); i++) {
-		if (is_equal(ex_to<basic>(ls.op(i)))) {
+	exmap::const_iterator it = m.find(*this);
+	if (it != m.end()) {
 
-			// Substitution index->index
-			if (is_ex_of_type(lr.op(i), idx))
-				return lr.op(i);
+		// Substitution index->index
+		if (is_a<idx>(it->second))
+			return it->second;
 
-			// Otherwise substitute value
-			idx *i_copy = static_cast<idx *>(duplicate());
-			i_copy->value = lr.op(i);
-			i_copy->clearflag(status_flags::hash_calculated);
-			return i_copy->setflag(status_flags::dynallocated);
-		}
+		// Otherwise substitute value
+		idx *i_copy = duplicate();
+		i_copy->value = it->second;
+		i_copy->clearflag(status_flags::hash_calculated);
+		return i_copy->setflag(status_flags::dynallocated);
 	}
 
 	// None, substitute objects in value (not in dimension)
-	const ex &subsed_value = value.subs(ls, lr, no_pattern);
+	const ex &subsed_value = value.subs(m, options);
 	if (are_ex_trivially_equal(value, subsed_value))
 		return *this;
 
-	idx *i_copy = static_cast<idx *>(duplicate());
+	idx *i_copy = duplicate();
 	i_copy->value = subsed_value;
 	i_copy->clearflag(status_flags::hash_calculated);
 	return i_copy->setflag(status_flags::dynallocated);
@@ -395,15 +398,19 @@ bool idx::is_dummy_pair_same_type(const basic & other) const
 	const idx &o = static_cast<const idx &>(other);
 
 	// Only pure symbols form dummy pairs, "2n+1" doesn't
-	if (!is_ex_of_type(value, symbol))
+	if (!is_a<symbol>(value))
 		return false;
 
 	// Value must be equal, of course
 	if (!value.is_equal(o.value))
 		return false;
 
-	// Also the dimension
-	return dim.is_equal(o.dim);
+	// Dimensions need not be equal but must be comparable (so we can
+	// determine the minimum dimension of contractions)
+	if (dim.is_equal(o.dim))
+		return true;
+
+	return (dim < o.dim || dim > o.dim || (is_exactly_a<numeric>(dim) && is_a<symbol>(o.dim)) || (is_a<symbol>(dim) && is_exactly_a<numeric>(o.dim)));
 }
 
 bool varidx::is_dummy_pair_same_type(const basic & other) const
@@ -433,25 +440,38 @@ bool spinidx::is_dummy_pair_same_type(const basic & other) const
 // non-virtual functions
 //////////
 
-ex varidx::toggle_variance(void) const
+ex idx::replace_dim(const ex & new_dim) const
 {
-	varidx *i_copy = static_cast<varidx *>(duplicate());
+	idx *i_copy = duplicate();
+	i_copy->dim = new_dim;
+	i_copy->clearflag(status_flags::hash_calculated);
+	return i_copy->setflag(status_flags::dynallocated);
+}
+
+ex idx::minimal_dim(const idx & other) const
+{
+	return GiNaC::minimal_dim(dim, other.dim);
+}
+
+ex varidx::toggle_variance() const
+{
+	varidx *i_copy = duplicate();
 	i_copy->covariant = !i_copy->covariant;
 	i_copy->clearflag(status_flags::hash_calculated);
 	return i_copy->setflag(status_flags::dynallocated);
 }
 
-ex spinidx::toggle_dot(void) const
+ex spinidx::toggle_dot() const
 {
-	spinidx *i_copy = static_cast<spinidx *>(duplicate());
+	spinidx *i_copy = duplicate();
 	i_copy->dotted = !i_copy->dotted;
 	i_copy->clearflag(status_flags::hash_calculated);
 	return i_copy->setflag(status_flags::dynallocated);
 }
 
-ex spinidx::toggle_variance_dot(void) const
+ex spinidx::toggle_variance_dot() const
 {
-	spinidx *i_copy = static_cast<spinidx *>(duplicate());
+	spinidx *i_copy = duplicate();
 	i_copy->covariant = !i_copy->covariant;
 	i_copy->dotted = !i_copy->dotted;
 	i_copy->clearflag(status_flags::hash_calculated);
@@ -475,7 +495,7 @@ bool is_dummy_pair(const idx & i1, const idx & i2)
 bool is_dummy_pair(const ex & e1, const ex & e2)
 {
 	// The expressions must be indices
-	if (!is_ex_of_type(e1, idx) || !is_ex_of_type(e2, idx))
+	if (!is_a<idx>(e1) || !is_a<idx>(e2))
 		return false;
 
 	return is_dummy_pair(ex_to<idx>(e1), ex_to<idx>(e2));
@@ -521,4 +541,17 @@ void find_free_and_dummy(exvector::const_iterator it, exvector::const_iterator i
 		out_free.push_back(*last);
 }
 
+ex minimal_dim(const ex & dim1, const ex & dim2)
+{
+	if (dim1.is_equal(dim2) || dim1 < dim2 || (is_exactly_a<numeric>(dim1) && is_a<symbol>(dim2)))
+		return dim1;
+	else if (dim1 > dim2 || (is_a<symbol>(dim1) && is_exactly_a<numeric>(dim2)))
+		return dim2;
+	else {
+		std::ostringstream s;
+		s << "minimal_dim(): index dimensions " << dim1 << " and " << dim2 << " cannot be ordered";
+		throw (std::runtime_error(s.str()));
+	}
+}
+
 } // namespace GiNaC
