@@ -3,7 +3,7 @@
  *  Implementation of symbolic matrices */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -20,7 +20,9 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#include <string>
 #include <iostream>
+#include <sstream>
 #include <algorithm>
 #include <map>
 #include <stdexcept>
@@ -30,39 +32,34 @@
 #include "lst.h"
 #include "idx.h"
 #include "indexed.h"
+#include "add.h"
 #include "power.h"
 #include "symbol.h"
+#include "operators.h"
 #include "normal.h"
-#include "print.h"
 #include "archive.h"
 #include "utils.h"
 
 namespace GiNaC {
 
-GINAC_IMPLEMENT_REGISTERED_CLASS(matrix, basic)
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(matrix, basic,
+  print_func<print_context>(&matrix::do_print).
+  print_func<print_latex>(&matrix::do_print_latex).
+  print_func<print_tree>(&matrix::do_print_tree).
+  print_func<print_python_repr>(&matrix::do_print_python_repr))
 
 //////////
-// default ctor, dtor, copy ctor, assignment operator and helpers:
+// default constructor
 //////////
 
 /** Default ctor.  Initializes to 1 x 1-dimensional zero-matrix. */
-matrix::matrix() : inherited(TINFO_matrix), row(1), col(1)
+matrix::matrix() : inherited(TINFO_matrix), row(1), col(1), m(1, _ex0)
 {
-	m.push_back(_ex0);
+	setflag(status_flags::not_shareable);
 }
 
-void matrix::copy(const matrix & other)
-{
-	inherited::copy(other);
-	row = other.row;
-	col = other.col;
-	m = other.m;  // STL's vector copying invoked here
-}
-
-DEFAULT_DESTROY(matrix)
-
 //////////
-// other ctors
+// other constructors
 //////////
 
 // public
@@ -72,32 +69,36 @@ DEFAULT_DESTROY(matrix)
  *  @param r number of rows
  *  @param c number of cols */
 matrix::matrix(unsigned r, unsigned c)
-  : inherited(TINFO_matrix), row(r), col(c)
+  : inherited(TINFO_matrix), row(r), col(c), m(r*c, _ex0)
 {
-	m.resize(r*c, _ex0);
+	setflag(status_flags::not_shareable);
 }
 
 // protected
 
 /** Ctor from representation, for internal use only. */
 matrix::matrix(unsigned r, unsigned c, const exvector & m2)
-  : inherited(TINFO_matrix), row(r), col(c), m(m2) {}
+  : inherited(TINFO_matrix), row(r), col(c), m(m2)
+{
+	setflag(status_flags::not_shareable);
+}
 
 /** Construct matrix from (flat) list of elements. If the list has fewer
  *  elements than the matrix, the remaining matrix elements are set to zero.
  *  If the list has more elements than the matrix, the excessive elements are
  *  thrown away. */
 matrix::matrix(unsigned r, unsigned c, const lst & l)
-  : inherited(TINFO_matrix), row(r), col(c)
+  : inherited(TINFO_matrix), row(r), col(c), m(r*c, _ex0)
 {
-	m.resize(r*c, _ex0);
+	setflag(status_flags::not_shareable);
 
-	for (unsigned i=0; i<l.nops(); i++) {
-		unsigned x = i % c;
-		unsigned y = i / c;
+	size_t i = 0;
+	for (lst::const_iterator it = l.begin(); it != l.end(); ++it, ++i) {
+		size_t x = i % c;
+		size_t y = i / c;
 		if (y >= r)
 			break; // matrix smaller than list: throw away excessive elements
-		m[y*c+x] = l.op(i);
+		m[y*c+x] = *it;
 	}
 }
 
@@ -105,8 +106,10 @@ matrix::matrix(unsigned r, unsigned c, const lst & l)
 // archiving
 //////////
 
-matrix::matrix(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+matrix::matrix(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
+	setflag(status_flags::not_shareable);
+
 	if (!(n.find_unsigned("row", row)) || !(n.find_unsigned("col", col)))
 		throw (std::runtime_error("unknown matrix dimensions in archive"));
 	m.reserve(row * col);
@@ -139,74 +142,63 @@ DEFAULT_UNARCHIVE(matrix)
 
 // public
 
-void matrix::print(const print_context & c, unsigned level) const
+void matrix::print_elements(const print_context & c, const char *row_start, const char *row_end, const char *row_sep, const char *col_sep) const
 {
-	if (is_a<print_tree>(c)) {
-
-		inherited::print(c, level);
-
-	} else {
-
-		if (is_a<print_python_repr>(c))
-			c.s << class_name() << '(';
-
-		if (is_a<print_latex>(c))
-			c.s << "\\left(\\begin{array}{" << std::string(col,'c') << "}";
-		else
-			c.s << "[";
-
-		for (unsigned ro=0; ro<row; ++ro) {
-			if (!is_a<print_latex>(c))
-				c.s << "[";
-			for (unsigned co=0; co<col; ++co) {
-				m[ro*col+co].print(c);
-				if (co<col-1) {
-					if (is_a<print_latex>(c))
-						c.s << "&";
-					else
-						c.s << ",";
-				} else {
-					if (!is_a<print_latex>(c))
-						c.s << "]";
-				}
-			}
-			if (ro<row-1) {
-				if (is_a<print_latex>(c))
-					c.s << "\\\\";
-				else
-					c.s << ",";
-			}
+	for (unsigned ro=0; ro<row; ++ro) {
+		c.s << row_start;
+		for (unsigned co=0; co<col; ++co) {
+			m[ro*col+co].print(c);
+			if (co < col-1)
+				c.s << col_sep;
+			else
+				c.s << row_end;
 		}
-
-		if (is_a<print_latex>(c))
-			c.s << "\\end{array}\\right)";
-		else
-			c.s << "]";
-
-		if (is_a<print_python_repr>(c))
-			c.s << ')';
-
+		if (ro < row-1)
+			c.s << row_sep;
 	}
 }
 
-/** nops is defined to be rows x columns. */
-unsigned matrix::nops() const
+void matrix::do_print(const print_context & c, unsigned level) const
 {
-	return row*col;
+	c.s << "[";
+	print_elements(c, "[", "]", ",", ",");
+	c.s << "]";
+}
+
+void matrix::do_print_latex(const print_latex & c, unsigned level) const
+{
+	c.s << "\\left(\\begin{array}{" << std::string(col,'c') << "}";
+	print_elements(c, "", "", "\\\\", "&");
+	c.s << "\\end{array}\\right)";
+}
+
+void matrix::do_print_python_repr(const print_python_repr & c, unsigned level) const
+{
+	c.s << class_name() << '(';
+	print_elements(c, "[", "]", ",", ",");
+	c.s << ')';
+}
+
+/** nops is defined to be rows x columns. */
+size_t matrix::nops() const
+{
+	return static_cast<size_t>(row) * static_cast<size_t>(col);
 }
 
 /** returns matrix entry at position (i/col, i%col). */
-ex matrix::op(int i) const
+ex matrix::op(size_t i) const
 {
+	GINAC_ASSERT(i<nops());
+	
 	return m[i];
 }
 
-/** returns matrix entry at position (i/col, i%col). */
-ex & matrix::let_op(int i)
+/** returns writable matrix entry at position (i/col, i%col). */
+ex & matrix::let_op(size_t i)
 {
-	GINAC_ASSERT(i>=0);
 	GINAC_ASSERT(i<nops());
 	
+	ensure_if_modifiable();
 	return m[i];
 }
 
@@ -229,17 +221,45 @@ ex matrix::eval(int level) const
 			m2[r*col+c] = m[r*col+c].eval(level);
 	
 	return (new matrix(row, col, m2))->setflag(status_flags::dynallocated |
-											   status_flags::evaluated );
+											   status_flags::evaluated);
 }
 
-ex matrix::subs(const lst & ls, const lst & lr, bool no_pattern) const
+ex matrix::subs(const exmap & mp, unsigned options) const
 {
 	exvector m2(row * col);
 	for (unsigned r=0; r<row; ++r)
 		for (unsigned c=0; c<col; ++c)
-			m2[r*col+c] = m[r*col+c].subs(ls, lr, no_pattern);
+			m2[r*col+c] = m[r*col+c].subs(mp, options);
 
-	return matrix(row, col, m2).basic::subs(ls, lr, no_pattern);
+	return matrix(row, col, m2).subs_one_level(mp, options);
+}
+
+/** Complex conjugate every matrix entry. */
+ex matrix::conjugate() const
+{
+	exvector * ev = 0;
+	for (exvector::const_iterator i=m.begin(); i!=m.end(); ++i) {
+		ex x = i->conjugate();
+		if (ev) {
+			ev->push_back(x);
+			continue;
+		}
+		if (are_ex_trivially_equal(x, *i)) {
+			continue;
+		}
+		ev = new exvector;
+		ev->reserve(m.size());
+		for (exvector::const_iterator j=m.begin(); j!=i; ++j) {
+			ev->push_back(*j);
+		}
+		ev->push_back(x);
+	}
+	if (ev) {
+		ex result = matrix(row, col, *ev);
+		delete ev;
+		return result;
+	}
+	return *this;
 }
 
 // protected
@@ -365,7 +385,7 @@ ex matrix::add_indexed(const ex & self, const ex & other) const
 	GINAC_ASSERT(self.nops() == 2 || self.nops() == 3);
 
 	// Only add two matrices
-	if (is_ex_of_type(other.op(0), matrix)) {
+	if (is_a<matrix>(other.op(0))) {
 		GINAC_ASSERT(other.nops() == 2 || other.nops() == 3);
 
 		const matrix &self_matrix = ex_to<matrix>(self.op(0));
@@ -416,7 +436,7 @@ bool matrix::contract_with(exvector::iterator self, exvector::iterator other, ex
 	GINAC_ASSERT(is_a<matrix>(self->op(0)));
 
 	// Only contract with other matrices
-	if (!is_ex_of_type(other->op(0), matrix))
+	if (!is_a<matrix>(other->op(0)))
 		return false;
 
 	GINAC_ASSERT(other->nops() == 2 || other->nops() == 3);
@@ -605,7 +625,7 @@ matrix matrix::pow(const ex & expn) const
 	if (col!=row)
 		throw (std::logic_error("matrix::pow(): matrix not square"));
 	
-	if (is_ex_exactly_of_type(expn, numeric)) {
+	if (is_exactly_a<numeric>(expn)) {
 		// Integer cases are computed by successive multiplication, using the
 		// obvious shortcut of storing temporaries, like A^4 == (A*A)*(A*A).
 		if (expn.info(info_flags::integer)) {
@@ -673,7 +693,7 @@ ex & matrix::operator() (unsigned ro, unsigned co)
 
 /** Transposed of an m x n matrix, producing a new n x m matrix object that
  *  represents the transposed. */
-matrix matrix::transpose(void) const
+matrix matrix::transpose() const
 {
 	exvector trans(this->cols()*this->rows());
 	
@@ -710,7 +730,7 @@ ex matrix::determinant(unsigned algo) const
 	unsigned sparse_count = 0;  // counts non-zero elements
 	exvector::const_iterator r = m.begin(), rend = m.end();
 	while (r != rend) {
-		lst srl;  // symbol replacement list
+		exmap srl;  // symbol replacement list
 		ex rtest = r->to_rational(srl);
 		if (!rtest.is_zero())
 			++sparse_count;
@@ -830,7 +850,7 @@ ex matrix::determinant(unsigned algo) const
  *
  *  @return    the sum of diagonal elements
  *  @exception logic_error (matrix not square) */
-ex matrix::trace(void) const
+ex matrix::trace() const
 {
 	if (row != col)
 		throw (std::logic_error("matrix::trace(): matrix not square"));
@@ -858,7 +878,7 @@ ex matrix::trace(void) const
  *  @return    characteristic polynomial as new expression
  *  @exception logic_error (matrix not square)
  *  @see       matrix::determinant() */
-ex matrix::charpoly(const symbol & lambda) const
+ex matrix::charpoly(const ex & lambda) const
 {
 	if (row != col)
 		throw (std::logic_error("matrix::charpoly(): matrix not square"));
@@ -875,27 +895,30 @@ ex matrix::charpoly(const symbol & lambda) const
 	// trapped and we use Leverrier's algorithm which goes as row^3 for
 	// every coefficient.  The expensive part is the matrix multiplication.
 	if (numeric_flag) {
+
 		matrix B(*this);
 		ex c = B.trace();
-		ex poly = power(lambda,row)-c*power(lambda,row-1);
+		ex poly = power(lambda, row) - c*power(lambda, row-1);
 		for (unsigned i=1; i<row; ++i) {
 			for (unsigned j=0; j<row; ++j)
 				B.m[j*col+j] -= c;
 			B = this->mul(B);
-			c = B.trace()/ex(i+1);
-			poly -= c*power(lambda,row-i-1);
+			c = B.trace() / ex(i+1);
+			poly -= c*power(lambda, row-i-1);
 		}
 		if (row%2)
 			return -poly;
 		else
 			return poly;
+
+	} else {
+	
+		matrix M(*this);
+		for (unsigned r=0; r<col; ++r)
+			M.m[r*col+r] -= lambda;
+	
+		return M.determinant().collect(lambda);
 	}
-	
-	matrix M(*this);
-	for (unsigned r=0; r<col; ++r)
-		M.m[r*col+r] -= lambda;
-	
-	return M.determinant().collect(lambda);
 }
 
 
@@ -904,7 +927,7 @@ ex matrix::charpoly(const symbol & lambda) const
  *  @return    the inverted matrix
  *  @exception logic_error (matrix not square)
  *  @exception runtime_error (singular matrix) */
-matrix matrix::inverse(void) const
+matrix matrix::inverse() const
 {
 	if (row != col)
 		throw (std::logic_error("matrix::inverse(): matrix not square"));
@@ -943,6 +966,7 @@ matrix matrix::inverse(void) const
  *
  *  @param vars n x p matrix, all elements must be symbols 
  *  @param rhs m x p matrix
+ *  @param algo selects the solving algorithm
  *  @return n x p solution matrix
  *  @exception logic_error (incompatible matrices)
  *  @exception invalid_argument (1st argument must be matrix of symbols)
@@ -1055,7 +1079,7 @@ matrix matrix::solve(const matrix & vars,
  *
  *  @return the determinant as a new expression (in expanded form)
  *  @see matrix::determinant() */
-ex matrix::determinant_minor(void) const
+ex matrix::determinant_minor() const
 {
 	// for small matrices the algorithm does not make any sense:
 	const unsigned n = this->cols();
@@ -1288,7 +1312,7 @@ int matrix::fraction_free_elimination(const bool det)
 	//
 	// Bareiss (fraction-free) elimination in addition divides that element
 	// by m[k-1](k-1,k-1) for k>1, where it can be shown by means of the
-	// Sylvester determinant that this really divides m[k+1](r,c).
+	// Sylvester identity that this really divides m[k+1](r,c).
 	//
 	// We also allow rational functions where the original prove still holds.
 	// However, we must care for numerator and denominator separately and
@@ -1325,7 +1349,7 @@ int matrix::fraction_free_elimination(const bool det)
 	// makes things more complicated than they need to be.
 	matrix tmp_n(*this);
 	matrix tmp_d(m,n);  // for denominators, if needed
-	lst srl;  // symbol replacement list
+	exmap srl;  // symbol replacement list
 	exvector::const_iterator cit = this->m.begin(), citend = this->m.end();
 	exvector::iterator tmp_n_it = tmp_n.m.begin(), tmp_d_it = tmp_d.m.begin();
 	while (cit != citend) {
@@ -1388,7 +1412,7 @@ int matrix::fraction_free_elimination(const bool det)
 	tmp_n_it = tmp_n.m.begin();
 	tmp_d_it = tmp_d.m.begin();
 	while (it != itend)
-		*it++ = ((*tmp_n_it++)/(*tmp_d_it++)).subs(srl);
+		*it++ = ((*tmp_n_it++)/(*tmp_d_it++)).subs(srl, subs_options::no_pattern);
 	
 	return sign;
 }
@@ -1416,11 +1440,11 @@ int matrix::pivot(unsigned ro, unsigned co, bool symbolic)
 			++k;
 	} else {
 		// search largest element in column co beginning at row ro
-		GINAC_ASSERT(is_a<numeric>(this->m[k*col+co]));
+		GINAC_ASSERT(is_exactly_a<numeric>(this->m[k*col+co]));
 		unsigned kmax = k+1;
 		numeric mmax = abs(ex_to<numeric>(m[kmax*col+co]));
 		while (kmax<row) {
-			GINAC_ASSERT(is_a<numeric>(this->m[kmax*col+co]));
+			GINAC_ASSERT(is_exactly_a<numeric>(this->m[kmax*col+co]));
 			numeric tmp = ex_to<numeric>(this->m[kmax*col+co]);
 			if (abs(tmp) > mmax) {
 				mmax = tmp;
@@ -1447,42 +1471,92 @@ int matrix::pivot(unsigned ro, unsigned co, bool symbolic)
 
 ex lst_to_matrix(const lst & l)
 {
+	lst::const_iterator itr, itc;
+
 	// Find number of rows and columns
-	unsigned rows = l.nops(), cols = 0, i, j;
-	for (i=0; i<rows; i++)
-		if (l.op(i).nops() > cols)
-			cols = l.op(i).nops();
+	size_t rows = l.nops(), cols = 0;
+	for (itr = l.begin(); itr != l.end(); ++itr) {
+		if (!is_a<lst>(*itr))
+			throw (std::invalid_argument("lst_to_matrix: argument must be a list of lists"));
+		if (itr->nops() > cols)
+			cols = itr->nops();
+	}
 
 	// Allocate and fill matrix
-	matrix &m = *new matrix(rows, cols);
-	m.setflag(status_flags::dynallocated);
-	for (i=0; i<rows; i++)
-		for (j=0; j<cols; j++)
-			if (l.op(i).nops() > j)
-				m(i, j) = l.op(i).op(j);
-			else
-				m(i, j) = _ex0;
-	return m;
+	matrix &M = *new matrix(rows, cols);
+	M.setflag(status_flags::dynallocated);
+
+	unsigned i;
+	for (itr = l.begin(), i = 0; itr != l.end(); ++itr, ++i) {
+		unsigned j;
+		for (itc = ex_to<lst>(*itr).begin(), j = 0; itc != ex_to<lst>(*itr).end(); ++itc, ++j)
+			M(i, j) = *itc;
+	}
+
+	return M;
 }
 
 ex diag_matrix(const lst & l)
 {
-	unsigned dim = l.nops();
+	lst::const_iterator it;
+	size_t dim = l.nops();
 
-	matrix &m = *new matrix(dim, dim);
-	m.setflag(status_flags::dynallocated);
-	for (unsigned i=0; i<dim; i++)
-		m(i, i) = l.op(i);
+	// Allocate and fill matrix
+	matrix &M = *new matrix(dim, dim);
+	M.setflag(status_flags::dynallocated);
 
-	return m;
+	unsigned i;
+	for (it = l.begin(), i = 0; it != l.end(); ++it, ++i)
+		M(i, i) = *it;
+
+	return M;
 }
 
 ex unit_matrix(unsigned r, unsigned c)
 {
-	matrix Id(r,c);
-	for (unsigned i=0; i<r && i<c; ++i)
+	matrix &Id = *new matrix(r, c);
+	Id.setflag(status_flags::dynallocated);
+	for (unsigned i=0; i<r && i<c; i++)
 		Id(i,i) = _ex1;
+
 	return Id;
 }
 
+ex symbolic_matrix(unsigned r, unsigned c, const std::string & base_name, const std::string & tex_base_name)
+{
+	matrix &M = *new matrix(r, c);
+	M.setflag(status_flags::dynallocated | status_flags::evaluated);
+
+	bool long_format = (r > 10 || c > 10);
+	bool single_row = (r == 1 || c == 1);
+
+	for (unsigned i=0; i<r; i++) {
+		for (unsigned j=0; j<c; j++) {
+			std::ostringstream s1, s2;
+			s1 << base_name;
+			s2 << tex_base_name << "_{";
+			if (single_row) {
+				if (c == 1) {
+					s1 << i;
+					s2 << i << '}';
+				} else {
+					s1 << j;
+					s2 << j << '}';
+				}
+			} else {
+				if (long_format) {
+					s1 << '_' << i << '_' << j;
+					s2 << i << ';' << j << "}";
+				} else {
+					s1 << i << j;
+					s2 << i << j << '}';
+				}
+			}
+			M(i, j) = symbol(s1.str(), s2.str());
+		}
+	}
+
+	return M;
+}
+
 } // namespace GiNaC
