@@ -3,7 +3,7 @@
  *  Interface to symbolic matrices */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,45 +24,113 @@
 #define __GINAC_MATRIX_H__
 
 #include <vector>
+#include <string>
 #include "basic.h"
 #include "ex.h"
 
 namespace GiNaC {
 
+
+/** Helper template to allow initialization of matrices via an overloaded
+ *  comma operator (idea stolen from Blitz++). */
+template <typename T, typename It>
+class matrix_init {
+public:
+	matrix_init(It i) : iter(i) {}
+
+	matrix_init<T, It> operator,(const T & x)
+	{
+		*iter = x;
+		return matrix_init<T, It>(++iter);
+	}
+
+	// The following specializations produce much tighter code than the
+	// general case above
+
+	matrix_init<T, It> operator,(int x)
+	{
+		*iter = T(x);
+		return matrix_init<T, It>(++iter);
+	}
+
+	matrix_init<T, It> operator,(unsigned int x)
+	{
+		*iter = T(x);
+		return matrix_init<T, It>(++iter);
+	}
+
+	matrix_init<T, It> operator,(long x)
+	{
+		*iter = T(x);
+		return matrix_init<T, It>(++iter);
+	}
+
+	matrix_init<T, It> operator,(unsigned long x)
+	{
+		*iter = T(x);
+		return matrix_init<T, It>(++iter);
+	}
+
+	matrix_init<T, It> operator,(double x)
+	{
+		*iter = T(x);
+		return matrix_init<T, It>(++iter);
+	}
+
+	matrix_init<T, It> operator,(const symbol & x)
+	{
+		*iter = T(x);
+		return matrix_init<T, It>(++iter);
+	}
+
+private:
+	matrix_init();
+	It iter;
+};
+
+
 /** Symbolic matrices. */
 class matrix : public basic
 {
 	GINAC_DECLARE_REGISTERED_CLASS(matrix, basic)
 	
-	// other ctors
+	// other constructors
 public:
 	matrix(unsigned r, unsigned c);
 	matrix(unsigned r, unsigned c, const exvector & m2);
 	matrix(unsigned r, unsigned c, const lst & l);
+
+	// First step of initialization of matrix with a comma-separated seqeuence
+	// of expressions. Subsequent steps are handled by matrix_init<>::operator,().
+	matrix_init<ex, exvector::iterator> operator=(const ex & x)
+	{
+		m[0] = x;
+		return matrix_init<ex, exvector::iterator>(++m.begin());
+	}
 	
 	// functions overriding virtual functions from base classes
 public:
-	void print(const print_context & c, unsigned level = 0) const;
-	unsigned nops() const;
-	ex op(int i) const;
-	ex & let_op(int i);
+	size_t nops() const;
+	ex op(size_t i) const;
+	ex & let_op(size_t i);
 	ex eval(int level=0) const;
-	ex evalm(void) const {return *this;}
-	ex subs(const lst & ls, const lst & lr, bool no_pattern = false) const;
+	ex evalm() const {return *this;}
+	ex subs(const exmap & m, unsigned options = 0) const;
 	ex eval_indexed(const basic & i) const;
 	ex add_indexed(const ex & self, const ex & other) const;
 	ex scalar_mul_indexed(const ex & self, const numeric & other) const;
 	bool contract_with(exvector::iterator self, exvector::iterator other, exvector & v) const;
+	ex conjugate() const;
 
 protected:
 	bool match_same_type(const basic & other) const;
-	unsigned return_type(void) const { return return_types::noncommutative; };
+	unsigned return_type() const { return return_types::noncommutative; };
 	
 	// non-virtual functions in this class
 public:
-	unsigned rows(void) const        /// Get number of rows.
+	unsigned rows() const        /// Get number of rows.
 		{ return row; }
-	unsigned cols(void) const        /// Get number of columns.
+	unsigned cols() const        /// Get number of columns.
 		{ return col; }
 	matrix add(const matrix & other) const;
 	matrix sub(const matrix & other) const;
@@ -73,19 +141,24 @@ public:
 	const ex & operator() (unsigned ro, unsigned co) const;
 	ex & operator() (unsigned ro, unsigned co);
 	matrix & set(unsigned ro, unsigned co, const ex & value) { (*this)(ro, co) = value; return *this; }
-	matrix transpose(void) const;
+	matrix transpose() const;
 	ex determinant(unsigned algo = determinant_algo::automatic) const;
-	ex trace(void) const;
-	ex charpoly(const symbol & lambda) const;
-	matrix inverse(void) const;
+	ex trace() const;
+	ex charpoly(const ex & lambda) const;
+	matrix inverse() const;
 	matrix solve(const matrix & vars, const matrix & rhs,
 	             unsigned algo = solve_algo::automatic) const;
 protected:
-	ex determinant_minor(void) const;
+	ex determinant_minor() const;
 	int gauss_elimination(const bool det = false);
 	int division_free_elimination(const bool det = false);
 	int fraction_free_elimination(const bool det = false);
 	int pivot(unsigned ro, unsigned co, bool symbolic = true);
+
+	void print_elements(const print_context & c, const char *row_start, const char *row_end, const char *row_sep, const char *col_sep) const;
+	void do_print(const print_context & c, unsigned level) const;
+	void do_print_latex(const print_latex & c, unsigned level) const;
+	void do_print_python_repr(const print_python_repr & c, unsigned level) const;
 	
 // member variables
 protected:
@@ -97,7 +170,7 @@ protected:
 
 // wrapper functions around member functions
 
-inline unsigned nops(const matrix & m)
+inline size_t nops(const matrix & m)
 { return m.nops(); }
 
 inline ex expand(const matrix & m, unsigned options = 0)
@@ -124,7 +197,7 @@ inline ex determinant(const matrix & m, unsigned options = determinant_algo::aut
 inline ex trace(const matrix & m)
 { return m.trace(); }
 
-inline ex charpoly(const matrix & m, const symbol & lambda)
+inline ex charpoly(const matrix & m, const ex & lambda)
 { return m.charpoly(lambda); }
 
 inline matrix inverse(const matrix & m)
@@ -144,13 +217,23 @@ extern ex lst_to_matrix(const lst & l);
 /** Convert list of diagonal elements to matrix. */
 extern ex diag_matrix(const lst & l);
 
-/** Create a r times c unit matrix. */
+/** Create an r times c unit matrix. */
 extern ex unit_matrix(unsigned r, unsigned c);
 
 /** Create a x times x unit matrix. */
 inline ex unit_matrix(unsigned x)
 { return unit_matrix(x, x); }
 
+/** Create an r times c matrix of newly generated symbols consisting of the
+ *  given base name plus the numeric row/column position of each element.
+ *  The base name for LaTeX output is specified separately. */
+extern ex symbolic_matrix(unsigned r, unsigned c, const std::string & base_name, const std::string & tex_base_name);
+
+/** Create an r times c matrix of newly generated symbols consisting of the
+ *  given base name plus the numeric row/column position of each element. */
+inline ex symbolic_matrix(unsigned r, unsigned c, const std::string & base_name)
+{ return symbolic_matrix(r, c, base_name, base_name); }
+
 } // namespace GiNaC
 
 #endif // ndef __GINAC_MATRIX_H__
