@@ -3,7 +3,7 @@
  *  Implementation of GiNaC's products of expressions. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,20 +23,29 @@
 #include <iostream>
 #include <vector>
 #include <stdexcept>
+#include <limits>
 
 #include "mul.h"
 #include "add.h"
 #include "power.h"
+#include "operators.h"
 #include "matrix.h"
+#include "lst.h"
 #include "archive.h"
 #include "utils.h"
 
 namespace GiNaC {
 
-GINAC_IMPLEMENT_REGISTERED_CLASS(mul, expairseq)
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(mul, expairseq,
+  print_func<print_context>(&mul::do_print).
+  print_func<print_latex>(&mul::do_print_latex).
+  print_func<print_csrc>(&mul::do_print_csrc).
+  print_func<print_tree>(&mul::do_print_tree).
+  print_func<print_python_repr>(&mul::do_print_python_repr))
+
 
 //////////
-// default ctor, dtor, copy ctor, assignment operator and helpers
+// default constructor
 //////////
 
 mul::mul()
@@ -44,11 +53,8 @@ mul::mul()
 	tinfo_key = TINFO_mul;
 }
 
-DEFAULT_COPY(mul)
-DEFAULT_DESTROY(mul)
-
 //////////
-// other ctors
+// other constructors
 //////////
 
 // public
@@ -85,13 +91,12 @@ mul::mul(const epvector & v, const ex & oc)
 	GINAC_ASSERT(is_canonical());
 }
 
-mul::mul(epvector * vp, const ex & oc)
+mul::mul(std::auto_ptr<epvector> vp, const ex & oc)
 {
 	tinfo_key = TINFO_mul;
-	GINAC_ASSERT(vp!=0);
+	GINAC_ASSERT(vp.get()!=0);
 	overall_coeff = oc;
 	construct_from_epvector(*vp);
-	delete vp;
 	GINAC_ASSERT(is_canonical());
 }
 
@@ -118,121 +123,154 @@ DEFAULT_ARCHIVING(mul)
 // functions overriding virtual functions from base classes
 //////////
 
-// public
-
-void mul::print(const print_context & c, unsigned level) const
+void mul::print_overall_coeff(const print_context & c, const char *mul_sym) const
 {
-	if (is_a<print_tree>(c)) {
-
-		inherited::print(c, level);
-
-	} else if (is_a<print_csrc>(c)) {
-
-		if (precedence() <= level)
-			c.s << "(";
-
-		if (!overall_coeff.is_equal(_ex1)) {
-			overall_coeff.print(c, precedence());
-			c.s << "*";
+	const numeric &coeff = ex_to<numeric>(overall_coeff);
+	if (coeff.csgn() == -1)
+		c.s << '-';
+	if (!coeff.is_equal(_num1) &&
+		!coeff.is_equal(_num_1)) {
+		if (coeff.is_rational()) {
+			if (coeff.is_negative())
+				(-coeff).print(c);
+			else
+				coeff.print(c);
+		} else {
+			if (coeff.csgn() == -1)
+				(-coeff).print(c, precedence());
+			else
+				coeff.print(c, precedence());
 		}
+		c.s << mul_sym;
+	}
+}
 
-		// Print arguments, separated by "*" or "/"
-		epvector::const_iterator it = seq.begin(), itend = seq.end();
-		while (it != itend) {
+void mul::do_print(const print_context & c, unsigned level) const
+{
+	if (precedence() <= level)
+		c.s << '(';
 
-			// If the first argument is a negative integer power, it gets printed as "1.0/<expr>"
-			if (it == seq.begin() && ex_to<numeric>(it->coeff).is_integer() && it->coeff.info(info_flags::negative)) {
-				if (is_a<print_csrc_cl_N>(c))
-					c.s << "recip(";
-				else
-					c.s << "1.0/";
-			}
+	print_overall_coeff(c, "*");
 
-			// If the exponent is 1 or -1, it is left out
-			if (it->coeff.is_equal(_ex1) || it->coeff.is_equal(_ex_1))
-				it->rest.print(c, precedence());
-			else {
-				// Outer parens around ex needed for broken gcc-2.95 parser:
-				(ex(power(it->rest, abs(ex_to<numeric>(it->coeff))))).print(c, level);
-			}
+	epvector::const_iterator it = seq.begin(), itend = seq.end();
+	bool first = true;
+	while (it != itend) {
+		if (!first)
+			c.s << '*';
+		else
+			first = false;
+		recombine_pair_to_ex(*it).print(c, precedence());
+		++it;
+	}
 
-			// Separator is "/" for negative integer powers, "*" otherwise
-			++it;
-			if (it != itend) {
-				if (ex_to<numeric>(it->coeff).is_integer() && it->coeff.info(info_flags::negative))
-					c.s << "/";
-				else
-					c.s << "*";
-			}
-		}
-
-		if (precedence() <= level)
-			c.s << ")";
-
-	} else if (is_a<print_python_repr>(c)) {
-		c.s << class_name() << '(';
-		op(0).print(c);
-		for (unsigned i=1; i<nops(); ++i) {
-			c.s << ',';
-			op(i).print(c);
-		}
+	if (precedence() <= level)
 		c.s << ')';
+}
+
+void mul::do_print_latex(const print_latex & c, unsigned level) const
+{
+	if (precedence() <= level)
+		c.s << "{(";
+
+	print_overall_coeff(c, " ");
+
+	// Separate factors into those with negative numeric exponent
+	// and all others
+	epvector::const_iterator it = seq.begin(), itend = seq.end();
+	exvector neg_powers, others;
+	while (it != itend) {
+		GINAC_ASSERT(is_exactly_a<numeric>(it->coeff));
+		if (ex_to<numeric>(it->coeff).is_negative())
+			neg_powers.push_back(recombine_pair_to_ex(expair(it->rest, -(it->coeff))));
+		else
+			others.push_back(recombine_pair_to_ex(*it));
+		++it;
+	}
+
+	if (!neg_powers.empty()) {
+
+		// Factors with negative exponent are printed as a fraction
+		c.s << "\\frac{";
+		mul(others).eval().print(c);
+		c.s << "}{";
+		mul(neg_powers).eval().print(c);
+		c.s << "}";
+
 	} else {
 
-		if (precedence() <= level) {
-			if (is_a<print_latex>(c))
-				c.s << "{(";
-			else
-				c.s << "(";
-		}
-
-		bool first = true;
-
-		// First print the overall numeric coefficient
-		numeric coeff = ex_to<numeric>(overall_coeff);
-		if (coeff.csgn() == -1)
-			c.s << '-';
-		if (!coeff.is_equal(_num1) &&
-			!coeff.is_equal(_num_1)) {
-			if (coeff.is_rational()) {
-				if (coeff.is_negative())
-					(-coeff).print(c);
-				else
-					coeff.print(c);
-			} else {
-				if (coeff.csgn() == -1)
-					(-coeff).print(c, precedence());
-				else
-					coeff.print(c, precedence());
-			}
-			if (is_a<print_latex>(c))
-				c.s << ' ';
-			else
-				c.s << '*';
-		}
-
-		// Then proceed with the remaining factors
-		epvector::const_iterator it = seq.begin(), itend = seq.end();
-		while (it != itend) {
-			if (!first) {
-				if (is_a<print_latex>(c))
-					c.s << ' ';
-				else
-					c.s << '*';
-			} else {
-				first = false;
-			}
-			recombine_pair_to_ex(*it).print(c, precedence());
-			++it;
-		}
-
-		if (precedence() <= level) {
-			if (is_a<print_latex>(c))
-				c.s << ")}";
-			else
-				c.s << ")";
+		// All other factors are printed in the ordinary way
+		exvector::const_iterator vit = others.begin(), vitend = others.end();
+		while (vit != vitend) {
+			c.s << ' ';
+			vit->print(c, precedence());
+			++vit;
 		}
 	}
+
+	if (precedence() <= level)
+		c.s << ")}";
+}
+
+void mul::do_print_csrc(const print_csrc & c, unsigned level) const
+{
+	if (precedence() <= level)
+		c.s << "(";
+
+	if (!overall_coeff.is_equal(_ex1)) {
+		overall_coeff.print(c, precedence());
+		c.s << "*";
+	}
+
+	// Print arguments, separated by "*" or "/"
+	epvector::const_iterator it = seq.begin(), itend = seq.end();
+	while (it != itend) {
+
+		// If the first argument is a negative integer power, it gets printed as "1.0/<expr>"
+		bool needclosingparenthesis = false;
+		if (it == seq.begin() && it->coeff.info(info_flags::negint)) {
+			if (is_a<print_csrc_cl_N>(c)) {
+				c.s << "recip(";
+				needclosingparenthesis = true;
+			} else
+				c.s << "1.0/";
+		}
+
+		// If the exponent is 1 or -1, it is left out
+		if (it->coeff.is_equal(_ex1) || it->coeff.is_equal(_ex_1))
+			it->rest.print(c, precedence());
+		else if (it->coeff.info(info_flags::negint))
+			// Outer parens around ex needed for broken GCC parser:
+			(ex(power(it->rest, -ex_to<numeric>(it->coeff)))).print(c, level);
+		else
+			// Outer parens around ex needed for broken GCC parser:
+			(ex(power(it->rest, ex_to<numeric>(it->coeff)))).print(c, level);
+
+		if (needclosingparenthesis)
+			c.s << ")";
+
+		// Separator is "/" for negative integer powers, "*" otherwise
+		++it;
+		if (it != itend) {
+			if (it->coeff.info(info_flags::negint))
+				c.s << "/";
+			else
+				c.s << "*";
+		}
+	}
+
+	if (precedence() <= level)
+		c.s << ")";
+}
+
+void mul::do_print_python_repr(const print_python_repr & c, unsigned level) const
+{
+	c.s << class_name() << '(';
+	op(0).print(c);
+	for (size_t i=1; i<nops(); ++i) {
+		c.s << ',';
+		op(i).print(c);
+	}
+	c.s << ')';
 }
 
 bool mul::info(unsigned inf) const
@@ -340,10 +378,10 @@ ex mul::coeff(const ex & s, int n) const
  *  @param level cut-off in recursive evaluation */
 ex mul::eval(int level) const
 {
-	epvector *evaled_seqp = evalchildren(level);
-	if (evaled_seqp) {
+	std::auto_ptr<epvector> evaled_seqp = evalchildren(level);
+	if (evaled_seqp.get()) {
 		// do more evaluation later
-		return (new mul(evaled_seqp,overall_coeff))->
+		return (new mul(evaled_seqp, overall_coeff))->
 		           setflag(status_flags::dynallocated);
 	}
 	
@@ -353,7 +391,7 @@ ex mul::eval(int level) const
 		GINAC_ASSERT((!is_exactly_a<mul>(i->rest)) ||
 		             (!(ex_to<numeric>(i->coeff).is_integer())));
 		GINAC_ASSERT(!(i->is_canonical_numeric()));
-		if (is_ex_exactly_of_type(recombine_pair_to_ex(*i), numeric))
+		if (is_exactly_a<numeric>(recombine_pair_to_ex(*i)))
 		    print(print_tree(std::cerr));
 		GINAC_ASSERT(!is_exactly_a<numeric>(recombine_pair_to_ex(*i)));
 		/* for paranoia */
@@ -382,11 +420,11 @@ ex mul::eval(int level) const
 		// *(x;1) -> x
 		return recombine_pair_to_ex(*(seq.begin()));
 	} else if ((seq_size==1) &&
-	           is_ex_exactly_of_type((*seq.begin()).rest,add) &&
+	           is_exactly_a<add>((*seq.begin()).rest) &&
 	           ex_to<numeric>((*seq.begin()).coeff).is_equal(_num1)) {
 		// *(+(x,y,...);c) -> +(*(x,c),*(y,c),...) (c numeric(), no powers of +())
 		const add & addref = ex_to<add>((*seq.begin()).rest);
-		epvector *distrseq = new epvector();
+		std::auto_ptr<epvector> distrseq(new epvector);
 		distrseq->reserve(addref.seq.size());
 		epvector::const_iterator i = addref.seq.begin(), end = addref.seq.end();
 		while (i != end) {
@@ -409,7 +447,7 @@ ex mul::evalf(int level) const
 	if (level==-max_recursion_level)
 		throw(std::runtime_error("max recursion level reached"));
 	
-	epvector *s = new epvector();
+	std::auto_ptr<epvector> s(new epvector);
 	s->reserve(seq.size());
 
 	--level;
@@ -422,17 +460,17 @@ ex mul::evalf(int level) const
 	return mul(s, overall_coeff.evalf(level));
 }
 
-ex mul::evalm(void) const
+ex mul::evalm() const
 {
 	// numeric*matrix
 	if (seq.size() == 1 && seq[0].coeff.is_equal(_ex1)
-	 && is_ex_of_type(seq[0].rest, matrix))
+	 && is_a<matrix>(seq[0].rest))
 		return ex_to<matrix>(seq[0].rest).mul(ex_to<numeric>(overall_coeff));
 
 	// Evaluate children first, look whether there are any matrices at all
 	// (there can be either no matrices or one matrix; if there were more
 	// than one matrix, it would be a non-commutative product)
-	epvector *s = new epvector;
+	std::auto_ptr<epvector> s(new epvector);
 	s->reserve(seq.size());
 
 	bool have_matrix = false;
@@ -442,7 +480,7 @@ ex mul::evalm(void) const
 	while (i != end) {
 		const ex &m = recombine_pair_to_ex(*i).evalm();
 		s->push_back(split_ex_to_pair(m));
-		if (is_ex_of_type(m, matrix)) {
+		if (is_a<matrix>(m)) {
 			have_matrix = true;
 			the_matrix = s->end() - 1;
 		}
@@ -462,19 +500,143 @@ ex mul::evalm(void) const
 		return (new mul(s, overall_coeff))->setflag(status_flags::dynallocated);
 }
 
-ex mul::simplify_ncmul(const exvector & v) const
+ex mul::eval_ncmul(const exvector & v) const
 {
 	if (seq.empty())
-		return inherited::simplify_ncmul(v);
+		return inherited::eval_ncmul(v);
 
-	// Find first noncommutative element and call its simplify_ncmul()
+	// Find first noncommutative element and call its eval_ncmul()
 	epvector::const_iterator i = seq.begin(), end = seq.end();
 	while (i != end) {
 		if (i->rest.return_type() == return_types::noncommutative)
-			return i->rest.simplify_ncmul(v);
+			return i->rest.eval_ncmul(v);
 		++i;
 	}
-	return inherited::simplify_ncmul(v);
+	return inherited::eval_ncmul(v);
+}
+
+bool tryfactsubs(const ex & origfactor, const ex & patternfactor, int & nummatches, lst & repls)
+{	
+	ex origbase;
+	int origexponent;
+	int origexpsign;
+
+	if (is_exactly_a<power>(origfactor) && origfactor.op(1).info(info_flags::integer)) {
+		origbase = origfactor.op(0);
+		int expon = ex_to<numeric>(origfactor.op(1)).to_int();
+		origexponent = expon > 0 ? expon : -expon;
+		origexpsign = expon > 0 ? 1 : -1;
+	} else {
+		origbase = origfactor;
+		origexponent = 1;
+		origexpsign = 1;
+	}
+
+	ex patternbase;
+	int patternexponent;
+	int patternexpsign;
+
+	if (is_exactly_a<power>(patternfactor) && patternfactor.op(1).info(info_flags::integer)) {
+		patternbase = patternfactor.op(0);
+		int expon = ex_to<numeric>(patternfactor.op(1)).to_int();
+		patternexponent = expon > 0 ? expon : -expon;
+		patternexpsign = expon > 0 ? 1 : -1;
+	} else {
+		patternbase = patternfactor;
+		patternexponent = 1;
+		patternexpsign = 1;
+	}
+
+	lst saverepls = repls;
+	if (origexponent < patternexponent || origexpsign != patternexpsign || !origbase.match(patternbase,saverepls))
+		return false;
+	repls = saverepls;
+
+	int newnummatches = origexponent / patternexponent;
+	if (newnummatches < nummatches)
+		nummatches = newnummatches;
+	return true;
+}
+
+ex mul::algebraic_subs_mul(const exmap & m, unsigned options) const
+{	
+	std::vector<bool> subsed(seq.size(), false);
+	exvector subsresult(seq.size());
+
+	for (exmap::const_iterator it = m.begin(); it != m.end(); ++it) {
+
+		if (is_exactly_a<mul>(it->first)) {
+
+			int nummatches = std::numeric_limits<int>::max();
+			std::vector<bool> currsubsed(seq.size(), false);
+			bool succeed = true;
+			lst repls;
+
+			for (size_t j=0; j<it->first.nops(); j++) {
+				bool found=false;
+				for (size_t k=0; k<nops(); k++) {
+					if (currsubsed[k] || subsed[k])
+						continue;
+					if (tryfactsubs(op(k), it->first.op(j), nummatches, repls)) {
+						currsubsed[k] = true;
+						found = true;
+						break;
+					}
+				}
+				if (!found) {
+					succeed = false;
+					break;
+				}
+			}
+			if (!succeed)
+				continue;
+
+			bool foundfirstsubsedfactor = false;
+			for (size_t j=0; j<subsed.size(); j++) {
+				if (currsubsed[j]) {
+					if (foundfirstsubsedfactor)
+						subsresult[j] = op(j);
+					else {
+						foundfirstsubsedfactor = true;
+						subsresult[j] = op(j) * power(it->second.subs(ex(repls), subs_options::no_pattern) / it->first.subs(ex(repls), subs_options::no_pattern), nummatches);
+					}
+					subsed[j] = true;
+				}
+			}
+
+		} else {
+
+			int nummatches = std::numeric_limits<int>::max();
+			lst repls;
+
+			for (size_t j=0; j<this->nops(); j++) {
+				if (!subsed[j] && tryfactsubs(op(j), it->first, nummatches, repls)) {
+					subsed[j] = true;
+					subsresult[j] = op(j) * power(it->second.subs(ex(repls), subs_options::no_pattern) / it->first.subs(ex(repls), subs_options::no_pattern), nummatches);
+				}
+			}
+		}
+	}
+
+	bool subsfound = false;
+	for (size_t i=0; i<subsed.size(); i++) {
+		if (subsed[i]) {
+			subsfound = true;
+			break;
+		}
+	}
+	if (!subsfound)
+		return subs_one_level(m, options | subs_options::algebraic);
+
+	exvector ev; ev.reserve(nops());
+	for (size_t i=0; i<nops(); i++) {
+		if (subsed[i])
+			ev.push_back(subsresult[i]);
+		else
+			ev.push_back(op(i));
+	}
+
+	return (new mul(ev))->setflag(status_flags::dynallocated);
 }
 
 // protected
@@ -483,7 +645,7 @@ ex mul::simplify_ncmul(const exvector & v) const
  *  @see ex::diff */
 ex mul::derivative(const symbol & s) const
 {
-	unsigned num = seq.size();
+	size_t num = seq.size();
 	exvector addseq;
 	addseq.reserve(num);
 	
@@ -507,12 +669,7 @@ int mul::compare_same_type(const basic & other) const
 	return inherited::compare_same_type(other);
 }
 
-bool mul::is_equal_same_type(const basic & other) const
-{
-	return inherited::is_equal_same_type(other);
-}
-
-unsigned mul::return_type(void) const
+unsigned mul::return_type() const
 {
 	if (seq.empty()) {
 		// mul without factors: should not happen, but commutes
@@ -545,7 +702,7 @@ unsigned mul::return_type(void) const
 	return all_commutative ? return_types::commutative : return_types::noncommutative;
 }
    
-unsigned mul::return_type_tinfo(void) const
+unsigned mul::return_type_tinfo() const
 {
 	if (seq.empty())
 		return tinfo_key;  // mul without factors: should not happen
@@ -566,16 +723,16 @@ ex mul::thisexpairseq(const epvector & v, const ex & oc) const
 	return (new mul(v, oc))->setflag(status_flags::dynallocated);
 }
 
-ex mul::thisexpairseq(epvector * vp, const ex & oc) const
+ex mul::thisexpairseq(std::auto_ptr<epvector> vp, const ex & oc) const
 {
 	return (new mul(vp, oc))->setflag(status_flags::dynallocated);
 }
 
 expair mul::split_ex_to_pair(const ex & e) const
 {
-	if (is_ex_exactly_of_type(e,power)) {
+	if (is_exactly_a<power>(e)) {
 		const power & powerref = ex_to<power>(e);
-		if (is_ex_exactly_of_type(powerref.exponent,numeric))
+		if (is_exactly_a<numeric>(powerref.exponent))
 			return expair(powerref.basis,powerref.exponent);
 	}
 	return expair(e,_ex1);
@@ -586,11 +743,11 @@ expair mul::combine_ex_with_coeff_to_pair(const ex & e,
 {
 	// to avoid duplication of power simplification rules,
 	// we create a temporary power object
-	// otherwise it would be hard to correctly simplify
+	// otherwise it would be hard to correctly evaluate
 	// expression like (4^(1/3))^(3/2)
-	if (are_ex_trivially_equal(c,_ex1))
+	if (c.is_equal(_ex1))
 		return split_ex_to_pair(e);
-	
+
 	return split_ex_to_pair(power(e,c));
 }
 	
@@ -599,11 +756,11 @@ expair mul::combine_pair_with_coeff_to_pair(const expair & p,
 {
 	// to avoid duplication of power simplification rules,
 	// we create a temporary power object
-	// otherwise it would be hard to correctly simplify
+	// otherwise it would be hard to correctly evaluate
 	// expression like (4^(1/3))^(3/2)
-	if (are_ex_trivially_equal(c,_ex1))
+	if (c.is_equal(_ex1))
 		return p;
-	
+
 	return split_ex_to_pair(power(recombine_pair_to_ex(p),c));
 }
 	
@@ -612,25 +769,25 @@ ex mul::recombine_pair_to_ex(const expair & p) const
 	if (ex_to<numeric>(p.coeff).is_equal(_num1)) 
 		return p.rest;
 	else
-		return power(p.rest,p.coeff);
+		return (new power(p.rest,p.coeff))->setflag(status_flags::dynallocated);
 }
 
 bool mul::expair_needs_further_processing(epp it)
 {
-	if (is_ex_exactly_of_type((*it).rest,mul) &&
-		ex_to<numeric>((*it).coeff).is_integer()) {
+	if (is_exactly_a<mul>(it->rest) &&
+		ex_to<numeric>(it->coeff).is_integer()) {
 		// combined pair is product with integer power -> expand it
 		*it = split_ex_to_pair(recombine_pair_to_ex(*it));
 		return true;
 	}
-	if (is_ex_exactly_of_type((*it).rest,numeric)) {
-		expair ep=split_ex_to_pair(recombine_pair_to_ex(*it));
+	if (is_exactly_a<numeric>(it->rest)) {
+		expair ep = split_ex_to_pair(recombine_pair_to_ex(*it));
 		if (!ep.is_equal(*it)) {
 			// combined pair is a numeric power which can be simplified
 			*it = ep;
 			return true;
 		}
-		if (ex_to<numeric>((*it).coeff).is_equal(_num1)) {
+		if (it->coeff.is_equal(_ex1)) {
 			// combined pair has coeff 1 and must be moved to the end
 			return true;
 		}
@@ -638,7 +795,7 @@ bool mul::expair_needs_further_processing(epp it)
 	return false;
 }       
 
-ex mul::default_overall_coeff(void) const
+ex mul::default_overall_coeff() const
 {
 	return _ex1;
 }
@@ -667,45 +824,40 @@ bool mul::can_make_flat(const expair & p) const
 	return ex_to<numeric>(p.coeff).is_equal(_num1);
 }
 
+bool mul::can_be_further_expanded(const ex & e)
+{
+	if (is_exactly_a<mul>(e)) {
+		for (epvector::const_iterator cit = ex_to<mul>(e).seq.begin(); cit != ex_to<mul>(e).seq.end(); ++cit) {
+			if (is_exactly_a<add>(cit->rest) && cit->coeff.info(info_flags::posint))
+				return true;
+		}
+	} else if (is_exactly_a<power>(e)) {
+		if (is_exactly_a<add>(e.op(0)) && e.op(1).info(info_flags::posint))
+			return true;
+	}
+	return false;
+}
+
 ex mul::expand(unsigned options) const
 {
 	// First, expand the children
-	epvector * expanded_seqp = expandchildren(options);
-	const epvector & expanded_seq = (expanded_seqp == NULL) ? seq : *expanded_seqp;
+	std::auto_ptr<epvector> expanded_seqp = expandchildren(options);
+	const epvector & expanded_seq = (expanded_seqp.get() ? *expanded_seqp : seq);
 
 	// Now, look for all the factors that are sums and multiply each one out
 	// with the next one that is found while collecting the factors which are
 	// not sums
-	int number_of_adds = 0;
 	ex last_expanded = _ex1;
+	bool need_reexpand = false;
+
 	epvector non_adds;
 	non_adds.reserve(expanded_seq.size());
-	epvector::const_iterator cit = expanded_seq.begin(), last = expanded_seq.end();
-	while (cit != last) {
-		if (is_ex_exactly_of_type(cit->rest, add) &&
+
+	for (epvector::const_iterator cit = expanded_seq.begin(); cit != expanded_seq.end(); ++cit) {
+		if (is_exactly_a<add>(cit->rest) &&
 			(cit->coeff.is_equal(_ex1))) {
-			++number_of_adds;
-			if (is_ex_exactly_of_type(last_expanded, add)) {
-#if 0
-				// Expand a product of two sums, simple and robust version.
-				const add & add1 = ex_to<add>(last_expanded);
-				const add & add2 = ex_to<add>(cit->rest);
-				const int n1 = add1.nops();
-				const int n2 = add2.nops();
-				ex tmp_accu;
-				exvector distrseq;
-				distrseq.reserve(n2);
-				for (int i1=0; i1<n1; ++i1) {
-					distrseq.clear();
-					// cache the first operand (for efficiency):
-					const ex op1 = add1.op(i1);
-					for (int i2=0; i2<n2; ++i2)
-						distrseq.push_back(op1 * add2.op(i2));
-					tmp_accu += (new add(distrseq))->
-					             setflag(status_flags::dynallocated);
-				}
-				last_expanded = tmp_accu;
-#else
+			if (is_exactly_a<add>(last_expanded)) {
+
 				// Expand a product of two sums, aggressive version.
 				// Caring for the overall coefficients in separate loops can
 				// sometimes give a performance gain of up to 15%!
@@ -721,6 +873,7 @@ ex mul::expand(unsigned options) const
 				const epvector::const_iterator add2end   = add2.seq.end();
 				epvector distrseq;
 				distrseq.reserve(add1.seq.size()+add2.seq.size());
+
 				// Multiply add2 with the overall coefficient of add1 and append it to distrseq:
 				if (!add1.overall_coeff.is_zero()) {
 					if (add1.overall_coeff.is_equal(_ex1))
@@ -729,6 +882,7 @@ ex mul::expand(unsigned options) const
 						for (epvector::const_iterator i=add2begin; i!=add2end; ++i)
 							distrseq.push_back(expair(i->rest, ex_to<numeric>(i->coeff).mul_dyn(ex_to<numeric>(add1.overall_coeff))));
 				}
+
 				// Multiply add1 with the overall coefficient of add2 and append it to distrseq:
 				if (!add2.overall_coeff.is_zero()) {
 					if (add2.overall_coeff.is_equal(_ex1))
@@ -737,8 +891,10 @@ ex mul::expand(unsigned options) const
 						for (epvector::const_iterator i=add1begin; i!=add1end; ++i)
 							distrseq.push_back(expair(i->rest, ex_to<numeric>(i->coeff).mul_dyn(ex_to<numeric>(add2.overall_coeff))));
 				}
+
 				// Compute the new overall coefficient and put it together:
 				ex tmp_accu = (new add(distrseq, add1.overall_coeff*add2.overall_coeff))->setflag(status_flags::dynallocated);
+
 				// Multiply explicitly all non-numeric terms of add1 and add2:
 				for (epvector::const_iterator i1=add1begin; i1!=add1end; ++i1) {
 					// We really have to combine terms here in order to compactify
@@ -749,7 +905,7 @@ ex mul::expand(unsigned options) const
 						// Don't push_back expairs which might have a rest that evaluates to a numeric,
 						// since that would violate an invariant of expairseq:
 						const ex rest = (new mul(i1->rest, i2->rest))->setflag(status_flags::dynallocated);
-						if (is_ex_exactly_of_type(rest, numeric))
+						if (is_exactly_a<numeric>(rest))
 							oc += ex_to<numeric>(rest).mul(ex_to<numeric>(i1->coeff).mul(ex_to<numeric>(i2->coeff)));
 						else
 							distrseq.push_back(expair(rest, ex_to<numeric>(i1->coeff).mul_dyn(ex_to<numeric>(i2->coeff))));
@@ -757,38 +913,51 @@ ex mul::expand(unsigned options) const
 					tmp_accu += (new add(distrseq, oc))->setflag(status_flags::dynallocated);
 				}
 				last_expanded = tmp_accu;
-#endif
+
 			} else {
-				non_adds.push_back(split_ex_to_pair(last_expanded));
+				if (!last_expanded.is_equal(_ex1))
+					non_adds.push_back(split_ex_to_pair(last_expanded));
 				last_expanded = cit->rest;
 			}
+
 		} else {
 			non_adds.push_back(*cit);
 		}
-		++cit;
 	}
-	if (expanded_seqp)
-		delete expanded_seqp;
-	
+
 	// Now the only remaining thing to do is to multiply the factors which
 	// were not sums into the "last_expanded" sum
-	if (is_ex_exactly_of_type(last_expanded, add)) {
-		const add & finaladd = ex_to<add>(last_expanded);
+	if (is_exactly_a<add>(last_expanded)) {
+		size_t n = last_expanded.nops();
 		exvector distrseq;
-		int n = finaladd.nops();
 		distrseq.reserve(n);
-		for (int i=0; i<n; ++i) {
+
+		for (size_t i=0; i<n; ++i) {
 			epvector factors = non_adds;
-			factors.push_back(split_ex_to_pair(finaladd.op(i)));
-			distrseq.push_back((new mul(factors, overall_coeff))->
-			                    setflag(status_flags::dynallocated | (options == 0 ? status_flags::expanded : 0)));
+			factors.push_back(split_ex_to_pair(last_expanded.op(i)));
+			ex term = (new mul(factors, overall_coeff))->setflag(status_flags::dynallocated);
+			if (can_be_further_expanded(term))
+				distrseq.push_back(term.expand());
+			else {
+				if (options == 0)
+					ex_to<basic>(term).setflag(status_flags::expanded);
+				distrseq.push_back(term);
+			}
 		}
+
 		return ((new add(distrseq))->
 		        setflag(status_flags::dynallocated | (options == 0 ? status_flags::expanded : 0)));
 	}
+
 	non_adds.push_back(split_ex_to_pair(last_expanded));
-	return (new mul(non_adds, overall_coeff))->
-	        setflag(status_flags::dynallocated | (options == 0 ? status_flags::expanded : 0));
+	ex result = (new mul(non_adds, overall_coeff))->setflag(status_flags::dynallocated);
+	if (can_be_further_expanded(result)) {
+		return result.expand();
+	} else {
+		if (options == 0)
+			ex_to<basic>(result).setflag(status_flags::expanded);
+		return result;
+	}
 }
 
   
@@ -810,7 +979,7 @@ ex mul::expand(unsigned options) const
  *  @see mul::expand()
  *  @return pointer to epvector containing expanded representation or zero
  *  pointer, if sequence is unchanged. */
-epvector * mul::expandchildren(unsigned options) const
+std::auto_ptr<epvector> mul::expandchildren(unsigned options) const
 {
 	const epvector::const_iterator last = seq.end();
 	epvector::const_iterator cit = seq.begin();
@@ -820,7 +989,7 @@ epvector * mul::expandchildren(unsigned options) const
 		if (!are_ex_trivially_equal(factor,expanded_factor)) {
 			
 			// something changed, copy seq, eval and return it
-			epvector *s = new epvector;
+			std::auto_ptr<epvector> s(new epvector);
 			s->reserve(seq.size());
 			
 			// copy parts of seq which are known not to have changed
@@ -829,9 +998,11 @@ epvector * mul::expandchildren(unsigned options) const
 				s->push_back(*cit2);
 				++cit2;
 			}
+
 			// copy first changed element
 			s->push_back(split_ex_to_pair(expanded_factor));
 			++cit2;
+
 			// copy rest
 			while (cit2!=last) {
 				s->push_back(split_ex_to_pair(recombine_pair_to_ex(*cit2).expand(options)));
@@ -842,7 +1013,7 @@ epvector * mul::expandchildren(unsigned options) const
 		++cit;
 	}
 	
-	return 0; // nothing has changed
+	return std::auto_ptr<epvector>(0); // nothing has changed
 }
 
 } // namespace GiNaC
