@@ -6,7 +6,7 @@
  *  computation, square-free factorization and rational function normalization. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -34,16 +34,19 @@ class ex;
 class symbol;
 
 // Quotient q(x) of polynomials a(x) and b(x) in Q[x], so that a(x)=b(x)*q(x)+r(x)
-extern ex quo(const ex &a, const ex &b, const symbol &x, bool check_args = true);
+extern ex quo(const ex &a, const ex &b, const ex &x, bool check_args = true);
 
 // Remainder r(x) of polynomials a(x) and b(x) in Q[x], so that a(x)=b(x)*q(x)+r(x)
-extern ex rem(const ex &a, const ex &b, const symbol &x, bool check_args = true);
+extern ex rem(const ex &a, const ex &b, const ex &x, bool check_args = true);
 
 // Decompose rational function a(x)=N(x)/D(x) into Q(x)+R(x)/D(x) with degree(R, x) < degree(D, x)
-extern ex decomp_rational(const ex &a, const symbol &x);
+extern ex decomp_rational(const ex &a, const ex &x);
 
-// Pseudo-remainder of polynomials a(x) and b(x) in Z[x]
-extern ex prem(const ex &a, const ex &b, const symbol &x, bool check_args = true);
+// Pseudo-remainder of polynomials a(x) and b(x) in Q[x]
+extern ex prem(const ex &a, const ex &b, const ex &x, bool check_args = true);
+
+// Pseudo-remainder of polynomials a(x) and b(x) in Q[x]
+extern ex sprem(const ex &a, const ex &b, const ex &x, bool check_args = true);
 
 // Exact polynomial division of a(X) by b(X) in Q[X] (quotient returned in q), returns false when exact division fails
 extern bool divide(const ex &a, const ex &b, ex &q, bool check_args = true);
@@ -60,6 +63,9 @@ extern ex sqrfree(const ex &a, const lst &l = lst());
 // Square-free partial fraction decomposition of a rational function a(x)
 extern ex sqrfree_parfrac(const ex & a, const symbol & x);
 
+// Collect common factors in sums.
+extern ex collect_common_factors(const ex & e);
+
 } // namespace GiNaC
 
 #endif // ndef __GINAC_NORMAL_H__
