@@ -3,7 +3,7 @@
  *  Makes the interface to the underlying bignum package available. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -26,9 +26,9 @@
 #include "basic.h"
 #include "ex.h"
 
-#include <cln/number.h>
-// forward decln of cln::cl_N, since cln/complex_class.h is not included:
-namespace cln { class cl_N; }
+#include <stdexcept>
+
+#include <cln/complex.h>
 
 #if defined(G__CINTVERSION) && !defined(__MAKECINT__)
 // Cint @$#$! doesn't like forward declaring classes used for casting operators
@@ -62,6 +62,17 @@ private:
 	static bool too_late;               ///< Already one object present
 };
 
+
+/** Exception class thrown when a singularity is encountered. */
+class pole_error : public std::domain_error {
+public:
+	explicit pole_error(const std::string& what_arg, int degree);
+	int degree() const;
+private:
+	int deg;
+};
+
+
 /** This class is a wrapper around CLN-numbers within the GiNaC class
  *  hierarchy. Objects of this type may directly be created by the user.*/
 class numeric : public basic
@@ -70,7 +81,7 @@ class numeric : public basic
 	
 // member functions
 	
-	// other ctors
+	// other constructors
 public:
 	numeric(int i);
 	numeric(unsigned int i);
@@ -82,8 +93,7 @@ public:
 	
 	// functions overriding virtual functions from base classes
 public:
-	void print(const print_context & c, unsigned level = 0) const;
-	unsigned precedence(void) const {return 30;}
+	unsigned precedence() const {return 30;}
 	bool info(unsigned inf) const;
 	int degree(const ex & s) const;
 	int ldegree(const ex & s) const;
@@ -91,17 +101,20 @@ public:
 	bool has(const ex &other) const;
 	ex eval(int level = 0) const;
 	ex evalf(int level = 0) const;
-	ex normal(lst &sym_lst, lst &repl_lst, int level = 0) const;
-	ex to_rational(lst &repl_lst) const;
-	numeric integer_content(void) const;
+	ex subs(const exmap & m, unsigned options = 0) const { return subs_one_level(m, options); } // overwrites basic::subs() for performance reasons
+	ex normal(exmap & repl, exmap & rev_lookup, int level = 0) const;
+	ex to_rational(exmap & repl) const;
+	ex to_polynomial(exmap & repl) const;
+	numeric integer_content() const;
 	ex smod(const numeric &xi) const;
-	numeric max_coefficient(void) const;
+	numeric max_coefficient() const;
+	ex conjugate() const;
 protected:
 	/** Implementation of ex::diff for a numeric always returns 0.
 	 *  @see ex::diff */
 	ex derivative(const symbol &s) const { return 0; }
 	bool is_equal_same_type(const basic &other) const;
-	unsigned calchash(void) const;
+	unsigned calchash() const;
 	
 	// new virtual functions which can be overridden by derived classes
 	// (none)
@@ -124,55 +137,62 @@ public:
 	const numeric & operator=(unsigned long i);
 	const numeric & operator=(double d);
 	const numeric & operator=(const char *s);
-	const numeric inverse(void) const;
-	int csgn(void) const;
+	const numeric inverse() const;
+	int csgn() const;
 	int compare(const numeric &other) const;
 	bool is_equal(const numeric &other) const;
-	bool is_zero(void) const;
-	bool is_positive(void) const;
-	bool is_negative(void) const;
-	bool is_integer(void) const;
-	bool is_pos_integer(void) const;
-	bool is_nonneg_integer(void) const;
-	bool is_even(void) const;
-	bool is_odd(void) const;
-	bool is_prime(void) const;
-	bool is_rational(void) const;
-	bool is_real(void) const;
-	bool is_cinteger(void) const;
-	bool is_crational(void) const;
+	bool is_zero() const;
+	bool is_positive() const;
+	bool is_negative() const;
+	bool is_integer() const;
+	bool is_pos_integer() const;
+	bool is_nonneg_integer() const;
+	bool is_even() const;
+	bool is_odd() const;
+	bool is_prime() const;
+	bool is_rational() const;
+	bool is_real() const;
+	bool is_cinteger() const;
+	bool is_crational() const;
 	bool operator==(const numeric &other) const;
 	bool operator!=(const numeric &other) const;
 	bool operator<(const numeric &other) const;
 	bool operator<=(const numeric &other) const;
 	bool operator>(const numeric &other) const;
 	bool operator>=(const numeric &other) const;
-	int to_int(void) const;
-	long to_long(void) const;
-	double to_double(void) const;
-	cln::cl_N to_cl_N(void) const;
-	const numeric real(void) const;
-	const numeric imag(void) const;
-	const numeric numer(void) const;
-	const numeric denom(void) const;
-	int int_length(void) const;
+	int to_int() const;
+	long to_long() const;
+	double to_double() const;
+	cln::cl_N to_cl_N() const;
+	const numeric real() const;
+	const numeric imag() const;
+	const numeric numer() const;
+	const numeric denom() const;
+	int int_length() const;
 	// converting routines for interfacing with CLN:
 	numeric(const cln::cl_N &z);
 
+protected:
+	void print_numeric(const print_context & c, const char *par_open, const char *par_close, const char *imag_sym, const char *mul_sym, unsigned level) const;
+	void do_print(const print_context & c, unsigned level) const;
+	void do_print_latex(const print_latex & c, unsigned level) const;
+	void do_print_csrc(const print_csrc & c, unsigned level) const;
+	void do_print_csrc_cl_N(const print_csrc_cl_N & c, unsigned level) const;
+	void do_print_tree(const print_tree & c, unsigned level) const;
+	void do_print_python_repr(const print_python_repr & c, unsigned level) const;
+
 // member variables
 
 protected:
-	cln::cl_number value;
+	cln::cl_N value;
 };
 
+
 // global constants
 
 extern const numeric I;
 extern _numeric_digits Digits;
 
-// deprecated macro, for internal use only
-#define is_a_numeric_hash(x) ((x)&0x80000000U)
-
 // global functions
 
 const numeric exp(const numeric &x);
@@ -282,9 +302,9 @@ inline const numeric denom(const numeric &x)
 
 // numeric evaluation functions for class constant objects:
 
-ex PiEvalf(void);
-ex EulerEvalf(void);
-ex CatalanEvalf(void);
+ex PiEvalf();
+ex EulerEvalf();
+ex CatalanEvalf();
 
 
 // utility functions
