@@ -3,7 +3,7 @@
  *  Implementation of GiNaC's symbolic exponentiation (basis^exponent). */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@
 #include <vector>
 #include <iostream>
 #include <stdexcept>
+#include <limits>
 
 #include "power.h"
 #include "expairseq.h"
@@ -31,36 +32,34 @@
 #include "ncmul.h"
 #include "numeric.h"
 #include "constant.h"
+#include "operators.h"
 #include "inifcns.h" // for log() in power::derivative()
 #include "matrix.h"
+#include "indexed.h"
 #include "symbol.h"
-#include "print.h"
+#include "lst.h"
 #include "archive.h"
 #include "utils.h"
 
 namespace GiNaC {
 
-GINAC_IMPLEMENT_REGISTERED_CLASS(power, basic)
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(power, basic,
+  print_func<print_dflt>(&power::do_print_dflt).
+  print_func<print_latex>(&power::do_print_latex).
+  print_func<print_csrc>(&power::do_print_csrc).
+  print_func<print_python>(&power::do_print_python).
+  print_func<print_python_repr>(&power::do_print_python_repr))
 
 typedef std::vector<int> intvector;
 
 //////////
-// default ctor, dtor, copy ctor, assignment operator and helpers
+// default constructor
 //////////
 
 power::power() : inherited(TINFO_power) { }
 
-void power::copy(const power & other)
-{
-	inherited::copy(other);
-	basis = other.basis;
-	exponent = other.exponent;
-}
-
-DEFAULT_DESTROY(power)
-
 //////////
-// other ctors
+// other constructors
 //////////
 
 // all inlined
@@ -69,7 +68,7 @@ DEFAULT_DESTROY(power)
 // archiving
 //////////
 
-power::power(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+power::power(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	n.find_ex("basis", basis, sym_lst);
 	n.find_ex("exponent", exponent, sym_lst);
@@ -90,11 +89,58 @@ DEFAULT_UNARCHIVE(power)
 
 // public
 
+void power::print_power(const print_context & c, const char *powersymbol, const char *openbrace, const char *closebrace, unsigned level) const
+{
+	// Ordinary output of powers using '^' or '**'
+	if (precedence() <= level)
+		c.s << openbrace << '(';
+	basis.print(c, precedence());
+	c.s << powersymbol;
+	c.s << openbrace;
+	exponent.print(c, precedence());
+	c.s << closebrace;
+	if (precedence() <= level)
+		c.s << ')' << closebrace;
+}
+
+void power::do_print_dflt(const print_dflt & c, unsigned level) const
+{
+	if (exponent.is_equal(_ex1_2)) {
+
+		// Square roots are printed in a special way
+		c.s << "sqrt(";
+		basis.print(c);
+		c.s << ')';
+
+	} else
+		print_power(c, "^", "", "", level);
+}
+
+void power::do_print_latex(const print_latex & c, unsigned level) const
+{
+	if (is_exactly_a<numeric>(exponent) && ex_to<numeric>(exponent).is_negative()) {
+
+		// Powers with negative numeric exponents are printed as fractions
+		c.s << "\\frac{1}{";
+		power(basis, -exponent).eval().print(c);
+		c.s << '}';
+
+	} else if (exponent.is_equal(_ex1_2)) {
+
+		// Square roots are printed in a special way
+		c.s << "\\sqrt{";
+		basis.print(c);
+		c.s << '}';
+
+	} else
+		print_power(c, "^", "{", "}", level);
+}
+
 static void print_sym_pow(const print_context & c, const symbol &x, int exp)
 {
 	// Optimal output of integer powers of symbols to aid compiler CSE.
 	// C.f. ISO/IEC 14882:1998, section 1.9 [intro execution], paragraph 15
-	// to learn why such a parenthisation is really necessary.
+	// to learn why such a parenthesation is really necessary.
 	if (exp == 1) {
 		x.print(c);
 	} else if (exp == 2) {
@@ -114,98 +160,60 @@ static void print_sym_pow(const print_context & c, const symbol &x, int exp)
 	}
 }
 
-void power::print(const print_context & c, unsigned level) const
+void power::do_print_csrc(const print_csrc & c, unsigned level) const
 {
-	if (is_a<print_tree>(c)) {
-
-		inherited::print(c, level);
-
-	} else if (is_a<print_csrc>(c)) {
-
-		// Integer powers of symbols are printed in a special, optimized way
-		if (exponent.info(info_flags::integer)
-		 && (is_exactly_a<symbol>(basis) || is_exactly_a<constant>(basis))) {
-			int exp = ex_to<numeric>(exponent).to_int();
-			if (exp > 0)
-				c.s << '(';
-			else {
-				exp = -exp;
-				if (is_a<print_csrc_cl_N>(c))
-					c.s << "recip(";
-				else
-					c.s << "1.0/(";
-			}
-			print_sym_pow(c, ex_to<symbol>(basis), exp);
-			c.s << ')';
-
-		// <expr>^-1 is printed as "1.0/<expr>" or with the recip() function of CLN
-		} else if (exponent.is_equal(_ex_1)) {
+	// Integer powers of symbols are printed in a special, optimized way
+	if (exponent.info(info_flags::integer)
+	 && (is_a<symbol>(basis) || is_a<constant>(basis))) {
+		int exp = ex_to<numeric>(exponent).to_int();
+		if (exp > 0)
+			c.s << '(';
+		else {
+			exp = -exp;
 			if (is_a<print_csrc_cl_N>(c))
 				c.s << "recip(";
 			else
 				c.s << "1.0/(";
-			basis.print(c);
-			c.s << ')';
-
-		// Otherwise, use the pow() or expt() (CLN) functions
-		} else {
-			if (is_a<print_csrc_cl_N>(c))
-				c.s << "expt(";
-			else
-				c.s << "pow(";
-			basis.print(c);
-			c.s << ',';
-			exponent.print(c);
-			c.s << ')';
 		}
+		print_sym_pow(c, ex_to<symbol>(basis), exp);
+		c.s << ')';
 
-	} else if (is_a<print_python_repr>(c)) {
+	// <expr>^-1 is printed as "1.0/<expr>" or with the recip() function of CLN
+	} else if (exponent.is_equal(_ex_1)) {
+		if (is_a<print_csrc_cl_N>(c))
+			c.s << "recip(";
+		else
+			c.s << "1.0/(";
+		basis.print(c);
+		c.s << ')';
 
-		c.s << class_name() << '(';
+	// Otherwise, use the pow() or expt() (CLN) functions
+	} else {
+		if (is_a<print_csrc_cl_N>(c))
+			c.s << "expt(";
+		else
+			c.s << "pow(";
 		basis.print(c);
 		c.s << ',';
 		exponent.print(c);
 		c.s << ')';
-
-	} else {
-
-		if (exponent.is_equal(_ex1_2)) {
-			if (is_a<print_latex>(c))
-				c.s << "\\sqrt{";
-			else
-				c.s << "sqrt(";
-			basis.print(c);
-			if (is_a<print_latex>(c))
-				c.s << '}';
-			else
-				c.s << ')';
-		} else {
-			if (precedence() <= level) {
-				if (is_a<print_latex>(c))
-					c.s << "{(";
-				else
-					c.s << "(";
-			}
-			basis.print(c, precedence());
-			if (is_a<print_python>(c))
-				c.s << "**";
-			else
-				c.s << '^';
-			if (is_a<print_latex>(c))
-				c.s << '{';
-			exponent.print(c, precedence());
-			if (is_a<print_latex>(c))
-				c.s << '}';
-			if (precedence() <= level) {
-				if (is_a<print_latex>(c))
-					c.s << ")}";
-				else
-					c.s << ')';
-			}
-		}
 	}
 }
 
+void power::do_print_python(const print_python & c, unsigned level) const
+{
+	print_power(c, "**", "", "", level);
+}
+
+void power::do_print_python_repr(const print_python_repr & c, unsigned level) const
+{
+	c.s << class_name() << '(';
+	basis.print(c);
+	c.s << ',';
+	exponent.print(c);
+	c.s << ')';
+}
+
 bool power::info(unsigned inf) const
 {
 	switch (inf) {
@@ -224,14 +232,13 @@ bool power::info(unsigned inf) const
 	return inherited::info(inf);
 }
 
-unsigned power::nops() const
+size_t power::nops() const
 {
 	return 2;
 }
 
-ex & power::let_op(int i)
+ex power::op(size_t i) const
 {
-	GINAC_ASSERT(i>=0);
 	GINAC_ASSERT(i<2);
 
 	return i==0 ? basis : exponent;
@@ -246,7 +253,7 @@ int power::degree(const ex & s) const
 {
 	if (is_equal(ex_to<basic>(s)))
 		return 1;
-	else if (is_ex_exactly_of_type(exponent, numeric) && ex_to<numeric>(exponent).is_integer()) {
+	else if (is_exactly_a<numeric>(exponent) && ex_to<numeric>(exponent).is_integer()) {
 		if (basis.is_equal(s))
 			return ex_to<numeric>(exponent).to_int();
 		else
@@ -261,7 +268,7 @@ int power::ldegree(const ex & s) const
 {
 	if (is_equal(ex_to<basic>(s)))
 		return 1;
-	else if (is_ex_exactly_of_type(exponent, numeric) && ex_to<numeric>(exponent).is_integer()) {
+	else if (is_exactly_a<numeric>(exponent) && ex_to<numeric>(exponent).is_integer()) {
 		if (basis.is_equal(s))
 			return ex_to<numeric>(exponent).to_int();
 		else
@@ -284,7 +291,7 @@ ex power::coeff(const ex & s, int n) const
 			return _ex0;
 	} else {
 		// basis equal to s
-		if (is_ex_exactly_of_type(exponent, numeric) && ex_to<numeric>(exponent).is_integer()) {
+		if (is_exactly_a<numeric>(exponent) && ex_to<numeric>(exponent).is_integer()) {
 			// integer exponent
 			int int_exp = ex_to<numeric>(exponent).to_int();
 			if (n == int_exp)
@@ -330,11 +337,11 @@ ex power::eval(int level) const
 	const numeric *num_basis;
 	const numeric *num_exponent;
 	
-	if (is_ex_exactly_of_type(ebasis, numeric)) {
+	if (is_exactly_a<numeric>(ebasis)) {
 		basis_is_numerical = true;
 		num_basis = &ex_to<numeric>(ebasis);
 	}
-	if (is_ex_exactly_of_type(eexponent, numeric)) {
+	if (is_exactly_a<numeric>(eexponent)) {
 		exponent_is_numerical = true;
 		num_exponent = &ex_to<numeric>(eexponent);
 	}
@@ -425,11 +432,11 @@ ex power::eval(int level) const
 		// ^(^(x,c1),c2) -> ^(x,c1*c2)
 		// (c1, c2 numeric(), c2 integer or -1 < c1 <= 1,
 		// case c1==1 should not happen, see below!)
-		if (is_ex_exactly_of_type(ebasis,power)) {
+		if (is_exactly_a<power>(ebasis)) {
 			const power & sub_power = ex_to<power>(ebasis);
 			const ex & sub_basis = sub_power.basis;
 			const ex & sub_exponent = sub_power.exponent;
-			if (is_ex_exactly_of_type(sub_exponent,numeric)) {
+			if (is_exactly_a<numeric>(sub_exponent)) {
 				const numeric & num_sub_exponent = ex_to<numeric>(sub_exponent);
 				GINAC_ASSERT(num_sub_exponent!=numeric(1));
 				if (num_exponent->is_integer() || (abs(num_sub_exponent) - _num1).is_negative())
@@ -438,13 +445,13 @@ ex power::eval(int level) const
 		}
 	
 		// ^(*(x,y,z),c1) -> *(x^c1,y^c1,z^c1) (c1 integer)
-		if (num_exponent->is_integer() && is_ex_exactly_of_type(ebasis,mul)) {
-			return expand_mul(ex_to<mul>(ebasis), *num_exponent);
+		if (num_exponent->is_integer() && is_exactly_a<mul>(ebasis)) {
+			return expand_mul(ex_to<mul>(ebasis), *num_exponent, 0);
 		}
 	
 		// ^(*(...,x;c1),c2) -> *(^(*(...,x;1),c2),c1^c2)  (c1, c2 numeric(), c1>0)
 		// ^(*(...,x;c1),c2) -> *(^(*(...,x;-1),c2),(-c1)^c2)  (c1, c2 numeric(), c1<0)
-		if (is_ex_exactly_of_type(ebasis,mul)) {
+		if (is_exactly_a<mul>(ebasis)) {
 			GINAC_ASSERT(!num_exponent->is_integer()); // should have been handled above
 			const mul & mulref = ex_to<mul>(ebasis);
 			if (!mulref.overall_coeff.is_equal(_ex1)) {
@@ -475,7 +482,7 @@ ex power::eval(int level) const
 		// ^(nc,c1) -> ncmul(nc,nc,...) (c1 positive integer, unless nc is a matrix)
 		if (num_exponent->is_pos_integer() &&
 		    ebasis.return_type() != return_types::commutative &&
-		    !is_ex_of_type(ebasis,matrix)) {
+		    !is_a<matrix>(ebasis)) {
 			return ncmul(exvector(num_exponent->to_int(), ebasis), true);
 		}
 	}
@@ -509,33 +516,56 @@ ex power::evalf(int level) const
 	return power(ebasis,eexponent);
 }
 
-ex power::evalm(void) const
+ex power::evalm() const
 {
 	const ex ebasis = basis.evalm();
 	const ex eexponent = exponent.evalm();
-	if (is_ex_of_type(ebasis,matrix)) {
-		if (is_ex_of_type(eexponent,numeric)) {
+	if (is_a<matrix>(ebasis)) {
+		if (is_exactly_a<numeric>(eexponent)) {
 			return (new matrix(ex_to<matrix>(ebasis).pow(eexponent)))->setflag(status_flags::dynallocated);
 		}
 	}
 	return (new power(ebasis, eexponent))->setflag(status_flags::dynallocated);
 }
 
-ex power::subs(const lst & ls, const lst & lr, bool no_pattern) const
-{
-	const ex &subsed_basis = basis.subs(ls, lr, no_pattern);
-	const ex &subsed_exponent = exponent.subs(ls, lr, no_pattern);
+// from mul.cpp
+extern bool tryfactsubs(const ex &, const ex &, int &, lst &);
 
-	if (are_ex_trivially_equal(basis, subsed_basis)
-	 && are_ex_trivially_equal(exponent, subsed_exponent))
-		return basic::subs(ls, lr, no_pattern);
-	else
-		return power(subsed_basis, subsed_exponent).basic::subs(ls, lr, no_pattern);
+ex power::subs(const exmap & m, unsigned options) const
+{	
+	const ex &subsed_basis = basis.subs(m, options);
+	const ex &subsed_exponent = exponent.subs(m, options);
+
+	if (!are_ex_trivially_equal(basis, subsed_basis)
+	 || !are_ex_trivially_equal(exponent, subsed_exponent)) 
+		return power(subsed_basis, subsed_exponent).subs_one_level(m, options);
+
+	if (!(options & subs_options::algebraic))
+		return subs_one_level(m, options);
+
+	for (exmap::const_iterator it = m.begin(); it != m.end(); ++it) {
+		int nummatches = std::numeric_limits<int>::max();
+		lst repls;
+		if (tryfactsubs(*this, it->first, nummatches, repls))
+			return (ex_to<basic>((*this) * power(it->second.subs(ex(repls), subs_options::no_pattern) / it->first.subs(ex(repls), subs_options::no_pattern), nummatches))).subs_one_level(m, options);
+	}
+
+	return subs_one_level(m, options);
 }
 
-ex power::simplify_ncmul(const exvector & v) const
+ex power::eval_ncmul(const exvector & v) const
 {
-	return inherited::simplify_ncmul(v);
+	return inherited::eval_ncmul(v);
+}
+
+ex power::conjugate() const
+{
+	ex newbasis = basis.conjugate();
+	ex newexponent = exponent.conjugate();
+	if (are_ex_trivially_equal(basis, newbasis) && are_ex_trivially_equal(exponent, newexponent)) {
+		return *this;
+	}
+	return (new power(newbasis, newexponent))->setflag(status_flags::dynallocated);
 }
 
 // protected
@@ -571,12 +601,12 @@ int power::compare_same_type(const basic & other) const
 		return exponent.compare(o.exponent);
 }
 
-unsigned power::return_type(void) const
+unsigned power::return_type() const
 {
 	return basis.return_type();
 }
    
-unsigned power::return_type_tinfo(void) const
+unsigned power::return_type_tinfo() const
 {
 	return basis.return_type_tinfo();
 }
@@ -590,7 +620,7 @@ ex power::expand(unsigned options) const
 	const ex expanded_exponent = exponent.expand(options);
 	
 	// x^(a+b) -> x^a * x^b
-	if (is_ex_exactly_of_type(expanded_exponent, add)) {
+	if (is_exactly_a<add>(expanded_exponent)) {
 		const add &a = ex_to<add>(expanded_exponent);
 		exvector distrseq;
 		distrseq.reserve(a.seq.size() + 1);
@@ -605,8 +635,8 @@ ex power::expand(unsigned options) const
 		if (ex_to<numeric>(a.overall_coeff).is_integer()) {
 			const numeric &num_exponent = ex_to<numeric>(a.overall_coeff);
 			int int_exponent = num_exponent.to_int();
-			if (int_exponent > 0 && is_ex_exactly_of_type(expanded_basis, add))
-				distrseq.push_back(expand_add(ex_to<add>(expanded_basis), int_exponent));
+			if (int_exponent > 0 && is_exactly_a<add>(expanded_basis))
+				distrseq.push_back(expand_add(ex_to<add>(expanded_basis), int_exponent, options));
 			else
 				distrseq.push_back(power(expanded_basis, a.overall_coeff));
 		} else
@@ -614,10 +644,10 @@ ex power::expand(unsigned options) const
 		
 		// Make sure that e.g. (x+y)^(1+a) -> x*(x+y)^a + y*(x+y)^a
 		ex r = (new mul(distrseq))->setflag(status_flags::dynallocated);
-		return r.expand();
+		return r.expand(options);
 	}
 	
-	if (!is_ex_exactly_of_type(expanded_exponent, numeric) ||
+	if (!is_exactly_a<numeric>(expanded_exponent) ||
 		!ex_to<numeric>(expanded_exponent).is_integer()) {
 		if (are_ex_trivially_equal(basis,expanded_basis) && are_ex_trivially_equal(exponent,expanded_exponent)) {
 			return this->hold();
@@ -631,12 +661,12 @@ ex power::expand(unsigned options) const
 	int int_exponent = num_exponent.to_int();
 	
 	// (x+y)^n, n>0
-	if (int_exponent > 0 && is_ex_exactly_of_type(expanded_basis,add))
-		return expand_add(ex_to<add>(expanded_basis), int_exponent);
+	if (int_exponent > 0 && is_exactly_a<add>(expanded_basis))
+		return expand_add(ex_to<add>(expanded_basis), int_exponent, options);
 	
 	// (x*y)^n -> x^n * y^n
-	if (is_ex_exactly_of_type(expanded_basis,mul))
-		return expand_mul(ex_to<mul>(expanded_basis), num_exponent);
+	if (is_exactly_a<mul>(expanded_basis))
+		return expand_mul(ex_to<mul>(expanded_basis), num_exponent, options, true);
 	
 	// cannot expand further
 	if (are_ex_trivially_equal(basis,expanded_basis) && are_ex_trivially_equal(exponent,expanded_exponent))
@@ -657,12 +687,12 @@ ex power::expand(unsigned options) const
 
 /** expand a^n where a is an add and n is a positive integer.
  *  @see power::expand */
-ex power::expand_add(const add & a, int n) const
+ex power::expand_add(const add & a, int n, unsigned options) const
 {
 	if (n==2)
-		return expand_add_2(a);
+		return expand_add_2(a, options);
 
-	const int m = a.nops();
+	const size_t m = a.nops();
 	exvector result;
 	// The number of terms will be the number of combinatorial compositions,
 	// i.e. the number of unordered arrangement of m nonnegative integers
@@ -674,7 +704,7 @@ ex power::expand_add(const add & a, int n) const
 	intvector upper_limit(m-1);
 	int l;
 
-	for (int l=0; l<m-1; ++l) {
+	for (size_t l=0; l<m-1; ++l) {
 		k[l] = 0;
 		k_cum[l] = 0;
 		upper_limit[l] = n;
@@ -692,8 +722,8 @@ ex power::expand_add(const add & a, int n) const
 			             !is_exactly_a<add>(ex_to<power>(b).basis) ||
 			             !is_exactly_a<mul>(ex_to<power>(b).basis) ||
 			             !is_exactly_a<power>(ex_to<power>(b).basis));
-			if (is_ex_exactly_of_type(b,mul))
-				term.push_back(expand_mul(ex_to<mul>(b),numeric(k[l])));
+			if (is_exactly_a<mul>(b))
+				term.push_back(expand_mul(ex_to<mul>(b), numeric(k[l]), options, true));
 			else
 				term.push_back(power(b,k[l]));
 		}
@@ -706,8 +736,8 @@ ex power::expand_add(const add & a, int n) const
 		             !is_exactly_a<add>(ex_to<power>(b).basis) ||
 		             !is_exactly_a<mul>(ex_to<power>(b).basis) ||
 		             !is_exactly_a<power>(ex_to<power>(b).basis));
-		if (is_ex_exactly_of_type(b,mul))
-			term.push_back(expand_mul(ex_to<mul>(b),numeric(n-k_cum[m-2])));
+		if (is_exactly_a<mul>(b))
+			term.push_back(expand_mul(ex_to<mul>(b), numeric(n-k_cum[m-2]), options, true));
 		else
 			term.push_back(power(b,n-k_cum[m-2]));
 
@@ -717,7 +747,7 @@ ex power::expand_add(const add & a, int n) const
 
 		term.push_back(f);
 
-		result.push_back((new mul(term))->setflag(status_flags::dynallocated));
+		result.push_back(ex((new mul(term))->setflag(status_flags::dynallocated)).expand(options));
 
 		// increment k[]
 		l = m-2;
@@ -730,10 +760,10 @@ ex power::expand_add(const add & a, int n) const
 		// recalc k_cum[] and upper_limit[]
 		k_cum[l] = (l==0 ? k[0] : k_cum[l-1]+k[l]);
 
-		for (int i=l+1; i<m-1; ++i)
+		for (size_t i=l+1; i<m-1; ++i)
 			k_cum[i] = k_cum[i-1]+k[i];
 
-		for (int i=l+1; i<m-1; ++i)
+		for (size_t i=l+1; i<m-1; ++i)
 			upper_limit[i] = n-k_cum[i-1];
 	}
 
@@ -744,10 +774,10 @@ ex power::expand_add(const add & a, int n) const
 
 /** Special case of power::expand_add. Expands a^2 where a is an add.
  *  @see power::expand_add */
-ex power::expand_add_2(const add & a) const
+ex power::expand_add_2(const add & a, unsigned options) const
 {
 	epvector sum;
-	unsigned a_nops = a.nops();
+	size_t a_nops = a.nops();
 	sum.reserve((a_nops*(a_nops+1))/2);
 	epvector::const_iterator last = a.seq.end();
 
@@ -765,24 +795,24 @@ ex power::expand_add_2(const add & a) const
 		             !is_exactly_a<mul>(ex_to<power>(r).basis) ||
 		             !is_exactly_a<power>(ex_to<power>(r).basis));
 		
-		if (are_ex_trivially_equal(c,_ex1)) {
-			if (is_ex_exactly_of_type(r,mul)) {
-				sum.push_back(expair(expand_mul(ex_to<mul>(r),_num2),
+		if (c.is_equal(_ex1)) {
+			if (is_exactly_a<mul>(r)) {
+				sum.push_back(expair(expand_mul(ex_to<mul>(r), _num2, options, true),
 				                     _ex1));
 			} else {
 				sum.push_back(expair((new power(r,_ex2))->setflag(status_flags::dynallocated),
 				                     _ex1));
 			}
 		} else {
-			if (is_ex_exactly_of_type(r,mul)) {
-				sum.push_back(expair(expand_mul(ex_to<mul>(r),_num2),
+			if (is_exactly_a<mul>(r)) {
+				sum.push_back(a.combine_ex_with_coeff_to_pair(expand_mul(ex_to<mul>(r), _num2, options, true),
 				                     ex_to<numeric>(c).power_dyn(_num2)));
 			} else {
-				sum.push_back(expair((new power(r,_ex2))->setflag(status_flags::dynallocated),
+				sum.push_back(a.combine_ex_with_coeff_to_pair((new power(r,_ex2))->setflag(status_flags::dynallocated),
 				                     ex_to<numeric>(c).power_dyn(_num2)));
 			}
 		}
-			
+
 		for (epvector::const_iterator cit1=cit0+1; cit1!=last; ++cit1) {
 			const ex & r1 = cit1->rest;
 			const ex & c1 = cit1->coeff;
@@ -810,7 +840,7 @@ ex power::expand_add_2(const add & a) const
 
 /** Expand factors of m in m^n where m is a mul and n is and integer.
  *  @see power::expand */
-ex power::expand_mul(const mul & m, const numeric & n) const
+ex power::expand_mul(const mul & m, const numeric & n, unsigned options, bool from_expand) const
 {
 	GINAC_ASSERT(n.is_integer());
 
@@ -819,19 +849,33 @@ ex power::expand_mul(const mul & m, const numeric & n) const
 
 	epvector distrseq;
 	distrseq.reserve(m.seq.size());
+	bool need_reexpand = false;
+
 	epvector::const_iterator last = m.seq.end();
 	epvector::const_iterator cit = m.seq.begin();
 	while (cit!=last) {
-		if (is_ex_exactly_of_type((*cit).rest,numeric)) {
-			distrseq.push_back(m.combine_pair_with_coeff_to_pair(*cit,n));
+		if (is_exactly_a<numeric>(cit->rest)) {
+			distrseq.push_back(m.combine_pair_with_coeff_to_pair(*cit, n));
 		} else {
 			// it is safe not to call mul::combine_pair_with_coeff_to_pair()
 			// since n is an integer
-			distrseq.push_back(expair((*cit).rest, ex_to<numeric>((*cit).coeff).mul(n)));
+			numeric new_coeff = ex_to<numeric>(cit->coeff).mul(n);
+			if (from_expand && is_exactly_a<add>(cit->rest) && new_coeff.is_pos_integer()) {
+				// this happens when e.g. (a+b)^(1/2) gets squared and
+				// the resulting product needs to be reexpanded
+				need_reexpand = true;
+			}
+			distrseq.push_back(expair(cit->rest, new_coeff));
 		}
 		++cit;
 	}
-	return (new mul(distrseq,ex_to<numeric>(m.overall_coeff).power_dyn(n)))->setflag(status_flags::dynallocated);
+
+	const mul & result = static_cast<const mul &>((new mul(distrseq, ex_to<numeric>(m.overall_coeff).power_dyn(n)))->setflag(status_flags::dynallocated));
+	if (need_reexpand)
+		return ex(result).expand(options);
+	if (from_expand)
+		return result.setflag(status_flags::expanded);
+	return result;
 }
 
 } // namespace GiNaC
