@@ -4,7 +4,7 @@
  *  methods for series expansion. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include <iostream>
+#include <numeric>
 #include <stdexcept>
 
 #include "pseries.h"
@@ -31,32 +31,27 @@
 #include "mul.h"
 #include "power.h"
 #include "relational.h"
+#include "operators.h"
 #include "symbol.h"
-#include "print.h"
 #include "archive.h"
 #include "utils.h"
 
 namespace GiNaC {
 
-GINAC_IMPLEMENT_REGISTERED_CLASS(pseries, basic)
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(pseries, basic,
+  print_func<print_context>(&pseries::do_print).
+  print_func<print_latex>(&pseries::do_print_latex).
+  print_func<print_tree>(&pseries::do_print_tree).
+  print_func<print_python>(&pseries::do_print_python).
+  print_func<print_python_repr>(&pseries::do_print_python_repr))
 
 
 /*
- *  Default ctor, dtor, copy ctor, assignment operator and helpers
+ *  Default constructor
  */
 
 pseries::pseries() : inherited(TINFO_pseries) { }
 
-void pseries::copy(const pseries &other)
-{
-	inherited::copy(other);
-	seq = other.seq;
-	var = other.var;
-	point = other.point;
-}
-
-DEFAULT_DESTROY(pseries)
-
 
 /*
  *  Other ctors
@@ -73,8 +68,8 @@ DEFAULT_DESTROY(pseries)
  *  @return newly constructed pseries */
 pseries::pseries(const ex &rel_, const epvector &ops_) : basic(TINFO_pseries), seq(ops_)
 {
-	GINAC_ASSERT(is_exactly_a<relational>(rel_));
-	GINAC_ASSERT(is_exactly_a<symbol>(rel_.lhs()));
+	GINAC_ASSERT(is_a<relational>(rel_));
+	GINAC_ASSERT(is_a<symbol>(rel_.lhs()));
 	point = rel_.rhs();
 	var = rel_.lhs();
 }
@@ -84,7 +79,7 @@ pseries::pseries(const ex &rel_, const epvector &ops_) : basic(TINFO_pseries), s
  *  Archiving
  */
 
-pseries::pseries(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+pseries::pseries(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	for (unsigned int i=0; true; ++i) {
 		ex rest;
@@ -117,106 +112,113 @@ DEFAULT_UNARCHIVE(pseries)
 // functions overriding virtual functions from base classes
 //////////
 
-void pseries::print(const print_context & c, unsigned level) const
+void pseries::print_series(const print_context & c, const char *openbrace, const char *closebrace, const char *mul_sym, const char *pow_sym, unsigned level) const
 {
-	if (is_a<print_tree>(c)) {
-
-		c.s << std::string(level, ' ') << class_name()
-		    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
-		    << std::endl;
-		unsigned delta_indent = static_cast<const print_tree &>(c).delta_indent;
-		unsigned num = seq.size();
-		for (unsigned i=0; i<num; ++i) {
-			seq[i].rest.print(c, level + delta_indent);
-			seq[i].coeff.print(c, level + delta_indent);
-			c.s << std::string(level + delta_indent, ' ') << "-----" << std::endl;
-		}
-		var.print(c, level + delta_indent);
-		point.print(c, level + delta_indent);
-
-	} else if (is_a<print_python_repr>(c)) {
-		c.s << class_name() << "(relational(";
-		var.print(c);
-		c.s << ',';
-		point.print(c);
-		c.s << "),[";
-		unsigned num = seq.size();
-		for (unsigned i=0; i<num; ++i) {
-			if (i)
-				c.s << ',';
-			c.s << '(';
-			seq[i].rest.print(c);
-			c.s << ',';
-			seq[i].coeff.print(c);
-			c.s << ')';
-		}
-		c.s << "])";
-	} else {
-
-		if (precedence() <= level)
-			c.s << "(";
+	if (precedence() <= level)
+		c.s << '(';
 		
-		std::string par_open = is_a<print_latex>(c) ? "{(" : "(";
-		std::string par_close = is_a<print_latex>(c) ? ")}" : ")";
-		
-		// objects of type pseries must not have any zero entries, so the
-		// trivial (zero) pseries needs a special treatment here:
-		if (seq.empty())
-			c.s << '0';
-		epvector::const_iterator i = seq.begin(), end = seq.end();
-		while (i != end) {
-			// print a sign, if needed
-			if (i != seq.begin())
-				c.s << '+';
-			if (!is_order_function(i->rest)) {
-				// print 'rest', i.e. the expansion coefficient
-				if (i->rest.info(info_flags::numeric) &&
-					i->rest.info(info_flags::positive)) {
-					i->rest.print(c);
-				} else {
-					c.s << par_open;
-					i->rest.print(c);
-					c.s << par_close;
-				}
-				// print 'coeff', something like (x-1)^42
-				if (!i->coeff.is_zero()) {
-					if (is_a<print_latex>(c))
-						c.s << ' ';
-					else
-						c.s << '*';
-					if (!point.is_zero()) {
-						c.s << par_open;
-						(var-point).print(c);
-						c.s << par_close;
+	// objects of type pseries must not have any zero entries, so the
+	// trivial (zero) pseries needs a special treatment here:
+	if (seq.empty())
+		c.s << '0';
+
+	epvector::const_iterator i = seq.begin(), end = seq.end();
+	while (i != end) {
+
+		// print a sign, if needed
+		if (i != seq.begin())
+			c.s << '+';
+
+		if (!is_order_function(i->rest)) {
+
+			// print 'rest', i.e. the expansion coefficient
+			if (i->rest.info(info_flags::numeric) &&
+				i->rest.info(info_flags::positive)) {
+				i->rest.print(c);
+			} else {
+				c.s << openbrace << '(';
+				i->rest.print(c);
+				c.s << ')' << closebrace;
+			}
+
+			// print 'coeff', something like (x-1)^42
+			if (!i->coeff.is_zero()) {
+				c.s << mul_sym;
+				if (!point.is_zero()) {
+					c.s << openbrace << '(';
+					(var-point).print(c);
+					c.s << ')' << closebrace;
+				} else
+					var.print(c);
+				if (i->coeff.compare(_ex1)) {
+					c.s << pow_sym;
+					c.s << openbrace;
+					if (i->coeff.info(info_flags::negative)) {
+						c.s << '(';
+						i->coeff.print(c);
+						c.s << ')';
 					} else
-						var.print(c);
-					if (i->coeff.compare(_ex1)) {
-						if (is_a<print_python>(c))
-							c.s << "**";
-						else
-							c.s << '^';
-						if (i->coeff.info(info_flags::negative)) {
-							c.s << par_open;
-							i->coeff.print(c);
-							c.s << par_close;
-						} else {
-							if (is_a<print_latex>(c)) {
-								c.s << '{';
-								i->coeff.print(c);
-								c.s << '}';
-							} else
-								i->coeff.print(c);
-						}
-					}
+						i->coeff.print(c);
+					c.s << closebrace;
 				}
-			} else
-				Order(power(var-point,i->coeff)).print(c);
-			++i;
-		}
-
-		if (precedence() <= level)
-			c.s << ")";
+			}
+		} else
+			Order(power(var-point,i->coeff)).print(c);
+		++i;
 	}
+
+	if (precedence() <= level)
+		c.s << ')';
+}
+
+void pseries::do_print(const print_context & c, unsigned level) const
+{
+	print_series(c, "", "", "*", "^", level);
+}
+
+void pseries::do_print_latex(const print_latex & c, unsigned level) const
+{
+	print_series(c, "{", "}", " ", "^", level);
+}
+
+void pseries::do_print_python(const print_python & c, unsigned level) const
+{
+	print_series(c, "", "", "*", "**", level);
+}
+
+void pseries::do_print_tree(const print_tree & c, unsigned level) const
+{
+	c.s << std::string(level, ' ') << class_name() << " @" << this
+	    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
+	    << std::endl;
+	size_t num = seq.size();
+	for (size_t i=0; i<num; ++i) {
+		seq[i].rest.print(c, level + c.delta_indent);
+		seq[i].coeff.print(c, level + c.delta_indent);
+		c.s << std::string(level + c.delta_indent, ' ') << "-----" << std::endl;
+	}
+	var.print(c, level + c.delta_indent);
+	point.print(c, level + c.delta_indent);
+}
+
+void pseries::do_print_python_repr(const print_python_repr & c, unsigned level) const
+{
+	c.s << class_name() << "(relational(";
+	var.print(c);
+	c.s << ',';
+	point.print(c);
+	c.s << "),[";
+	size_t num = seq.size();
+	for (size_t i=0; i<num; ++i) {
+		if (i)
+			c.s << ',';
+		c.s << '(';
+		seq[i].rest.print(c);
+		c.s << ',';
+		seq[i].coeff.print(c);
+		c.s << ')';
+	}
+	c.s << "])";
 }
 
 int pseries::compare_same_type(const basic & other) const
@@ -253,22 +255,18 @@ int pseries::compare_same_type(const basic & other) const
 }
 
 /** Return the number of operands including a possible order term. */
-unsigned pseries::nops(void) const
+size_t pseries::nops() const
 {
 	return seq.size();
 }
 
 /** Return the ith term in the series when represented as a sum. */
-ex pseries::op(int i) const
+ex pseries::op(size_t i) const
 {
-	if (i < 0 || unsigned(i) >= seq.size())
+	if (i >= seq.size())
 		throw (std::out_of_range("op() out of range"));
-	return seq[i].rest * power(var - point, seq[i].coeff);
-}
 
-ex &pseries::let_op(int i)
-{
-	throw (std::logic_error("let_op not defined for pseries"));
+	return seq[i].rest * power(var - point, seq[i].coeff);
 }
 
 /** Return degree of highest power of the series.  This is usually the exponent
@@ -409,13 +407,30 @@ ex pseries::evalf(int level) const
 	return (new pseries(relational(var,point), new_seq))->setflag(status_flags::dynallocated | status_flags::evaluated);
 }
 
-ex pseries::subs(const lst & ls, const lst & lr, bool no_pattern) const
+ex pseries::conjugate() const
+{
+	epvector * newseq = conjugateepvector(seq);
+	ex newvar = var.conjugate();
+	ex newpoint = point.conjugate();
+
+	if (!newseq	&& are_ex_trivially_equal(newvar, var) && are_ex_trivially_equal(point, newpoint)) {
+		return *this;
+	}
+
+	ex result = (new pseries(newvar==newpoint, newseq ? *newseq : seq))->setflag(status_flags::dynallocated);
+	if (newseq) {
+		delete newseq;
+	}
+	return result;
+}
+
+ex pseries::subs(const exmap & m, unsigned options) const
 {
 	// If expansion variable is being substituted, convert the series to a
 	// polynomial and do the substitution there because the result might
 	// no longer be a power series
-	if (ls.has(var))
-		return convert_to_poly(true).subs(ls, lr, no_pattern);
+	if (m.find(var) != m.end())
+		return convert_to_poly(true).subs(m, options);
 	
 	// Otherwise construct a new series with substituted coefficients and
 	// expansion point
@@ -423,10 +438,10 @@ ex pseries::subs(const lst & ls, const lst & lr, bool no_pattern) const
 	newseq.reserve(seq.size());
 	epvector::const_iterator it = seq.begin(), itend = seq.end();
 	while (it != itend) {
-		newseq.push_back(expair(it->rest.subs(ls, lr, no_pattern), it->coeff));
+		newseq.push_back(expair(it->rest.subs(m, options), it->coeff));
 		++it;
 	}
-	return (new pseries(relational(var,point.subs(ls, lr, no_pattern)), newseq))->setflag(status_flags::dynallocated);
+	return (new pseries(relational(var,point.subs(m, options)), newseq))->setflag(status_flags::dynallocated);
 }
 
 /** Implementation of ex::expand() for a power series.  It expands all the
@@ -445,14 +460,14 @@ ex pseries::expand(unsigned options) const
 	        ->setflag(status_flags::dynallocated | (options == 0 ? status_flags::expanded : 0));
 }
 
-/** Implementation of ex::diff() for a power series.  It treats the series as a
- *  polynomial.
+/** Implementation of ex::diff() for a power series.
  *  @see ex::diff */
 ex pseries::derivative(const symbol & s) const
 {
+	epvector new_seq;
+	epvector::const_iterator it = seq.begin(), itend = seq.end();
+
 	if (s == var) {
-		epvector new_seq;
-		epvector::const_iterator it = seq.begin(), itend = seq.end();
 		
 		// FIXME: coeff might depend on var
 		while (it != itend) {
@@ -465,10 +480,22 @@ ex pseries::derivative(const symbol & s) const
 			}
 			++it;
 		}
-		return pseries(relational(var,point), new_seq);
+
 	} else {
-		return *this;
+
+		while (it != itend) {
+			if (is_order_function(it->rest)) {
+				new_seq.push_back(*it);
+			} else {
+				ex c = it->rest.diff(s);
+				if (!c.is_zero())
+					new_seq.push_back(expair(c, it->coeff));
+			}
+			++it;
+		}
 	}
+
+	return pseries(relational(var,point), new_seq);
 }
 
 ex pseries::convert_to_poly(bool no_order) const
@@ -487,7 +514,7 @@ ex pseries::convert_to_poly(bool no_order) const
 	return e;
 }
 
-bool pseries::is_terminating(void) const
+bool pseries::is_terminating() const
 {
 	return seq.empty() || !is_order_function((seq.end()-1)->rest);
 }
@@ -502,14 +529,23 @@ bool pseries::is_terminating(void) const
 ex basic::series(const relational & r, int order, unsigned options) const
 {
 	epvector seq;
+	const symbol &s = ex_to<symbol>(r.lhs());
+
+	// default for order-values that make no sense for Taylor expansion
+	if ((order <= 0) && this->has(s)) {
+		seq.push_back(expair(Order(_ex1), order));
+		return pseries(r, seq);
+	}
+
+	// do Taylor expansion
 	numeric fac = 1;
 	ex deriv = *this;
-	ex coeff = deriv.subs(r);
-	const symbol &s = ex_to<symbol>(r.lhs());
-	
-	if (!coeff.is_zero())
+	ex coeff = deriv.subs(r, subs_options::no_pattern);
+
+	if (!coeff.is_zero()) {
 		seq.push_back(expair(coeff, _ex0));
-	
+	}
+
 	int n;
 	for (n=1; n<order; ++n) {
 		fac = fac.mul(n);
@@ -520,7 +556,7 @@ ex basic::series(const relational & r, int order, unsigned options) const
 		if (deriv.is_zero())  // Series terminates
 			return pseries(r, seq);
 
-		coeff = deriv.subs(r);
+		coeff = deriv.subs(r, subs_options::no_pattern);
 		if (!coeff.is_zero())
 			seq.push_back(expair(fac.inverse() * coeff, n));
 	}
@@ -539,7 +575,7 @@ ex symbol::series(const relational & r, int order, unsigned options) const
 {
 	epvector seq;
 	const ex point = r.rhs();
-	GINAC_ASSERT(is_exactly_a<symbol>(r.lhs()));
+	GINAC_ASSERT(is_a<symbol>(r.lhs()));
 
 	if (this->is_equal_same_type(ex_to<symbol>(r.lhs()))) {
 		if (order > 0 && !point.is_zero())
@@ -643,7 +679,7 @@ ex add::series(const relational & r, int order, unsigned options) const
 	epvector::const_iterator itend = seq.end();
 	for (; it!=itend; ++it) {
 		ex op;
-		if (is_ex_exactly_of_type(it->rest, pseries))
+		if (is_exactly_a<pseries>(it->rest))
 			op = it->rest;
 		else
 			op = it->rest.series(r, order, options);
@@ -738,11 +774,58 @@ ex mul::series(const relational & r, int order, unsigned options) const
 {
 	pseries acc; // Series accumulator
 
-	// Multiply with remaining terms
+	GINAC_ASSERT(is_a<symbol>(r.lhs()));
+	const ex& sym = r.lhs();
+		
+	// holds ldegrees of the series of individual factors
+	std::vector<int> ldegrees;
+
+	// find minimal degrees
 	const epvector::const_iterator itbeg = seq.begin();
 	const epvector::const_iterator itend = seq.end();
 	for (epvector::const_iterator it=itbeg; it!=itend; ++it) {
-		ex op = recombine_pair_to_ex(*it).series(r, order, options);
+
+		ex expon = it->coeff;
+		int factor = 1;
+		ex buf;
+		if (expon.info(info_flags::integer)) {
+			buf = it->rest;
+			factor = ex_to<numeric>(expon).to_int();
+		} else {
+			buf = recombine_pair_to_ex(*it);
+		}
+
+		int real_ldegree = 0;
+		try {
+			real_ldegree = buf.expand().ldegree(sym-r.rhs());
+		}
+		catch (std::runtime_error) {}
+
+		if (real_ldegree == 0) {
+			int orderloop = 0;
+			do {
+				orderloop++;
+				real_ldegree = buf.series(r, orderloop, options).ldegree(sym);
+			} while (real_ldegree == orderloop);
+		}
+
+		ldegrees.push_back(factor * real_ldegree);
+	}
+
+	int degsum = std::accumulate(ldegrees.begin(), ldegrees.end(), 0);
+
+	if (degsum>order) {
+		epvector epv;
+		epv.push_back(expair(Order(_ex1), order));
+		return (new pseries(r, epv))->setflag(status_flags::dynallocated);
+	}
+
+	// Multiply with remaining terms
+	std::vector<int>::const_iterator itd = ldegrees.begin();
+	for (epvector::const_iterator it=itbeg; it!=itend; ++it, ++itd) {
+
+		// do series expansion with adjusted order
+		ex op = recombine_pair_to_ex(*it).series(r, order-degsum+(*itd), options);
 
 		// Series multiplication
 		if (it==itbeg)
@@ -750,6 +833,7 @@ ex mul::series(const relational & r, int order, unsigned options) const
 		else
 			acc = ex_to<pseries>(acc.mul_series(ex_to<pseries>(op)));
 	}
+
 	return acc.mul_const(ex_to<numeric>(overall_coeff));
 }
 
@@ -796,6 +880,9 @@ ex pseries::power_const(const numeric &p, int deg) const
 	if (!(p*ldeg).is_integer())
 		throw std::runtime_error("pseries::power_const(): trying to assemble a Puiseux series");
 
+	// adjust number of coefficients
+	deg = deg - p.to_int()*ldeg;
+	
 	// O(x^n)^(-m) is undefined
 	if (seq.size() == 1 && is_order_function(seq[0].rest) && p.real().is_negative())
 		throw pole_error("pseries::power_const(): division by zero",1);
@@ -804,7 +891,6 @@ ex pseries::power_const(const numeric &p, int deg) const
 	exvector co;
 	co.reserve(deg);
 	co.push_back(power(coeff(var, ldeg), p));
-	bool all_sums_zero = true;
 	for (int i=1; i<deg; ++i) {
 		ex sum = _ex0;
 		for (int j=1; j<=i; ++j) {
@@ -815,8 +901,6 @@ ex pseries::power_const(const numeric &p, int deg) const
 			} else
 				sum += (p * j - (i - j)) * co[i - j] * c;
 		}
-		if (!sum.is_zero())
-			all_sums_zero = false;
 		co.push_back(sum / coeff(var, ldeg) / i);
 	}
 	
@@ -831,8 +915,9 @@ ex pseries::power_const(const numeric &p, int deg) const
 			break;
 		}
 	}
-	if (!higher_order && !all_sums_zero)
+	if (!higher_order)
 		new_seq.push_back(expair(Order(_ex1), p * ldeg + deg));
+
 	return pseries(relational(var,point), new_seq);
 }
 
@@ -856,23 +941,23 @@ pseries pseries::shift_exponents(int deg) const
 ex power::series(const relational & r, int order, unsigned options) const
 {
 	// If basis is already a series, just power it
-	if (is_ex_exactly_of_type(basis, pseries))
+	if (is_exactly_a<pseries>(basis))
 		return ex_to<pseries>(basis).power_const(ex_to<numeric>(exponent), order);
 
 	// Basis is not a series, may there be a singularity?
 	bool must_expand_basis = false;
 	try {
-		basis.subs(r);
+		basis.subs(r, subs_options::no_pattern);
 	} catch (pole_error) {
 		must_expand_basis = true;
 	}
-		
+
 	// Is the expression of type something^(-int)?
-	if (!must_expand_basis && !exponent.info(info_flags::negint))
+	if (!must_expand_basis && !exponent.info(info_flags::negint) && !is_a<add>(basis))
 		return basic::series(r, order, options);
-		
+
 	// Is the expression of type 0^something?
-	if (!must_expand_basis && !basis.subs(r).is_zero())
+	if (!must_expand_basis && !basis.subs(r, subs_options::no_pattern).is_zero())
 		return basic::series(r, order, options);
 
 	// Singularity encountered, is the basis equal to (var - point)?
@@ -886,8 +971,32 @@ ex power::series(const relational & r, int order, unsigned options) const
 	}
 
 	// No, expand basis into series
-	ex e = basis.series(r, order, options);
-	return ex_to<pseries>(e).power_const(ex_to<numeric>(exponent), order);
+
+	int intexp = ex_to<numeric>(exponent).to_int();
+	const ex& sym = r.lhs();
+	// find existing minimal degree
+	int real_ldegree = basis.expand().ldegree(sym-r.rhs());
+	if (real_ldegree == 0) {
+		int orderloop = 0;
+		do {
+			orderloop++;
+			real_ldegree = basis.series(r, orderloop, options).ldegree(sym);
+		} while (real_ldegree == orderloop);
+	}
+
+	ex e = basis.series(r, order + real_ldegree*(1-intexp), options);
+	
+	ex result;
+	try {
+		result = ex_to<pseries>(e).power_const(intexp, order);
+	}
+	catch (pole_error) {
+		epvector ser;
+		ser.push_back(expair(Order(_ex1), order));
+		result = pseries(r, ser);
+	}
+
+	return result;
 }
 
 
@@ -895,7 +1004,7 @@ ex power::series(const relational & r, int order, unsigned options) const
 ex pseries::series(const relational & r, int order, unsigned options) const
 {
 	const ex p = r.rhs();
-	GINAC_ASSERT(is_exactly_a<symbol>(r.lhs()));
+	GINAC_ASSERT(is_a<symbol>(r.lhs()));
 	const symbol &s = ex_to<symbol>(r.lhs());
 	
 	if (var.is_equal(s) && point.is_equal(p)) {
@@ -931,13 +1040,12 @@ ex pseries::series(const relational & r, int order, unsigned options) const
  *  @return an expression holding a pseries object */
 ex ex::series(const ex & r, int order, unsigned options) const
 {
-	GINAC_ASSERT(bp!=0);
 	ex e;
 	relational rel_;
 	
-	if (is_ex_exactly_of_type(r,relational))
+	if (is_a<relational>(r))
 		rel_ = ex_to<relational>(r);
-	else if (is_ex_exactly_of_type(r,symbol))
+	else if (is_a<symbol>(r))
 		rel_ = relational(r,_ex0);
 	else
 		throw (std::logic_error("ex::series(): expansion point has unknown type"));
