@@ -3,7 +3,7 @@
  *  Implementation of GiNaC's special tensors. */
 
 /*
- *  GiNaC Copyright (C) 1999-2002 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2004 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -29,33 +29,48 @@
 #include "indexed.h"
 #include "symmetry.h"
 #include "relational.h"
+#include "operators.h"
 #include "lst.h"
 #include "numeric.h"
 #include "matrix.h"
-#include "print.h"
 #include "archive.h"
 #include "utils.h"
 
 namespace GiNaC {
 
 GINAC_IMPLEMENT_REGISTERED_CLASS(tensor, basic)
-GINAC_IMPLEMENT_REGISTERED_CLASS(tensdelta, tensor)
-GINAC_IMPLEMENT_REGISTERED_CLASS(tensmetric, tensor)
-GINAC_IMPLEMENT_REGISTERED_CLASS(minkmetric, tensmetric)
-GINAC_IMPLEMENT_REGISTERED_CLASS(spinmetric, tensmetric)
-GINAC_IMPLEMENT_REGISTERED_CLASS(tensepsilon, tensor)
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(tensdelta, tensor,
+  print_func<print_dflt>(&tensdelta::do_print).
+  print_func<print_latex>(&tensdelta::do_print_latex))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(tensmetric, tensor,
+  print_func<print_dflt>(&tensmetric::do_print).
+  print_func<print_latex>(&tensmetric::do_print))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(minkmetric, tensmetric,
+  print_func<print_dflt>(&minkmetric::do_print).
+  print_func<print_latex>(&minkmetric::do_print_latex))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(spinmetric, tensmetric,
+  print_func<print_dflt>(&spinmetric::do_print).
+  print_func<print_latex>(&spinmetric::do_print_latex))
+
+GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(tensepsilon, tensor,
+  print_func<print_dflt>(&tensepsilon::do_print).
+  print_func<print_latex>(&tensepsilon::do_print_latex))
 
 //////////
-// default ctor, dtor, copy ctor, assignment operator and helpers
+// constructors
 //////////
 
-DEFAULT_CTORS(tensor)
-DEFAULT_CTORS(tensdelta)
-DEFAULT_CTORS(tensmetric)
-DEFAULT_COPY(spinmetric)
-DEFAULT_DESTROY(spinmetric)
-DEFAULT_DESTROY(minkmetric)
-DEFAULT_DESTROY(tensepsilon)
+tensor::tensor() : inherited(TINFO_tensor)
+{
+	setflag(status_flags::evaluated | status_flags::expanded);
+}
+
+DEFAULT_CTOR(tensdelta)
+DEFAULT_CTOR(tensmetric)
 
 minkmetric::minkmetric() : pos_sig(false)
 {
@@ -72,12 +87,6 @@ minkmetric::minkmetric(bool ps) : pos_sig(ps)
 	tinfo_key = TINFO_minkmetric;
 }
 
-void minkmetric::copy(const minkmetric & other)
-{
-	inherited::copy(other);
-	pos_sig = other.pos_sig;
-}
-
 tensepsilon::tensepsilon() : minkowski(false), pos_sig(false)
 {
 	tinfo_key = TINFO_tensepsilon;
@@ -88,13 +97,6 @@ tensepsilon::tensepsilon(bool mink, bool ps) : minkowski(mink), pos_sig(ps)
 	tinfo_key = TINFO_tensepsilon;
 }
 
-void tensepsilon::copy(const tensepsilon & other)
-{
-	inherited::copy(other);
-	minkowski = other.minkowski;
-	pos_sig = other.pos_sig;
-}
-
 //////////
 // archiving
 //////////
@@ -106,7 +108,7 @@ DEFAULT_ARCHIVING(spinmetric)
 DEFAULT_UNARCHIVE(minkmetric)
 DEFAULT_UNARCHIVE(tensepsilon)
 
-minkmetric::minkmetric(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+minkmetric::minkmetric(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	n.find_bool("pos_sig", pos_sig);
 }
@@ -117,7 +119,7 @@ void minkmetric::archive(archive_node &n) const
 	n.add_bool("pos_sig", pos_sig);
 }
 
-tensepsilon::tensepsilon(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+tensepsilon::tensepsilon(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
 {
 	n.find_bool("minkowski", minkowski);
 	n.find_bool("pos_sig", pos_sig);
@@ -179,9 +181,24 @@ ex tensdelta::eval_indexed(const basic & i) const
 	const idx & i1 = ex_to<idx>(i.op(1));
 	const idx & i2 = ex_to<idx>(i.op(2));
 
-	// Trace of delta tensor is the dimension of the space
-	if (is_dummy_pair(i1, i2))
-		return i1.get_dim();
+	// The dimension of the indices must be equal, otherwise we use the minimal
+	// dimension
+	if (!i1.get_dim().is_equal(i2.get_dim())) {
+		ex min_dim = i1.minimal_dim(i2);
+		exmap m;
+		m[i1] = i1.replace_dim(min_dim);
+		m[i2] = i2.replace_dim(min_dim);
+		return i.subs(m, subs_options::no_pattern);
+	}
+
+	// Trace of delta tensor is the (effective) dimension of the space
+	if (is_dummy_pair(i1, i2)) {
+		try {
+			return i1.minimal_dim(i2);
+		} catch (std::exception &e) {
+			return i.hold();
+		}
+	}
 
 	// Numeric evaluation
 	if (static_cast<const indexed &>(i).all_index_values_are(info_flags::integer)) {
@@ -208,6 +225,16 @@ ex tensmetric::eval_indexed(const basic & i) const
 	const varidx & i1 = ex_to<varidx>(i.op(1));
 	const varidx & i2 = ex_to<varidx>(i.op(2));
 
+	// The dimension of the indices must be equal, otherwise we use the minimal
+	// dimension
+	if (!i1.get_dim().is_equal(i2.get_dim())) {
+		ex min_dim = i1.minimal_dim(i2);
+		exmap m;
+		m[i1] = i1.replace_dim(min_dim);
+		m[i2] = i2.replace_dim(min_dim);
+		return i.subs(m, subs_options::no_pattern);
+	}
+
 	// A metric tensor with one covariant and one contravariant index gets
 	// replaced by a delta tensor
 	if (i1.is_covariant() != i2.is_covariant())
@@ -293,15 +320,15 @@ ex tensepsilon::eval_indexed(const basic & i) const
 		// a canonic order but we can't assume what exactly that order is)
 		std::vector<int> v;
 		v.reserve(i.nops() - 1);
-		for (unsigned j=1; j<i.nops(); j++)
+		for (size_t j=1; j<i.nops(); j++)
 			v.push_back(ex_to<numeric>(ex_to<idx>(i.op(j)).get_value()).to_int());
 		int sign = permutation_sign(v.begin(), v.end());
 
 		// In a Minkowski space, check for covariant indices
 		if (minkowski) {
-			for (unsigned j=1; j<i.nops(); j++) {
+			for (size_t j=1; j<i.nops(); j++) {
 				const ex & x = i.op(j);
-				if (!is_ex_of_type(x, varidx))
+				if (!is_a<varidx>(x))
 					throw(std::runtime_error("indices of epsilon tensor in Minkowski space must be of type varidx"));
 				if (ex_to<varidx>(x).is_covariant())
 					if (ex_to<idx>(x).get_value().is_zero())
@@ -318,6 +345,46 @@ ex tensepsilon::eval_indexed(const basic & i) const
 	return i.hold();
 }
 
+bool tensor::replace_contr_index(exvector::iterator self, exvector::iterator other) const
+{
+	// Try to contract the first index
+	const idx *self_idx = &ex_to<idx>(self->op(1));
+	const idx *free_idx = &ex_to<idx>(self->op(2));
+	bool first_index_tried = false;
+
+again:
+	if (self_idx->is_symbolic()) {
+		for (size_t i=1; i<other->nops(); i++) {
+			const idx &other_idx = ex_to<idx>(other->op(i));
+			if (is_dummy_pair(*self_idx, other_idx)) {
+
+				// Contraction found, remove this tensor and substitute the
+				// index in the second object
+				try {
+					// minimal_dim() throws an exception when index dimensions are not comparable
+					ex min_dim = self_idx->minimal_dim(other_idx);
+					*other = other->subs(other_idx == free_idx->replace_dim(min_dim));
+					*self = _ex1; // *other is assigned first because assigning *self invalidates free_idx
+					return true;
+				} catch (std::exception &e) {
+					return false;
+				}
+			}
+		}
+	}
+
+	if (!first_index_tried) {
+
+		// No contraction with the first index found, try the second index
+		self_idx = &ex_to<idx>(self->op(2));
+		free_idx = &ex_to<idx>(self->op(1));
+		first_index_tried = true;
+		goto again;
+	}
+
+	return false;
+}
+
 /** Contraction of an indexed delta tensor with something else. */
 bool tensdelta::contract_with(exvector::iterator self, exvector::iterator other, exvector & v) const
 {
@@ -326,36 +393,9 @@ bool tensdelta::contract_with(exvector::iterator self, exvector::iterator other,
 	GINAC_ASSERT(self->nops() == 3);
 	GINAC_ASSERT(is_a<tensdelta>(self->op(0)));
 
-	// Try to contract first index
-	const idx *self_idx = &ex_to<idx>(self->op(1));
-	const idx *free_idx = &ex_to<idx>(self->op(2));
-	bool first_index_tried = false;
-
-again:
-	if (self_idx->is_symbolic()) {
-		for (unsigned i=1; i<other->nops(); i++) {
-			const idx &other_idx = ex_to<idx>(other->op(i));
-			if (is_dummy_pair(*self_idx, other_idx)) {
-
-				// Contraction found, remove delta tensor and substitute
-				// index in second object
-				*self = _ex1;
-				*other = other->subs(other_idx == *free_idx);
-				return true;
-			}
-		}
-	}
-
-	if (!first_index_tried) {
-
-		// No contraction with first index found, try second index
-		self_idx = &ex_to<idx>(self->op(2));
-		free_idx = &ex_to<idx>(self->op(1));
-		first_index_tried = true;
-		goto again;
-	}
-
-	return false;
+	// Replace the dummy index with this tensor's other index and remove
+	// the tensor (this is valid for contractions with all other tensors)
+	return replace_contr_index(self, other);
 }
 
 /** Contraction of an indexed metric tensor with something else. */
@@ -368,39 +408,12 @@ bool tensmetric::contract_with(exvector::iterator self, exvector::iterator other
 
 	// If contracting with the delta tensor, let the delta do it
 	// (don't raise/lower delta indices)
-	if (is_ex_of_type(other->op(0), tensdelta))
+	if (is_a<tensdelta>(other->op(0)))
 		return false;
 
-	// Try to contract first index
-	const idx *self_idx = &ex_to<idx>(self->op(1));
-	const idx *free_idx = &ex_to<idx>(self->op(2));
-	bool first_index_tried = false;
-
-again:
-	if (self_idx->is_symbolic()) {
-		for (unsigned i=1; i<other->nops(); i++) {
-			const idx &other_idx = ex_to<idx>(other->op(i));
-			if (is_dummy_pair(*self_idx, other_idx)) {
-
-				// Contraction found, remove metric tensor and substitute
-				// index in second object
-				*self = _ex1;
-				*other = other->subs(other_idx == *free_idx);
-				return true;
-			}
-		}
-	}
-
-	if (!first_index_tried) {
-
-		// No contraction with first index found, try second index
-		self_idx = &ex_to<idx>(self->op(2));
-		free_idx = &ex_to<idx>(self->op(1));
-		first_index_tried = true;
-		goto again;
-	}
-
-	return false;
+	// Replace the dummy index with this tensor's other index and remove
+	// the tensor
+	return replace_contr_index(self, other);
 }
 
 /** Contraction of an indexed spinor metric with something else. */
@@ -412,7 +425,7 @@ bool spinmetric::contract_with(exvector::iterator self, exvector::iterator other
 	GINAC_ASSERT(is_a<spinmetric>(self->op(0)));
 
 	// Contractions between spinor metrics
-	if (is_ex_of_type(other->op(0), spinmetric)) {
+	if (is_a<spinmetric>(other->op(0))) {
 		const idx &self_i1 = ex_to<idx>(self->op(1));
 		const idx &self_i2 = ex_to<idx>(self->op(2));
 		const idx &other_i1 = ex_to<idx>(other->op(1));
@@ -445,7 +458,7 @@ bool spinmetric::contract_with(exvector::iterator self, exvector::iterator other
 
 	// If contracting with the delta tensor, let the delta do it
 	// (don't raise/lower delta indices)
-	if (is_ex_of_type(other->op(0), tensdelta))
+	if (is_a<tensdelta>(other->op(0)))
 		return false;
 
 	// Try to contract first index
@@ -456,14 +469,15 @@ bool spinmetric::contract_with(exvector::iterator self, exvector::iterator other
 
 again:
 	if (self_idx->is_symbolic()) {
-		for (unsigned i=1; i<other->nops(); i++) {
+		for (size_t i=1; i<other->nops(); i++) {
 			const idx &other_idx = ex_to<idx>(other->op(i));
 			if (is_dummy_pair(*self_idx, other_idx)) {
 
 				// Contraction found, remove metric tensor and substitute
-				// index in second object
-				*self = (static_cast<const spinidx *>(self_idx)->is_covariant() ? sign : -sign);
+				// index in second object (assign *self last because this
+				// invalidates free_idx)
 				*other = other->subs(other_idx == *free_idx);
+				*self = (static_cast<const spinidx *>(self_idx)->is_covariant() ? sign : -sign);
 				return true;
 			}
 		}
@@ -488,63 +502,27 @@ bool tensepsilon::contract_with(exvector::iterator self, exvector::iterator othe
 	GINAC_ASSERT(is_a<indexed>(*self));
 	GINAC_ASSERT(is_a<indexed>(*other));
 	GINAC_ASSERT(is_a<tensepsilon>(self->op(0)));
-	unsigned num = self->nops() - 1;
+	size_t num = self->nops() - 1;
 
-	if (is_ex_exactly_of_type(other->op(0), tensepsilon) && num+1 == other->nops()) {
+	if (is_exactly_a<tensepsilon>(other->op(0)) && num+1 == other->nops()) {
 
 		// Contraction of two epsilon tensors is a determinant
-		ex dim = ex_to<idx>(self->op(1)).get_dim();
+		bool variance = is_a<varidx>(self->op(1));
 		matrix M(num, num);
-		for (int i=0; i<num; i++) {
-			for (int j=0; j<num; j++) {
+		for (size_t i=0; i<num; i++) {
+			for (size_t j=0; j<num; j++) {
 				if (minkowski)
 					M(i, j) = lorentz_g(self->op(i+1), other->op(j+1), pos_sig);
-				else
+				else if (variance)
 					M(i, j) = metric_tensor(self->op(i+1), other->op(j+1));
+				else
+					M(i, j) = delta_tensor(self->op(i+1), other->op(j+1));
 			}
 		}
 		int sign = minkowski ? -1 : 1;
 		*self = sign * M.determinant().simplify_indexed();
 		*other = _ex1;
 		return true;
-
-	} else if (other->return_type() == return_types::commutative) {
-
-#if 0
-		// This handles eps.i.j.k * p.j * p.k = 0
-		// Maybe something like this should go to simplify_indexed() because
-		// such relations are true for any antisymmetric tensors...
-		exvector c;
-
-		// Handle all indices of the epsilon tensor
-		for (int i=0; i<num; i++) {
-			ex idx = self->op(i+1);
-
-			// Look whether there's a contraction with this index
-			exvector::const_iterator ait, aitend = v.end();
-			for (ait = v.begin(); ait != aitend; ait++) {
-				if (ait == self)
-					continue;
-				if (is_a<indexed>(*ait) && ait->return_type() == return_types::commutative && ex_to<indexed>(*ait).has_dummy_index_for(idx) && ait->nops() == 2) {
-
-					// Yes, did we already have another contraction with the same base expression?
-					ex base = ait->op(0);
-					if (std::find_if(c.begin(), c.end(), bind2nd(ex_is_equal(), base)) == c.end()) {
-
-						// No, add the base expression to the list
-						c.push_back(base);
-
-					} else {
-
-						// Yes, the contraction is zero
-						*self = _ex0;
-						*other = _ex0;
-						return true;
-					}
-				}
-			}
-		}
-#endif
 	}
 
 	return false;
@@ -556,7 +534,7 @@ bool tensepsilon::contract_with(exvector::iterator self, exvector::iterator othe
 
 ex delta_tensor(const ex & i1, const ex & i2)
 {
-	if (!is_ex_of_type(i1, idx) || !is_ex_of_type(i2, idx))
+	if (!is_a<idx>(i1) || !is_a<idx>(i2))
 		throw(std::invalid_argument("indices of delta tensor must be of type idx"));
 
 	return indexed(tensdelta(), sy_symm(), i1, i2);
@@ -564,7 +542,7 @@ ex delta_tensor(const ex & i1, const ex & i2)
 
 ex metric_tensor(const ex & i1, const ex & i2)
 {
-	if (!is_ex_of_type(i1, varidx) || !is_ex_of_type(i2, varidx))
+	if (!is_a<varidx>(i1) || !is_a<varidx>(i2))
 		throw(std::invalid_argument("indices of metric tensor must be of type varidx"));
 
 	return indexed(tensmetric(), sy_symm(), i1, i2);
@@ -572,7 +550,7 @@ ex metric_tensor(const ex & i1, const ex & i2)
 
 ex lorentz_g(const ex & i1, const ex & i2, bool pos_sig)
 {
-	if (!is_ex_of_type(i1, varidx) || !is_ex_of_type(i2, varidx))
+	if (!is_a<varidx>(i1) || !is_a<varidx>(i2))
 		throw(std::invalid_argument("indices of metric tensor must be of type varidx"));
 
 	return indexed(minkmetric(pos_sig), sy_symm(), i1, i2);
@@ -580,7 +558,7 @@ ex lorentz_g(const ex & i1, const ex & i2, bool pos_sig)
 
 ex spinor_metric(const ex & i1, const ex & i2)
 {
-	if (!is_ex_of_type(i1, spinidx) || !is_ex_of_type(i2, spinidx))
+	if (!is_a<spinidx>(i1) || !is_a<spinidx>(i2))
 		throw(std::invalid_argument("indices of spinor metric must be of type spinidx"));
 	if (!ex_to<idx>(i1).get_dim().is_equal(2) || !ex_to<idx>(i2).get_dim().is_equal(2))
 		throw(std::runtime_error("index dimension for spinor metric must be 2"));
@@ -590,7 +568,7 @@ ex spinor_metric(const ex & i1, const ex & i2)
 
 ex epsilon_tensor(const ex & i1, const ex & i2)
 {
-	if (!is_ex_of_type(i1, idx) || !is_ex_of_type(i2, idx))
+	if (!is_a<idx>(i1) || !is_a<idx>(i2))
 		throw(std::invalid_argument("indices of epsilon tensor must be of type idx"));
 
 	ex dim = ex_to<idx>(i1).get_dim();
@@ -604,7 +582,7 @@ ex epsilon_tensor(const ex & i1, const ex & i2)
 
 ex epsilon_tensor(const ex & i1, const ex & i2, const ex & i3)
 {
-	if (!is_ex_of_type(i1, idx) || !is_ex_of_type(i2, idx) || !is_ex_of_type(i3, idx))
+	if (!is_a<idx>(i1) || !is_a<idx>(i2) || !is_a<idx>(i3))
 		throw(std::invalid_argument("indices of epsilon tensor must be of type idx"));
 
 	ex dim = ex_to<idx>(i1).get_dim();
@@ -618,7 +596,7 @@ ex epsilon_tensor(const ex & i1, const ex & i2, const ex & i3)
 
 ex lorentz_eps(const ex & i1, const ex & i2, const ex & i3, const ex & i4, bool pos_sig)
 {
-	if (!is_ex_of_type(i1, varidx) || !is_ex_of_type(i2, varidx) || !is_ex_of_type(i3, varidx) || !is_ex_of_type(i4, varidx))
+	if (!is_a<varidx>(i1) || !is_a<varidx>(i2) || !is_a<varidx>(i3) || !is_a<varidx>(i4))
 		throw(std::invalid_argument("indices of Lorentz epsilon tensor must be of type varidx"));
 
 	ex dim = ex_to<idx>(i1).get_dim();
@@ -630,16 +608,4 @@ ex lorentz_eps(const ex & i1, const ex & i2, const ex & i3, const ex & i4, bool
 	return indexed(tensepsilon(true, pos_sig), sy_anti(), i1, i2, i3, i4);
 }
 
-ex eps0123(const ex & i1, const ex & i2, const ex & i3, const ex & i4, bool pos_sig)
-{
-	if (!is_ex_of_type(i1, varidx) || !is_ex_of_type(i2, varidx) || !is_ex_of_type(i3, varidx) || !is_ex_of_type(i4, varidx))
-		throw(std::invalid_argument("indices of epsilon tensor must be of type varidx"));
-
-	ex dim = ex_to<idx>(i1).get_dim();
-	if (dim.is_equal(4))
-		return lorentz_eps(i1, i2, i3, i4, pos_sig);
-	else
-		return indexed(tensepsilon(true, pos_sig), sy_anti(), i1, i2, i3, i4);
-}
-
 } // namespace GiNaC
